/* Automatically @generated by tree-sitter */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 418
#define LARGE_STATE_COUNT 2
#define SYMBOL_COUNT 216
#define ALIAS_COUNT 0
#define TOKEN_COUNT 108
#define EXTERNAL_TOKEN_COUNT 10
#define FIELD_COUNT 9
#define MAX_ALIAS_SEQUENCE_LENGTH 5
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 19
#define SUPERTYPE_COUNT 0

enum ts_symbol_identifiers {
  aux_sym_from_instruction_token1 = 1,
  aux_sym_from_instruction_token2 = 2,
  aux_sym_run_instruction_token1 = 3,
  aux_sym_cmd_instruction_token1 = 4,
  aux_sym_label_instruction_token1 = 5,
  aux_sym_expose_instruction_token1 = 6,
  aux_sym_env_instruction_token1 = 7,
  aux_sym_add_instruction_token1 = 8,
  aux_sym_copy_instruction_token1 = 9,
  aux_sym_entrypoint_instruction_token1 = 10,
  aux_sym_volume_instruction_token1 = 11,
  aux_sym_user_instruction_token1 = 12,
  anon_sym_COLON = 13,
  aux_sym__user_name_or_group_token1 = 14,
  aux_sym__immediate_user_name_or_group_fragment_token1 = 15,
  aux_sym_workdir_instruction_token1 = 16,
  aux_sym_arg_instruction_token1 = 17,
  aux_sym_arg_pair_token1 = 18,
  anon_sym_EQ = 19,
  aux_sym_onbuild_instruction_token1 = 20,
  aux_sym_stopsignal_instruction_token1 = 21,
  aux_sym__stopsignal_value_token1 = 22,
  aux_sym__stopsignal_value_token2 = 23,
  aux_sym_healthcheck_instruction_token1 = 24,
  anon_sym_NONE = 25,
  aux_sym__healthcheck_raw_fragment_atom_token1 = 26,
  aux_sym__healthcheck_raw_fragment_atom_token2 = 27,
  aux_sym__healthcheck_raw_fragment_atom_token3 = 28,
  aux_sym__healthcheck_raw_fragment_atom_token4 = 29,
  aux_sym__healthcheck_raw_fragment_atom_token5 = 30,
  aux_sym__healthcheck_raw_fragment_atom_token6 = 31,
  aux_sym__healthcheck_raw_fragment_atom_token7 = 32,
  aux_sym__healthcheck_raw_fragment_atom_token8 = 33,
  aux_sym__healthcheck_raw_fragment_atom_token9 = 34,
  aux_sym__healthcheck_raw_fragment_atom_token10 = 35,
  aux_sym__healthcheck_raw_fragment_atom_token11 = 36,
  aux_sym__healthcheck_raw_fragment_atom_token12 = 37,
  aux_sym_shell_instruction_token1 = 38,
  aux_sym_maintainer_instruction_token1 = 39,
  aux_sym_maintainer_instruction_token2 = 40,
  aux_sym_cross_build_instruction_token1 = 41,
  anon_sym_LF = 42,
  aux_sym_path_token1 = 43,
  aux_sym_path_token2 = 44,
  aux_sym_path_with_heredoc_token1 = 45,
  anon_sym_DOLLAR = 46,
  anon_sym_DOLLAR2 = 47,
  anon_sym_LBRACE = 48,
  aux_sym__expansion_body_token1 = 49,
  anon_sym_RBRACE = 50,
  sym_variable = 51,
  sym__env_trailing_single_quotes = 52,
  aux_sym__spaced_env_pair_token1 = 53,
  aux_sym__env_key_token1 = 54,
  aux_sym_expose_port_token1 = 55,
  anon_sym_SLASHtcp = 56,
  anon_sym_SLASHudp = 57,
  aux_sym__label_key_token1 = 58,
  aux_sym_image_name_token1 = 59,
  aux_sym_image_name_token2 = 60,
  aux_sym_image_tag_token1 = 61,
  anon_sym_AT = 62,
  aux_sym_image_digest_token1 = 63,
  anon_sym_DASH_DASH = 64,
  aux_sym_param_token1 = 65,
  aux_sym_param_token2 = 66,
  anon_sym_mount = 67,
  anon_sym_COMMA = 68,
  aux_sym_mount_param_param_token1 = 69,
  aux_sym_image_alias_token1 = 70,
  aux_sym_image_alias_token2 = 71,
  aux_sym__shell_fragment_atom_token1 = 72,
  anon_sym_DQUOTE = 73,
  aux_sym__shell_double_quoted_fragment_token1 = 74,
  aux_sym__shell_double_quoted_fragment_token2 = 75,
  anon_sym_DQUOTE2 = 76,
  anon_sym_SQUOTE = 77,
  aux_sym__shell_single_quoted_fragment_token1 = 78,
  anon_sym_SQUOTE2 = 79,
  anon_sym_LBRACK = 80,
  anon_sym_COMMA2 = 81,
  anon_sym_RBRACK = 82,
  aux_sym_json_string_token1 = 83,
  anon_sym_BQUOTE = 84,
  sym_json_escape_sequence = 85,
  aux_sym_double_quoted_string_token1 = 86,
  aux_sym_double_quoted_string_token2 = 87,
  anon_sym_BSLASH = 88,
  anon_sym_BQUOTE2 = 89,
  aux_sym_single_quoted_string_token1 = 90,
  aux_sym_unquoted_string_token1 = 91,
  aux_sym_unquoted_string_token2 = 92,
  aux_sym_unquoted_string_token3 = 93,
  aux_sym__spaced_label_value_fragment_token1 = 94,
  sym_double_quoted_escape_sequence = 95,
  sym_single_quoted_escape_sequence = 96,
  sym__non_newline_whitespace = 97,
  sym_comment = 98,
  sym_line_continuation = 99,
  sym_required_line_continuation = 100,
  sym__newline = 101,
  sym__invalid_json_array_shell_command = 102,
  sym_heredoc_marker = 103,
  sym__heredoc_line = 104,
  sym_heredoc_end = 105,
  sym__heredoc_nl = 106,
  sym_error_sentinel = 107,
  sym_source_file = 108,
  sym__instruction = 109,
  sym_from_instruction = 110,
  sym_run_instruction = 111,
  sym_cmd_instruction = 112,
  sym_label_instruction = 113,
  sym_expose_instruction = 114,
  sym_env_instruction = 115,
  sym_add_instruction = 116,
  sym_copy_instruction = 117,
  sym_entrypoint_instruction = 118,
  sym_volume_instruction = 119,
  sym_user_instruction = 120,
  sym__user_name_or_group = 121,
  aux_sym__immediate_user_name_or_group = 122,
  sym__immediate_user_name_or_group_fragment = 123,
  sym_workdir_instruction = 124,
  sym_arg_instruction = 125,
  sym_arg_pair = 126,
  sym_onbuild_instruction = 127,
  sym_stopsignal_instruction = 128,
  sym__stopsignal_value = 129,
  sym_healthcheck_instruction = 130,
  sym__healthcheck_cmd_instruction = 131,
  sym__healthcheck_raw_command = 132,
  sym__healthcheck_raw_fragment = 133,
  sym__healthcheck_raw_fragment_atom = 134,
  sym_shell_instruction = 135,
  sym_maintainer_instruction = 136,
  sym_cross_build_instruction = 137,
  sym_heredoc_block = 138,
  sym_path = 139,
  sym_path_with_heredoc = 140,
  sym_expansion = 141,
  sym__immediate_expansion = 142,
  sym__imm_expansion = 143,
  sym__expansion_body = 144,
  sym_env_pair = 145,
  sym__env_assignment_value = 146,
  sym__env_single_quoted_string_with_trailing_quote = 147,
  sym__spaced_env_pair = 148,
  sym__env_key = 149,
  sym_expose_port = 150,
  sym_label_pair = 151,
  sym__spaced_label_pair = 152,
  sym__label_key = 153,
  sym_image_spec = 154,
  sym_image_name = 155,
  sym_image_tag = 156,
  sym_image_digest = 157,
  sym_param = 158,
  sym_mount_param = 159,
  sym_mount_param_param = 160,
  sym_image_alias = 161,
  sym__json_or_shell_command = 162,
  sym_shell_command = 163,
  aux_sym__shell_fragment = 164,
  sym__shell_fragment_atom = 165,
  sym__shell_double_quoted_fragment = 166,
  sym__shell_single_quoted_fragment = 167,
  sym_json_string_array = 168,
  sym_json_string = 169,
  sym_double_quoted_string = 170,
  sym_single_quoted_string = 171,
  sym_unquoted_string = 172,
  sym__spaced_label_value = 173,
  sym__continued_spaced_label_value = 174,
  sym__spaced_label_value_continuation = 175,
  sym__spaced_label_value_atom = 176,
  sym__spaced_label_value_fragment = 177,
  sym__spaced_env_value = 178,
  sym__continued_spaced_env_value = 179,
  sym__spaced_env_value_continuation = 180,
  sym__spaced_env_value_atom = 181,
  sym__spaced_env_value_fragment = 182,
  aux_sym_source_file_repeat1 = 183,
  aux_sym_source_file_repeat2 = 184,
  aux_sym_run_instruction_repeat1 = 185,
  aux_sym_run_instruction_repeat2 = 186,
  aux_sym_label_instruction_repeat1 = 187,
  aux_sym_expose_instruction_repeat1 = 188,
  aux_sym_env_instruction_repeat1 = 189,
  aux_sym_add_instruction_repeat1 = 190,
  aux_sym_add_instruction_repeat2 = 191,
  aux_sym_volume_instruction_repeat1 = 192,
  aux_sym__user_name_or_group_repeat1 = 193,
  aux_sym_arg_instruction_repeat1 = 194,
  aux_sym__stopsignal_value_repeat1 = 195,
  aux_sym__healthcheck_raw_command_repeat1 = 196,
  aux_sym__healthcheck_raw_fragment_repeat1 = 197,
  aux_sym_heredoc_block_repeat1 = 198,
  aux_sym_path_repeat1 = 199,
  aux_sym_image_name_repeat1 = 200,
  aux_sym_image_tag_repeat1 = 201,
  aux_sym_image_digest_repeat1 = 202,
  aux_sym_mount_param_repeat1 = 203,
  aux_sym_image_alias_repeat1 = 204,
  aux_sym__shell_double_quoted_fragment_repeat1 = 205,
  aux_sym__shell_single_quoted_fragment_repeat1 = 206,
  aux_sym_json_string_array_repeat1 = 207,
  aux_sym_json_string_repeat1 = 208,
  aux_sym_double_quoted_string_repeat1 = 209,
  aux_sym_single_quoted_string_repeat1 = 210,
  aux_sym_unquoted_string_repeat1 = 211,
  aux_sym__spaced_label_value_repeat1 = 212,
  aux_sym__spaced_label_value_repeat2 = 213,
  aux_sym__spaced_env_value_repeat1 = 214,
  aux_sym__spaced_env_value_repeat2 = 215,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [aux_sym_from_instruction_token1] = "FROM",
  [aux_sym_from_instruction_token2] = "AS",
  [aux_sym_run_instruction_token1] = "RUN",
  [aux_sym_cmd_instruction_token1] = "CMD",
  [aux_sym_label_instruction_token1] = "LABEL",
  [aux_sym_expose_instruction_token1] = "EXPOSE",
  [aux_sym_env_instruction_token1] = "ENV",
  [aux_sym_add_instruction_token1] = "ADD",
  [aux_sym_copy_instruction_token1] = "COPY",
  [aux_sym_entrypoint_instruction_token1] = "ENTRYPOINT",
  [aux_sym_volume_instruction_token1] = "VOLUME",
  [aux_sym_user_instruction_token1] = "USER",
  [anon_sym_COLON] = ":",
  [aux_sym__user_name_or_group_token1] = "_user_name_or_group_token1",
  [aux_sym__immediate_user_name_or_group_fragment_token1] = "_immediate_user_name_or_group_fragment_token1",
  [aux_sym_workdir_instruction_token1] = "WORKDIR",
  [aux_sym_arg_instruction_token1] = "ARG",
  [aux_sym_arg_pair_token1] = "unquoted_string",
  [anon_sym_EQ] = "=",
  [aux_sym_onbuild_instruction_token1] = "ONBUILD",
  [aux_sym_stopsignal_instruction_token1] = "STOPSIGNAL",
  [aux_sym__stopsignal_value_token1] = "_stopsignal_value_token1",
  [aux_sym__stopsignal_value_token2] = "_stopsignal_value_token2",
  [aux_sym_healthcheck_instruction_token1] = "HEALTHCHECK",
  [anon_sym_NONE] = "NONE",
  [aux_sym__healthcheck_raw_fragment_atom_token1] = "_healthcheck_raw_fragment_atom_token1",
  [aux_sym__healthcheck_raw_fragment_atom_token2] = "_healthcheck_raw_fragment_atom_token2",
  [aux_sym__healthcheck_raw_fragment_atom_token3] = "_healthcheck_raw_fragment_atom_token3",
  [aux_sym__healthcheck_raw_fragment_atom_token4] = "_healthcheck_raw_fragment_atom_token4",
  [aux_sym__healthcheck_raw_fragment_atom_token5] = "_healthcheck_raw_fragment_atom_token5",
  [aux_sym__healthcheck_raw_fragment_atom_token6] = "_healthcheck_raw_fragment_atom_token6",
  [aux_sym__healthcheck_raw_fragment_atom_token7] = "_healthcheck_raw_fragment_atom_token7",
  [aux_sym__healthcheck_raw_fragment_atom_token8] = "_healthcheck_raw_fragment_atom_token8",
  [aux_sym__healthcheck_raw_fragment_atom_token9] = "_healthcheck_raw_fragment_atom_token9",
  [aux_sym__healthcheck_raw_fragment_atom_token10] = "_healthcheck_raw_fragment_atom_token10",
  [aux_sym__healthcheck_raw_fragment_atom_token11] = "_healthcheck_raw_fragment_atom_token11",
  [aux_sym__healthcheck_raw_fragment_atom_token12] = "_healthcheck_raw_fragment_atom_token12",
  [aux_sym_shell_instruction_token1] = "SHELL",
  [aux_sym_maintainer_instruction_token1] = "MAINTAINER",
  [aux_sym_maintainer_instruction_token2] = "maintainer_instruction_token2",
  [aux_sym_cross_build_instruction_token1] = "CROSS_BUILD",
  [anon_sym_LF] = "\n",
  [aux_sym_path_token1] = "path_token1",
  [aux_sym_path_token2] = "path_token2",
  [aux_sym_path_with_heredoc_token1] = "path_with_heredoc_token1",
  [anon_sym_DOLLAR] = "$",
  [anon_sym_DOLLAR2] = "$",
  [anon_sym_LBRACE] = "{",
  [aux_sym__expansion_body_token1] = "variable",
  [anon_sym_RBRACE] = "}",
  [sym_variable] = "variable",
  [sym__env_trailing_single_quotes] = "_env_trailing_single_quotes",
  [aux_sym__spaced_env_pair_token1] = "_spaced_env_pair_token1",
  [aux_sym__env_key_token1] = "unquoted_string",
  [aux_sym_expose_port_token1] = "expose_port_token1",
  [anon_sym_SLASHtcp] = "/tcp",
  [anon_sym_SLASHudp] = "/udp",
  [aux_sym__label_key_token1] = "unquoted_string",
  [aux_sym_image_name_token1] = "image_name_token1",
  [aux_sym_image_name_token2] = "image_name_token2",
  [aux_sym_image_tag_token1] = "image_tag_token1",
  [anon_sym_AT] = "@",
  [aux_sym_image_digest_token1] = "image_digest_token1",
  [anon_sym_DASH_DASH] = "--",
  [aux_sym_param_token1] = "param_token1",
  [aux_sym_param_token2] = "param_token2",
  [anon_sym_mount] = "mount",
  [anon_sym_COMMA] = ",",
  [aux_sym_mount_param_param_token1] = "mount_param_param_token1",
  [aux_sym_image_alias_token1] = "image_alias_token1",
  [aux_sym_image_alias_token2] = "image_alias_token2",
  [aux_sym__shell_fragment_atom_token1] = "_shell_fragment_atom_token1",
  [anon_sym_DQUOTE] = "\"",
  [aux_sym__shell_double_quoted_fragment_token1] = "_shell_double_quoted_fragment_token1",
  [aux_sym__shell_double_quoted_fragment_token2] = "_shell_double_quoted_fragment_token2",
  [anon_sym_DQUOTE2] = "\"",
  [anon_sym_SQUOTE] = "'",
  [aux_sym__shell_single_quoted_fragment_token1] = "_shell_single_quoted_fragment_token1",
  [anon_sym_SQUOTE2] = "'",
  [anon_sym_LBRACK] = "[",
  [anon_sym_COMMA2] = ",",
  [anon_sym_RBRACK] = "]",
  [aux_sym_json_string_token1] = "json_string_token1",
  [anon_sym_BQUOTE] = "`",
  [sym_json_escape_sequence] = "escape_sequence",
  [aux_sym_double_quoted_string_token1] = "double_quoted_string_token1",
  [aux_sym_double_quoted_string_token2] = "double_quoted_string_token2",
  [anon_sym_BSLASH] = "\\",
  [anon_sym_BQUOTE2] = "`",
  [aux_sym_single_quoted_string_token1] = "single_quoted_string_token1",
  [aux_sym_unquoted_string_token1] = "unquoted_string_token1",
  [aux_sym_unquoted_string_token2] = "unquoted_string_token2",
  [aux_sym_unquoted_string_token3] = "unquoted_string_token3",
  [aux_sym__spaced_label_value_fragment_token1] = "_spaced_label_value_fragment_token1",
  [sym_double_quoted_escape_sequence] = "escape_sequence",
  [sym_single_quoted_escape_sequence] = "escape_sequence",
  [sym__non_newline_whitespace] = "_non_newline_whitespace",
  [sym_comment] = "comment",
  [sym_line_continuation] = "line_continuation",
  [sym_required_line_continuation] = "line_continuation",
  [sym__newline] = "_newline",
  [sym__invalid_json_array_shell_command] = "shell_command",
  [sym_heredoc_marker] = "heredoc_marker",
  [sym__heredoc_line] = "_heredoc_line",
  [sym_heredoc_end] = "heredoc_end",
  [sym__heredoc_nl] = "_heredoc_nl",
  [sym_error_sentinel] = "error_sentinel",
  [sym_source_file] = "source_file",
  [sym__instruction] = "_instruction",
  [sym_from_instruction] = "from_instruction",
  [sym_run_instruction] = "run_instruction",
  [sym_cmd_instruction] = "cmd_instruction",
  [sym_label_instruction] = "label_instruction",
  [sym_expose_instruction] = "expose_instruction",
  [sym_env_instruction] = "env_instruction",
  [sym_add_instruction] = "add_instruction",
  [sym_copy_instruction] = "copy_instruction",
  [sym_entrypoint_instruction] = "entrypoint_instruction",
  [sym_volume_instruction] = "volume_instruction",
  [sym_user_instruction] = "user_instruction",
  [sym__user_name_or_group] = "unquoted_string",
  [aux_sym__immediate_user_name_or_group] = "_immediate_user_name_or_group",
  [sym__immediate_user_name_or_group_fragment] = "_immediate_user_name_or_group_fragment",
  [sym_workdir_instruction] = "workdir_instruction",
  [sym_arg_instruction] = "arg_instruction",
  [sym_arg_pair] = "arg_pair",
  [sym_onbuild_instruction] = "onbuild_instruction",
  [sym_stopsignal_instruction] = "stopsignal_instruction",
  [sym__stopsignal_value] = "_stopsignal_value",
  [sym_healthcheck_instruction] = "healthcheck_instruction",
  [sym__healthcheck_cmd_instruction] = "cmd_instruction",
  [sym__healthcheck_raw_command] = "shell_command",
  [sym__healthcheck_raw_fragment] = "_healthcheck_raw_fragment",
  [sym__healthcheck_raw_fragment_atom] = "_healthcheck_raw_fragment_atom",
  [sym_shell_instruction] = "shell_instruction",
  [sym_maintainer_instruction] = "maintainer_instruction",
  [sym_cross_build_instruction] = "cross_build_instruction",
  [sym_heredoc_block] = "heredoc_block",
  [sym_path] = "path",
  [sym_path_with_heredoc] = "path",
  [sym_expansion] = "expansion",
  [sym__immediate_expansion] = "_immediate_expansion",
  [sym__imm_expansion] = "expansion",
  [sym__expansion_body] = "_expansion_body",
  [sym_env_pair] = "env_pair",
  [sym__env_assignment_value] = "_env_assignment_value",
  [sym__env_single_quoted_string_with_trailing_quote] = "_env_single_quoted_string_with_trailing_quote",
  [sym__spaced_env_pair] = "env_pair",
  [sym__env_key] = "_env_key",
  [sym_expose_port] = "expose_port",
  [sym_label_pair] = "label_pair",
  [sym__spaced_label_pair] = "label_pair",
  [sym__label_key] = "_label_key",
  [sym_image_spec] = "image_spec",
  [sym_image_name] = "image_name",
  [sym_image_tag] = "image_tag",
  [sym_image_digest] = "image_digest",
  [sym_param] = "param",
  [sym_mount_param] = "mount_param",
  [sym_mount_param_param] = "mount_param_param",
  [sym_image_alias] = "image_alias",
  [sym__json_or_shell_command] = "_json_or_shell_command",
  [sym_shell_command] = "shell_command",
  [aux_sym__shell_fragment] = "_shell_fragment",
  [sym__shell_fragment_atom] = "_shell_fragment_atom",
  [sym__shell_double_quoted_fragment] = "_shell_double_quoted_fragment",
  [sym__shell_single_quoted_fragment] = "_shell_single_quoted_fragment",
  [sym_json_string_array] = "json_string_array",
  [sym_json_string] = "json_string",
  [sym_double_quoted_string] = "double_quoted_string",
  [sym_single_quoted_string] = "single_quoted_string",
  [sym_unquoted_string] = "unquoted_string",
  [sym__spaced_label_value] = "unquoted_string",
  [sym__continued_spaced_label_value] = "unquoted_string",
  [sym__spaced_label_value_continuation] = "_spaced_label_value_continuation",
  [sym__spaced_label_value_atom] = "_spaced_label_value_atom",
  [sym__spaced_label_value_fragment] = "_spaced_label_value_fragment",
  [sym__spaced_env_value] = "unquoted_string",
  [sym__continued_spaced_env_value] = "unquoted_string",
  [sym__spaced_env_value_continuation] = "_spaced_env_value_continuation",
  [sym__spaced_env_value_atom] = "_spaced_env_value_atom",
  [sym__spaced_env_value_fragment] = "_spaced_env_value_fragment",
  [aux_sym_source_file_repeat1] = "source_file_repeat1",
  [aux_sym_source_file_repeat2] = "source_file_repeat2",
  [aux_sym_run_instruction_repeat1] = "run_instruction_repeat1",
  [aux_sym_run_instruction_repeat2] = "run_instruction_repeat2",
  [aux_sym_label_instruction_repeat1] = "label_instruction_repeat1",
  [aux_sym_expose_instruction_repeat1] = "expose_instruction_repeat1",
  [aux_sym_env_instruction_repeat1] = "env_instruction_repeat1",
  [aux_sym_add_instruction_repeat1] = "add_instruction_repeat1",
  [aux_sym_add_instruction_repeat2] = "add_instruction_repeat2",
  [aux_sym_volume_instruction_repeat1] = "volume_instruction_repeat1",
  [aux_sym__user_name_or_group_repeat1] = "_user_name_or_group_repeat1",
  [aux_sym_arg_instruction_repeat1] = "arg_instruction_repeat1",
  [aux_sym__stopsignal_value_repeat1] = "_stopsignal_value_repeat1",
  [aux_sym__healthcheck_raw_command_repeat1] = "_healthcheck_raw_command_repeat1",
  [aux_sym__healthcheck_raw_fragment_repeat1] = "_healthcheck_raw_fragment_repeat1",
  [aux_sym_heredoc_block_repeat1] = "heredoc_block_repeat1",
  [aux_sym_path_repeat1] = "path_repeat1",
  [aux_sym_image_name_repeat1] = "image_name_repeat1",
  [aux_sym_image_tag_repeat1] = "image_tag_repeat1",
  [aux_sym_image_digest_repeat1] = "image_digest_repeat1",
  [aux_sym_mount_param_repeat1] = "mount_param_repeat1",
  [aux_sym_image_alias_repeat1] = "image_alias_repeat1",
  [aux_sym__shell_double_quoted_fragment_repeat1] = "_shell_double_quoted_fragment_repeat1",
  [aux_sym__shell_single_quoted_fragment_repeat1] = "_shell_single_quoted_fragment_repeat1",
  [aux_sym_json_string_array_repeat1] = "json_string_array_repeat1",
  [aux_sym_json_string_repeat1] = "json_string_repeat1",
  [aux_sym_double_quoted_string_repeat1] = "double_quoted_string_repeat1",
  [aux_sym_single_quoted_string_repeat1] = "single_quoted_string_repeat1",
  [aux_sym_unquoted_string_repeat1] = "unquoted_string_repeat1",
  [aux_sym__spaced_label_value_repeat1] = "_spaced_label_value_repeat1",
  [aux_sym__spaced_label_value_repeat2] = "_spaced_label_value_repeat2",
  [aux_sym__spaced_env_value_repeat1] = "_spaced_env_value_repeat1",
  [aux_sym__spaced_env_value_repeat2] = "_spaced_env_value_repeat2",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [aux_sym_from_instruction_token1] = aux_sym_from_instruction_token1,
  [aux_sym_from_instruction_token2] = aux_sym_from_instruction_token2,
  [aux_sym_run_instruction_token1] = aux_sym_run_instruction_token1,
  [aux_sym_cmd_instruction_token1] = aux_sym_cmd_instruction_token1,
  [aux_sym_label_instruction_token1] = aux_sym_label_instruction_token1,
  [aux_sym_expose_instruction_token1] = aux_sym_expose_instruction_token1,
  [aux_sym_env_instruction_token1] = aux_sym_env_instruction_token1,
  [aux_sym_add_instruction_token1] = aux_sym_add_instruction_token1,
  [aux_sym_copy_instruction_token1] = aux_sym_copy_instruction_token1,
  [aux_sym_entrypoint_instruction_token1] = aux_sym_entrypoint_instruction_token1,
  [aux_sym_volume_instruction_token1] = aux_sym_volume_instruction_token1,
  [aux_sym_user_instruction_token1] = aux_sym_user_instruction_token1,
  [anon_sym_COLON] = anon_sym_COLON,
  [aux_sym__user_name_or_group_token1] = aux_sym__user_name_or_group_token1,
  [aux_sym__immediate_user_name_or_group_fragment_token1] = aux_sym__immediate_user_name_or_group_fragment_token1,
  [aux_sym_workdir_instruction_token1] = aux_sym_workdir_instruction_token1,
  [aux_sym_arg_instruction_token1] = aux_sym_arg_instruction_token1,
  [aux_sym_arg_pair_token1] = sym_unquoted_string,
  [anon_sym_EQ] = anon_sym_EQ,
  [aux_sym_onbuild_instruction_token1] = aux_sym_onbuild_instruction_token1,
  [aux_sym_stopsignal_instruction_token1] = aux_sym_stopsignal_instruction_token1,
  [aux_sym__stopsignal_value_token1] = aux_sym__stopsignal_value_token1,
  [aux_sym__stopsignal_value_token2] = aux_sym__stopsignal_value_token2,
  [aux_sym_healthcheck_instruction_token1] = aux_sym_healthcheck_instruction_token1,
  [anon_sym_NONE] = anon_sym_NONE,
  [aux_sym__healthcheck_raw_fragment_atom_token1] = aux_sym__healthcheck_raw_fragment_atom_token1,
  [aux_sym__healthcheck_raw_fragment_atom_token2] = aux_sym__healthcheck_raw_fragment_atom_token2,
  [aux_sym__healthcheck_raw_fragment_atom_token3] = aux_sym__healthcheck_raw_fragment_atom_token3,
  [aux_sym__healthcheck_raw_fragment_atom_token4] = aux_sym__healthcheck_raw_fragment_atom_token4,
  [aux_sym__healthcheck_raw_fragment_atom_token5] = aux_sym__healthcheck_raw_fragment_atom_token5,
  [aux_sym__healthcheck_raw_fragment_atom_token6] = aux_sym__healthcheck_raw_fragment_atom_token6,
  [aux_sym__healthcheck_raw_fragment_atom_token7] = aux_sym__healthcheck_raw_fragment_atom_token7,
  [aux_sym__healthcheck_raw_fragment_atom_token8] = aux_sym__healthcheck_raw_fragment_atom_token8,
  [aux_sym__healthcheck_raw_fragment_atom_token9] = aux_sym__healthcheck_raw_fragment_atom_token9,
  [aux_sym__healthcheck_raw_fragment_atom_token10] = aux_sym__healthcheck_raw_fragment_atom_token10,
  [aux_sym__healthcheck_raw_fragment_atom_token11] = aux_sym__healthcheck_raw_fragment_atom_token11,
  [aux_sym__healthcheck_raw_fragment_atom_token12] = aux_sym__healthcheck_raw_fragment_atom_token12,
  [aux_sym_shell_instruction_token1] = aux_sym_shell_instruction_token1,
  [aux_sym_maintainer_instruction_token1] = aux_sym_maintainer_instruction_token1,
  [aux_sym_maintainer_instruction_token2] = aux_sym_maintainer_instruction_token2,
  [aux_sym_cross_build_instruction_token1] = aux_sym_cross_build_instruction_token1,
  [anon_sym_LF] = anon_sym_LF,
  [aux_sym_path_token1] = aux_sym_path_token1,
  [aux_sym_path_token2] = aux_sym_path_token2,
  [aux_sym_path_with_heredoc_token1] = aux_sym_path_with_heredoc_token1,
  [anon_sym_DOLLAR] = anon_sym_DOLLAR,
  [anon_sym_DOLLAR2] = anon_sym_DOLLAR,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [aux_sym__expansion_body_token1] = sym_variable,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [sym_variable] = sym_variable,
  [sym__env_trailing_single_quotes] = sym__env_trailing_single_quotes,
  [aux_sym__spaced_env_pair_token1] = aux_sym__spaced_env_pair_token1,
  [aux_sym__env_key_token1] = sym_unquoted_string,
  [aux_sym_expose_port_token1] = aux_sym_expose_port_token1,
  [anon_sym_SLASHtcp] = anon_sym_SLASHtcp,
  [anon_sym_SLASHudp] = anon_sym_SLASHudp,
  [aux_sym__label_key_token1] = sym_unquoted_string,
  [aux_sym_image_name_token1] = aux_sym_image_name_token1,
  [aux_sym_image_name_token2] = aux_sym_image_name_token2,
  [aux_sym_image_tag_token1] = aux_sym_image_tag_token1,
  [anon_sym_AT] = anon_sym_AT,
  [aux_sym_image_digest_token1] = aux_sym_image_digest_token1,
  [anon_sym_DASH_DASH] = anon_sym_DASH_DASH,
  [aux_sym_param_token1] = aux_sym_param_token1,
  [aux_sym_param_token2] = aux_sym_param_token2,
  [anon_sym_mount] = anon_sym_mount,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [aux_sym_mount_param_param_token1] = aux_sym_mount_param_param_token1,
  [aux_sym_image_alias_token1] = aux_sym_image_alias_token1,
  [aux_sym_image_alias_token2] = aux_sym_image_alias_token2,
  [aux_sym__shell_fragment_atom_token1] = aux_sym__shell_fragment_atom_token1,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [aux_sym__shell_double_quoted_fragment_token1] = aux_sym__shell_double_quoted_fragment_token1,
  [aux_sym__shell_double_quoted_fragment_token2] = aux_sym__shell_double_quoted_fragment_token2,
  [anon_sym_DQUOTE2] = anon_sym_DQUOTE,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [aux_sym__shell_single_quoted_fragment_token1] = aux_sym__shell_single_quoted_fragment_token1,
  [anon_sym_SQUOTE2] = anon_sym_SQUOTE,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_COMMA2] = anon_sym_COMMA,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [aux_sym_json_string_token1] = aux_sym_json_string_token1,
  [anon_sym_BQUOTE] = anon_sym_BQUOTE,
  [sym_json_escape_sequence] = sym_json_escape_sequence,
  [aux_sym_double_quoted_string_token1] = aux_sym_double_quoted_string_token1,
  [aux_sym_double_quoted_string_token2] = aux_sym_double_quoted_string_token2,
  [anon_sym_BSLASH] = anon_sym_BSLASH,
  [anon_sym_BQUOTE2] = anon_sym_BQUOTE,
  [aux_sym_single_quoted_string_token1] = aux_sym_single_quoted_string_token1,
  [aux_sym_unquoted_string_token1] = aux_sym_unquoted_string_token1,
  [aux_sym_unquoted_string_token2] = aux_sym_unquoted_string_token2,
  [aux_sym_unquoted_string_token3] = aux_sym_unquoted_string_token3,
  [aux_sym__spaced_label_value_fragment_token1] = aux_sym__spaced_label_value_fragment_token1,
  [sym_double_quoted_escape_sequence] = sym_json_escape_sequence,
  [sym_single_quoted_escape_sequence] = sym_json_escape_sequence,
  [sym__non_newline_whitespace] = sym__non_newline_whitespace,
  [sym_comment] = sym_comment,
  [sym_line_continuation] = sym_line_continuation,
  [sym_required_line_continuation] = sym_line_continuation,
  [sym__newline] = sym__newline,
  [sym__invalid_json_array_shell_command] = sym_shell_command,
  [sym_heredoc_marker] = sym_heredoc_marker,
  [sym__heredoc_line] = sym__heredoc_line,
  [sym_heredoc_end] = sym_heredoc_end,
  [sym__heredoc_nl] = sym__heredoc_nl,
  [sym_error_sentinel] = sym_error_sentinel,
  [sym_source_file] = sym_source_file,
  [sym__instruction] = sym__instruction,
  [sym_from_instruction] = sym_from_instruction,
  [sym_run_instruction] = sym_run_instruction,
  [sym_cmd_instruction] = sym_cmd_instruction,
  [sym_label_instruction] = sym_label_instruction,
  [sym_expose_instruction] = sym_expose_instruction,
  [sym_env_instruction] = sym_env_instruction,
  [sym_add_instruction] = sym_add_instruction,
  [sym_copy_instruction] = sym_copy_instruction,
  [sym_entrypoint_instruction] = sym_entrypoint_instruction,
  [sym_volume_instruction] = sym_volume_instruction,
  [sym_user_instruction] = sym_user_instruction,
  [sym__user_name_or_group] = sym_unquoted_string,
  [aux_sym__immediate_user_name_or_group] = aux_sym__immediate_user_name_or_group,
  [sym__immediate_user_name_or_group_fragment] = sym__immediate_user_name_or_group_fragment,
  [sym_workdir_instruction] = sym_workdir_instruction,
  [sym_arg_instruction] = sym_arg_instruction,
  [sym_arg_pair] = sym_arg_pair,
  [sym_onbuild_instruction] = sym_onbuild_instruction,
  [sym_stopsignal_instruction] = sym_stopsignal_instruction,
  [sym__stopsignal_value] = sym__stopsignal_value,
  [sym_healthcheck_instruction] = sym_healthcheck_instruction,
  [sym__healthcheck_cmd_instruction] = sym_cmd_instruction,
  [sym__healthcheck_raw_command] = sym_shell_command,
  [sym__healthcheck_raw_fragment] = sym__healthcheck_raw_fragment,
  [sym__healthcheck_raw_fragment_atom] = sym__healthcheck_raw_fragment_atom,
  [sym_shell_instruction] = sym_shell_instruction,
  [sym_maintainer_instruction] = sym_maintainer_instruction,
  [sym_cross_build_instruction] = sym_cross_build_instruction,
  [sym_heredoc_block] = sym_heredoc_block,
  [sym_path] = sym_path,
  [sym_path_with_heredoc] = sym_path,
  [sym_expansion] = sym_expansion,
  [sym__immediate_expansion] = sym__immediate_expansion,
  [sym__imm_expansion] = sym_expansion,
  [sym__expansion_body] = sym__expansion_body,
  [sym_env_pair] = sym_env_pair,
  [sym__env_assignment_value] = sym__env_assignment_value,
  [sym__env_single_quoted_string_with_trailing_quote] = sym__env_single_quoted_string_with_trailing_quote,
  [sym__spaced_env_pair] = sym_env_pair,
  [sym__env_key] = sym__env_key,
  [sym_expose_port] = sym_expose_port,
  [sym_label_pair] = sym_label_pair,
  [sym__spaced_label_pair] = sym_label_pair,
  [sym__label_key] = sym__label_key,
  [sym_image_spec] = sym_image_spec,
  [sym_image_name] = sym_image_name,
  [sym_image_tag] = sym_image_tag,
  [sym_image_digest] = sym_image_digest,
  [sym_param] = sym_param,
  [sym_mount_param] = sym_mount_param,
  [sym_mount_param_param] = sym_mount_param_param,
  [sym_image_alias] = sym_image_alias,
  [sym__json_or_shell_command] = sym__json_or_shell_command,
  [sym_shell_command] = sym_shell_command,
  [aux_sym__shell_fragment] = aux_sym__shell_fragment,
  [sym__shell_fragment_atom] = sym__shell_fragment_atom,
  [sym__shell_double_quoted_fragment] = sym__shell_double_quoted_fragment,
  [sym__shell_single_quoted_fragment] = sym__shell_single_quoted_fragment,
  [sym_json_string_array] = sym_json_string_array,
  [sym_json_string] = sym_json_string,
  [sym_double_quoted_string] = sym_double_quoted_string,
  [sym_single_quoted_string] = sym_single_quoted_string,
  [sym_unquoted_string] = sym_unquoted_string,
  [sym__spaced_label_value] = sym_unquoted_string,
  [sym__continued_spaced_label_value] = sym_unquoted_string,
  [sym__spaced_label_value_continuation] = sym__spaced_label_value_continuation,
  [sym__spaced_label_value_atom] = sym__spaced_label_value_atom,
  [sym__spaced_label_value_fragment] = sym__spaced_label_value_fragment,
  [sym__spaced_env_value] = sym_unquoted_string,
  [sym__continued_spaced_env_value] = sym_unquoted_string,
  [sym__spaced_env_value_continuation] = sym__spaced_env_value_continuation,
  [sym__spaced_env_value_atom] = sym__spaced_env_value_atom,
  [sym__spaced_env_value_fragment] = sym__spaced_env_value_fragment,
  [aux_sym_source_file_repeat1] = aux_sym_source_file_repeat1,
  [aux_sym_source_file_repeat2] = aux_sym_source_file_repeat2,
  [aux_sym_run_instruction_repeat1] = aux_sym_run_instruction_repeat1,
  [aux_sym_run_instruction_repeat2] = aux_sym_run_instruction_repeat2,
  [aux_sym_label_instruction_repeat1] = aux_sym_label_instruction_repeat1,
  [aux_sym_expose_instruction_repeat1] = aux_sym_expose_instruction_repeat1,
  [aux_sym_env_instruction_repeat1] = aux_sym_env_instruction_repeat1,
  [aux_sym_add_instruction_repeat1] = aux_sym_add_instruction_repeat1,
  [aux_sym_add_instruction_repeat2] = aux_sym_add_instruction_repeat2,
  [aux_sym_volume_instruction_repeat1] = aux_sym_volume_instruction_repeat1,
  [aux_sym__user_name_or_group_repeat1] = aux_sym__user_name_or_group_repeat1,
  [aux_sym_arg_instruction_repeat1] = aux_sym_arg_instruction_repeat1,
  [aux_sym__stopsignal_value_repeat1] = aux_sym__stopsignal_value_repeat1,
  [aux_sym__healthcheck_raw_command_repeat1] = aux_sym__healthcheck_raw_command_repeat1,
  [aux_sym__healthcheck_raw_fragment_repeat1] = aux_sym__healthcheck_raw_fragment_repeat1,
  [aux_sym_heredoc_block_repeat1] = aux_sym_heredoc_block_repeat1,
  [aux_sym_path_repeat1] = aux_sym_path_repeat1,
  [aux_sym_image_name_repeat1] = aux_sym_image_name_repeat1,
  [aux_sym_image_tag_repeat1] = aux_sym_image_tag_repeat1,
  [aux_sym_image_digest_repeat1] = aux_sym_image_digest_repeat1,
  [aux_sym_mount_param_repeat1] = aux_sym_mount_param_repeat1,
  [aux_sym_image_alias_repeat1] = aux_sym_image_alias_repeat1,
  [aux_sym__shell_double_quoted_fragment_repeat1] = aux_sym__shell_double_quoted_fragment_repeat1,
  [aux_sym__shell_single_quoted_fragment_repeat1] = aux_sym__shell_single_quoted_fragment_repeat1,
  [aux_sym_json_string_array_repeat1] = aux_sym_json_string_array_repeat1,
  [aux_sym_json_string_repeat1] = aux_sym_json_string_repeat1,
  [aux_sym_double_quoted_string_repeat1] = aux_sym_double_quoted_string_repeat1,
  [aux_sym_single_quoted_string_repeat1] = aux_sym_single_quoted_string_repeat1,
  [aux_sym_unquoted_string_repeat1] = aux_sym_unquoted_string_repeat1,
  [aux_sym__spaced_label_value_repeat1] = aux_sym__spaced_label_value_repeat1,
  [aux_sym__spaced_label_value_repeat2] = aux_sym__spaced_label_value_repeat2,
  [aux_sym__spaced_env_value_repeat1] = aux_sym__spaced_env_value_repeat1,
  [aux_sym__spaced_env_value_repeat2] = aux_sym__spaced_env_value_repeat2,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [aux_sym_from_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_from_instruction_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_run_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_cmd_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_label_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_expose_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_env_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_add_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_copy_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_entrypoint_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_volume_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_user_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__user_name_or_group_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__immediate_user_name_or_group_fragment_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_workdir_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_arg_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_arg_pair_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_onbuild_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_stopsignal_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__stopsignal_value_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__stopsignal_value_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_healthcheck_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NONE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__healthcheck_raw_fragment_atom_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__healthcheck_raw_fragment_atom_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__healthcheck_raw_fragment_atom_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__healthcheck_raw_fragment_atom_token4] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__healthcheck_raw_fragment_atom_token5] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__healthcheck_raw_fragment_atom_token6] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__healthcheck_raw_fragment_atom_token7] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__healthcheck_raw_fragment_atom_token8] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__healthcheck_raw_fragment_atom_token9] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__healthcheck_raw_fragment_atom_token10] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__healthcheck_raw_fragment_atom_token11] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__healthcheck_raw_fragment_atom_token12] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_shell_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_maintainer_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_maintainer_instruction_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cross_build_instruction_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LF] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_path_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_path_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_path_with_heredoc_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DOLLAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__expansion_body_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [sym_variable] = {
    .visible = true,
    .named = true,
  },
  [sym__env_trailing_single_quotes] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__spaced_env_pair_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__env_key_token1] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_expose_port_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SLASHtcp] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASHudp] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__label_key_token1] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_image_name_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_image_name_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_image_tag_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_image_digest_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_param_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_param_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_mount] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_mount_param_param_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_image_alias_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_image_alias_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__shell_fragment_atom_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__shell_double_quoted_fragment_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__shell_double_quoted_fragment_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__shell_single_quoted_fragment_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_json_string_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_BQUOTE] = {
    .visible = true,
    .named = false,
  },
  [sym_json_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_double_quoted_string_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_double_quoted_string_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_BSLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_single_quoted_string_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_unquoted_string_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_unquoted_string_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_unquoted_string_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__spaced_label_value_fragment_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_double_quoted_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [sym_single_quoted_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [sym__non_newline_whitespace] = {
    .visible = false,
    .named = true,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_line_continuation] = {
    .visible = true,
    .named = true,
  },
  [sym_required_line_continuation] = {
    .visible = true,
    .named = true,
  },
  [sym__newline] = {
    .visible = false,
    .named = true,
  },
  [sym__invalid_json_array_shell_command] = {
    .visible = true,
    .named = true,
  },
  [sym_heredoc_marker] = {
    .visible = true,
    .named = true,
  },
  [sym__heredoc_line] = {
    .visible = false,
    .named = true,
  },
  [sym_heredoc_end] = {
    .visible = true,
    .named = true,
  },
  [sym__heredoc_nl] = {
    .visible = false,
    .named = true,
  },
  [sym_error_sentinel] = {
    .visible = true,
    .named = true,
  },
  [sym_source_file] = {
    .visible = true,
    .named = true,
  },
  [sym__instruction] = {
    .visible = false,
    .named = true,
  },
  [sym_from_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_run_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_cmd_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_label_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_expose_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_env_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_add_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_copy_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_entrypoint_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_volume_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_user_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym__user_name_or_group] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__immediate_user_name_or_group] = {
    .visible = false,
    .named = false,
  },
  [sym__immediate_user_name_or_group_fragment] = {
    .visible = false,
    .named = true,
  },
  [sym_workdir_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_arg_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_arg_pair] = {
    .visible = true,
    .named = true,
  },
  [sym_onbuild_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_stopsignal_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym__stopsignal_value] = {
    .visible = false,
    .named = true,
  },
  [sym_healthcheck_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym__healthcheck_cmd_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym__healthcheck_raw_command] = {
    .visible = true,
    .named = true,
  },
  [sym__healthcheck_raw_fragment] = {
    .visible = false,
    .named = true,
  },
  [sym__healthcheck_raw_fragment_atom] = {
    .visible = false,
    .named = true,
  },
  [sym_shell_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_maintainer_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_cross_build_instruction] = {
    .visible = true,
    .named = true,
  },
  [sym_heredoc_block] = {
    .visible = true,
    .named = true,
  },
  [sym_path] = {
    .visible = true,
    .named = true,
  },
  [sym_path_with_heredoc] = {
    .visible = true,
    .named = true,
  },
  [sym_expansion] = {
    .visible = true,
    .named = true,
  },
  [sym__immediate_expansion] = {
    .visible = false,
    .named = true,
  },
  [sym__imm_expansion] = {
    .visible = true,
    .named = true,
  },
  [sym__expansion_body] = {
    .visible = false,
    .named = true,
  },
  [sym_env_pair] = {
    .visible = true,
    .named = true,
  },
  [sym__env_assignment_value] = {
    .visible = false,
    .named = true,
  },
  [sym__env_single_quoted_string_with_trailing_quote] = {
    .visible = false,
    .named = true,
  },
  [sym__spaced_env_pair] = {
    .visible = true,
    .named = true,
  },
  [sym__env_key] = {
    .visible = false,
    .named = true,
  },
  [sym_expose_port] = {
    .visible = true,
    .named = true,
  },
  [sym_label_pair] = {
    .visible = true,
    .named = true,
  },
  [sym__spaced_label_pair] = {
    .visible = true,
    .named = true,
  },
  [sym__label_key] = {
    .visible = false,
    .named = true,
  },
  [sym_image_spec] = {
    .visible = true,
    .named = true,
  },
  [sym_image_name] = {
    .visible = true,
    .named = true,
  },
  [sym_image_tag] = {
    .visible = true,
    .named = true,
  },
  [sym_image_digest] = {
    .visible = true,
    .named = true,
  },
  [sym_param] = {
    .visible = true,
    .named = true,
  },
  [sym_mount_param] = {
    .visible = true,
    .named = true,
  },
  [sym_mount_param_param] = {
    .visible = true,
    .named = true,
  },
  [sym_image_alias] = {
    .visible = true,
    .named = true,
  },
  [sym__json_or_shell_command] = {
    .visible = false,
    .named = true,
  },
  [sym_shell_command] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__shell_fragment] = {
    .visible = false,
    .named = false,
  },
  [sym__shell_fragment_atom] = {
    .visible = false,
    .named = true,
  },
  [sym__shell_double_quoted_fragment] = {
    .visible = false,
    .named = true,
  },
  [sym__shell_single_quoted_fragment] = {
    .visible = false,
    .named = true,
  },
  [sym_json_string_array] = {
    .visible = true,
    .named = true,
  },
  [sym_json_string] = {
    .visible = true,
    .named = true,
  },
  [sym_double_quoted_string] = {
    .visible = true,
    .named = true,
  },
  [sym_single_quoted_string] = {
    .visible = true,
    .named = true,
  },
  [sym_unquoted_string] = {
    .visible = true,
    .named = true,
  },
  [sym__spaced_label_value] = {
    .visible = true,
    .named = true,
  },
  [sym__continued_spaced_label_value] = {
    .visible = true,
    .named = true,
  },
  [sym__spaced_label_value_continuation] = {
    .visible = false,
    .named = true,
  },
  [sym__spaced_label_value_atom] = {
    .visible = false,
    .named = true,
  },
  [sym__spaced_label_value_fragment] = {
    .visible = false,
    .named = true,
  },
  [sym__spaced_env_value] = {
    .visible = true,
    .named = true,
  },
  [sym__continued_spaced_env_value] = {
    .visible = true,
    .named = true,
  },
  [sym__spaced_env_value_continuation] = {
    .visible = false,
    .named = true,
  },
  [sym__spaced_env_value_atom] = {
    .visible = false,
    .named = true,
  },
  [sym__spaced_env_value_fragment] = {
    .visible = false,
    .named = true,
  },
  [aux_sym_source_file_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_source_file_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_run_instruction_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_run_instruction_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_label_instruction_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expose_instruction_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_env_instruction_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_add_instruction_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_add_instruction_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_volume_instruction_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__user_name_or_group_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_arg_instruction_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__stopsignal_value_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__healthcheck_raw_command_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__healthcheck_raw_fragment_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_heredoc_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_path_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_image_name_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_image_tag_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_image_digest_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_mount_param_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_image_alias_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__shell_double_quoted_fragment_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__shell_single_quoted_fragment_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_json_string_array_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_json_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_double_quoted_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_single_quoted_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_unquoted_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__spaced_label_value_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__spaced_label_value_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__spaced_env_value_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__spaced_env_value_repeat2] = {
    .visible = false,
    .named = false,
  },
};

enum ts_field_identifiers {
  field_as = 1,
  field_default = 2,
  field_digest = 3,
  field_group = 4,
  field_key = 5,
  field_name = 6,
  field_tag = 7,
  field_user = 8,
  field_value = 9,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_as] = "as",
  [field_default] = "default",
  [field_digest] = "digest",
  [field_group] = "group",
  [field_key] = "key",
  [field_name] = "name",
  [field_tag] = "tag",
  [field_user] = "user",
  [field_value] = "value",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [1] = {.index = 0, .length = 1},
  [2] = {.index = 1, .length = 2},
  [3] = {.index = 3, .length = 2},
  [4] = {.index = 5, .length = 1},
  [5] = {.index = 6, .length = 1},
  [6] = {.index = 7, .length = 2},
  [7] = {.index = 9, .length = 2},
  [8] = {.index = 11, .length = 2},
  [9] = {.index = 13, .length = 2},
  [10] = {.index = 15, .length = 1},
  [11] = {.index = 16, .length = 3},
  [12] = {.index = 19, .length = 2},
  [13] = {.index = 21, .length = 2},
  [14] = {.index = 23, .length = 2},
  [15] = {.index = 25, .length = 2},
  [16] = {.index = 27, .length = 1},
  [17] = {.index = 28, .length = 2},
  [18] = {.index = 30, .length = 3},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_name, 0},
  [1] =
    {field_key, 1, .inherited = true},
    {field_value, 1, .inherited = true},
  [3] =
    {field_name, 1, .inherited = true},
    {field_value, 1, .inherited = true},
  [5] =
    {field_user, 1},
  [6] =
    {field_name, 1},
  [7] =
    {field_name, 0},
    {field_tag, 1},
  [9] =
    {field_digest, 1},
    {field_name, 0},
  [11] =
    {field_key, 0},
    {field_value, 1},
  [13] =
    {field_name, 0},
    {field_value, 1},
  [15] =
    {field_as, 3},
  [16] =
    {field_digest, 2},
    {field_name, 0},
    {field_tag, 1},
  [19] =
    {field_key, 0},
    {field_value, 2},
  [21] =
    {field_name, 0},
    {field_value, 2},
  [23] =
    {field_group, 3},
    {field_user, 1},
  [25] =
    {field_name, 1},
    {field_value, 3},
  [27] =
    {field_as, 4},
  [28] =
    {field_default, 2},
    {field_name, 0},
  [30] =
    {field_name, 1},
    {field_value, 3},
    {field_value, 4},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [14] = {
    [3] = sym_unquoted_string,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  aux_sym__immediate_user_name_or_group, 2,
    aux_sym__immediate_user_name_or_group,
    sym_unquoted_string,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 4,
  [5] = 5,
  [6] = 6,
  [7] = 7,
  [8] = 8,
  [9] = 9,
  [10] = 10,
  [11] = 11,
  [12] = 12,
  [13] = 13,
  [14] = 14,
  [15] = 15,
  [16] = 16,
  [17] = 17,
  [18] = 18,
  [19] = 14,
  [20] = 20,
  [21] = 21,
  [22] = 22,
  [23] = 23,
  [24] = 24,
  [25] = 17,
  [26] = 26,
  [27] = 20,
  [28] = 21,
  [29] = 29,
  [30] = 30,
  [31] = 31,
  [32] = 31,
  [33] = 33,
  [34] = 33,
  [35] = 35,
  [36] = 36,
  [37] = 37,
  [38] = 38,
  [39] = 39,
  [40] = 40,
  [41] = 41,
  [42] = 42,
  [43] = 43,
  [44] = 44,
  [45] = 45,
  [46] = 46,
  [47] = 47,
  [48] = 29,
  [49] = 49,
  [50] = 50,
  [51] = 51,
  [52] = 52,
  [53] = 53,
  [54] = 54,
  [55] = 44,
  [56] = 56,
  [57] = 49,
  [58] = 50,
  [59] = 43,
  [60] = 45,
  [61] = 30,
  [62] = 62,
  [63] = 63,
  [64] = 64,
  [65] = 65,
  [66] = 66,
  [67] = 67,
  [68] = 68,
  [69] = 69,
  [70] = 12,
  [71] = 71,
  [72] = 72,
  [73] = 73,
  [74] = 74,
  [75] = 72,
  [76] = 76,
  [77] = 77,
  [78] = 74,
  [79] = 79,
  [80] = 80,
  [81] = 65,
  [82] = 82,
  [83] = 83,
  [84] = 69,
  [85] = 65,
  [86] = 69,
  [87] = 87,
  [88] = 72,
  [89] = 74,
  [90] = 72,
  [91] = 74,
  [92] = 72,
  [93] = 74,
  [94] = 94,
  [95] = 95,
  [96] = 96,
  [97] = 97,
  [98] = 98,
  [99] = 99,
  [100] = 100,
  [101] = 101,
  [102] = 102,
  [103] = 103,
  [104] = 104,
  [105] = 98,
  [106] = 106,
  [107] = 107,
  [108] = 18,
  [109] = 109,
  [110] = 99,
  [111] = 99,
  [112] = 100,
  [113] = 100,
  [114] = 114,
  [115] = 115,
  [116] = 116,
  [117] = 117,
  [118] = 118,
  [119] = 119,
  [120] = 120,
  [121] = 121,
  [122] = 100,
  [123] = 123,
  [124] = 124,
  [125] = 125,
  [126] = 126,
  [127] = 127,
  [128] = 128,
  [129] = 129,
  [130] = 130,
  [131] = 99,
  [132] = 132,
  [133] = 133,
  [134] = 99,
  [135] = 135,
  [136] = 29,
  [137] = 100,
  [138] = 99,
  [139] = 139,
  [140] = 99,
  [141] = 100,
  [142] = 100,
  [143] = 143,
  [144] = 97,
  [145] = 145,
  [146] = 146,
  [147] = 30,
  [148] = 132,
  [149] = 99,
  [150] = 150,
  [151] = 100,
  [152] = 152,
  [153] = 100,
  [154] = 154,
  [155] = 99,
  [156] = 156,
  [157] = 157,
  [158] = 94,
  [159] = 159,
  [160] = 154,
  [161] = 143,
  [162] = 154,
  [163] = 163,
  [164] = 143,
  [165] = 154,
  [166] = 98,
  [167] = 143,
  [168] = 168,
  [169] = 154,
  [170] = 143,
  [171] = 171,
  [172] = 172,
  [173] = 173,
  [174] = 174,
  [175] = 175,
  [176] = 176,
  [177] = 177,
  [178] = 178,
  [179] = 179,
  [180] = 132,
  [181] = 181,
  [182] = 99,
  [183] = 100,
  [184] = 184,
  [185] = 132,
  [186] = 99,
  [187] = 100,
  [188] = 188,
  [189] = 189,
  [190] = 190,
  [191] = 99,
  [192] = 192,
  [193] = 193,
  [194] = 194,
  [195] = 195,
  [196] = 196,
  [197] = 100,
  [198] = 198,
  [199] = 199,
  [200] = 200,
  [201] = 201,
  [202] = 202,
  [203] = 203,
  [204] = 204,
  [205] = 205,
  [206] = 206,
  [207] = 177,
  [208] = 208,
  [209] = 206,
  [210] = 184,
  [211] = 211,
  [212] = 212,
  [213] = 213,
  [214] = 214,
  [215] = 211,
  [216] = 216,
  [217] = 217,
  [218] = 218,
  [219] = 219,
  [220] = 220,
  [221] = 221,
  [222] = 222,
  [223] = 223,
  [224] = 224,
  [225] = 225,
  [226] = 226,
  [227] = 227,
  [228] = 228,
  [229] = 229,
  [230] = 230,
  [231] = 231,
  [232] = 232,
  [233] = 233,
  [234] = 234,
  [235] = 235,
  [236] = 236,
  [237] = 132,
  [238] = 238,
  [239] = 239,
  [240] = 176,
  [241] = 241,
  [242] = 179,
  [243] = 243,
  [244] = 100,
  [245] = 174,
  [246] = 175,
  [247] = 247,
  [248] = 132,
  [249] = 99,
  [250] = 100,
  [251] = 251,
  [252] = 132,
  [253] = 100,
  [254] = 254,
  [255] = 255,
  [256] = 99,
  [257] = 257,
  [258] = 258,
  [259] = 259,
  [260] = 260,
  [261] = 261,
  [262] = 100,
  [263] = 263,
  [264] = 264,
  [265] = 265,
  [266] = 266,
  [267] = 205,
  [268] = 260,
  [269] = 269,
  [270] = 219,
  [271] = 271,
  [272] = 219,
  [273] = 205,
  [274] = 174,
  [275] = 175,
  [276] = 219,
  [277] = 175,
  [278] = 278,
  [279] = 279,
  [280] = 29,
  [281] = 281,
  [282] = 282,
  [283] = 205,
  [284] = 260,
  [285] = 285,
  [286] = 174,
  [287] = 287,
  [288] = 288,
  [289] = 289,
  [290] = 290,
  [291] = 291,
  [292] = 292,
  [293] = 269,
  [294] = 132,
  [295] = 289,
  [296] = 296,
  [297] = 99,
  [298] = 265,
  [299] = 299,
  [300] = 300,
  [301] = 260,
  [302] = 269,
  [303] = 260,
  [304] = 269,
  [305] = 300,
  [306] = 260,
  [307] = 269,
  [308] = 308,
  [309] = 269,
  [310] = 260,
  [311] = 269,
  [312] = 260,
  [313] = 269,
  [314] = 260,
  [315] = 260,
  [316] = 260,
  [317] = 260,
  [318] = 260,
  [319] = 260,
  [320] = 260,
  [321] = 321,
  [322] = 322,
  [323] = 323,
  [324] = 324,
  [325] = 325,
  [326] = 326,
  [327] = 327,
  [328] = 328,
  [329] = 329,
  [330] = 264,
  [331] = 331,
  [332] = 332,
  [333] = 333,
  [334] = 334,
  [335] = 335,
  [336] = 336,
  [337] = 337,
  [338] = 338,
  [339] = 30,
  [340] = 340,
  [341] = 341,
  [342] = 296,
  [343] = 343,
  [344] = 344,
  [345] = 345,
  [346] = 346,
  [347] = 347,
  [348] = 348,
  [349] = 349,
  [350] = 350,
  [351] = 351,
  [352] = 352,
  [353] = 353,
  [354] = 354,
  [355] = 355,
  [356] = 356,
  [357] = 287,
  [358] = 358,
  [359] = 359,
  [360] = 360,
  [361] = 361,
  [362] = 362,
  [363] = 363,
  [364] = 364,
  [365] = 365,
  [366] = 366,
  [367] = 205,
  [368] = 368,
  [369] = 368,
  [370] = 370,
  [371] = 371,
  [372] = 370,
  [373] = 174,
  [374] = 374,
  [375] = 375,
  [376] = 236,
  [377] = 368,
  [378] = 368,
  [379] = 370,
  [380] = 368,
  [381] = 381,
  [382] = 175,
  [383] = 368,
  [384] = 384,
  [385] = 385,
  [386] = 370,
  [387] = 387,
  [388] = 368,
  [389] = 368,
  [390] = 368,
  [391] = 371,
  [392] = 368,
  [393] = 393,
  [394] = 368,
  [395] = 219,
  [396] = 368,
  [397] = 371,
  [398] = 368,
  [399] = 385,
  [400] = 368,
  [401] = 368,
  [402] = 366,
  [403] = 366,
  [404] = 366,
  [405] = 366,
  [406] = 366,
  [407] = 366,
  [408] = 366,
  [409] = 366,
  [410] = 366,
  [411] = 366,
  [412] = 366,
  [413] = 366,
  [414] = 366,
  [415] = 366,
  [416] = 366,
  [417] = 368,
};

static const TSCharacterRange aux_sym__shell_fragment_atom_token1_character_set_1[] = {
  {0, 0x08}, {0x0e, 0x1f}, {'!', '!'}, {'$', '&'}, {'(', '+'}, {'.', '<'}, {'>', 'Z'}, {']', '_'},
  {'a', 0x10ffff},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(166);
      ADVANCE_MAP(
        '\n', 222,
        '"', 267,
        '$', 228,
        '\'', 271,
        ',', 257,
        '-', 200,
        ':', 180,
        '<', 241,
        '=', 189,
        '@', 246,
        '[', 272,
        '\\', 282,
        ']', 274,
        '_', 232,
        '`', 277,
        '{', 229,
        '}', 231,
      );
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(144);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(184);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(183);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(223);
      END_STATE();
    case 1:
      if (lookahead == ' ') ADVANCE(293);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(294);
      END_STATE();
    case 2:
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(139);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(265);
      END_STATE();
    case 3:
      if (lookahead == '"') ADVANCE(263);
      if (lookahead == '$') ADVANCE(228);
      if (lookahead == '\\') ADVANCE(283);
      if (lookahead == '`') ADVANCE(286);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') ADVANCE(279);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n') ADVANCE(280);
      END_STATE();
    case 4:
      if (lookahead == '"') ADVANCE(263);
      if (lookahead == '$') ADVANCE(226);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(5);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(254);
      if (lookahead == '+' ||
          lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z')) ADVANCE(192);
      END_STATE();
    case 5:
      if (lookahead == '"') ADVANCE(263);
      if (lookahead == '$') ADVANCE(226);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(5);
      if (lookahead == '+' ||
          lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z')) ADVANCE(192);
      END_STATE();
    case 6:
      if (lookahead == '"') ADVANCE(263);
      if (lookahead == '\\') ADVANCE(34);
      if (lookahead == '`') ADVANCE(277);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') ADVANCE(275);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n') ADVANCE(276);
      END_STATE();
    case 7:
      ADVANCE_MAP(
        '"', 264,
        '\'', 269,
        ',', 257,
        '-', 200,
        '<', 261,
        '=', 189,
        '[', 272,
        '\\', 212,
        '`', 212,
      );
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(12);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '"' &&
          lookahead != '#') ADVANCE(262);
      END_STATE();
    case 8:
      ADVANCE_MAP(
        '"', 264,
        '\'', 269,
        ',', 257,
        '-', 200,
        '<', 261,
        '=', 199,
        '[', 272,
        '\\', 212,
        '`', 212,
      );
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(12);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '"' &&
          lookahead != '#') ADVANCE(262);
      END_STATE();
    case 9:
      ADVANCE_MAP(
        '"', 264,
        '\'', 269,
        ',', 199,
        '-', 200,
        '<', 204,
        '=', 189,
        'C', 210,
        'c', 210,
        '\\', 212,
        '`', 212,
      );
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(11);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '"' &&
          lookahead != '#' &&
          lookahead != '[' &&
          lookahead != '\\') ADVANCE(205);
      END_STATE();
    case 10:
      ADVANCE_MAP(
        '"', 264,
        '\'', 269,
        ',', 199,
        '-', 200,
        '<', 261,
        '=', 189,
        '[', 272,
        '\\', 212,
        '`', 212,
      );
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(12);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '"' &&
          lookahead != '#') ADVANCE(262);
      END_STATE();
    case 11:
      ADVANCE_MAP(
        '"', 264,
        '\'', 269,
        '-', 200,
        '<', 204,
        ',', 199,
        '=', 199,
        'C', 210,
        'c', 210,
        '\\', 212,
        '`', 212,
      );
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(11);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '"' &&
          lookahead != '#' &&
          lookahead != '[' &&
          lookahead != '\\') ADVANCE(205);
      END_STATE();
    case 12:
      ADVANCE_MAP(
        '"', 264,
        '\'', 269,
        '-', 200,
        '<', 261,
        '[', 272,
        ',', 199,
        '=', 199,
        '\\', 212,
        '`', 212,
      );
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(12);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '"' &&
          lookahead != '#') ADVANCE(262);
      END_STATE();
    case 13:
      if (lookahead == '"') ADVANCE(197);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(140);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(13);
      END_STATE();
    case 14:
      if (lookahead == '$') ADVANCE(226);
      if (lookahead == '-') ADVANCE(27);
      if (lookahead == '<') ADVANCE(131);
      if (lookahead == '=') ADVANCE(189);
      if (lookahead == '[') ADVANCE(272);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(15);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(223);
      END_STATE();
    case 15:
      if (lookahead == '$') ADVANCE(226);
      if (lookahead == '-') ADVANCE(27);
      if (lookahead == '<') ADVANCE(131);
      if (lookahead == '[') ADVANCE(272);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(15);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(223);
      END_STATE();
    case 16:
      if (lookahead == '$') ADVANCE(226);
      if (lookahead == '-') ADVANCE(27);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(16);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ':' &&
          lookahead != '@') ADVANCE(241);
      END_STATE();
    case 17:
      if (lookahead == '$') ADVANCE(226);
      if (lookahead == '<') ADVANCE(131);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(17);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '-') ADVANCE(223);
      END_STATE();
    case 18:
      if (lookahead == '$') ADVANCE(226);
      if (lookahead == '<') ADVANCE(142);
      if (lookahead == '[') ADVANCE(272);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(18);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '-') ADVANCE(223);
      END_STATE();
    case 19:
      if (lookahead == '$') ADVANCE(226);
      if (lookahead == '<') ADVANCE(142);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(19);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '-') ADVANCE(223);
      END_STATE();
    case 20:
      if (lookahead == '$') ADVANCE(226);
      if (lookahead == '=') ADVANCE(189);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(23);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '-' &&
          lookahead != ':' &&
          lookahead != '@') ADVANCE(241);
      END_STATE();
    case 21:
      if (lookahead == '$') ADVANCE(226);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(21);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(259);
      END_STATE();
    case 22:
      if (lookahead == '$') ADVANCE(226);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(22);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(181);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(182);
      END_STATE();
    case 23:
      if (lookahead == '$') ADVANCE(226);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(23);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '-' &&
          lookahead != ':' &&
          lookahead != '@') ADVANCE(241);
      END_STATE();
    case 24:
      if (lookahead == '\'') ADVANCE(271);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(139);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(270);
      END_STATE();
    case 25:
      if (lookahead == '\'') ADVANCE(268);
      if (lookahead == '\\') ADVANCE(284);
      if (lookahead == '`') ADVANCE(287);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') ADVANCE(288);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n') ADVANCE(289);
      END_STATE();
    case 26:
      if (lookahead == '\'') ADVANCE(198);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(141);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(26);
      END_STATE();
    case 27:
      if (lookahead == '-') ADVANCE(249);
      END_STATE();
    case 28:
      if (lookahead == '_') ADVANCE(43);
      END_STATE();
    case 29:
      if (lookahead == 'c') ADVANCE(31);
      END_STATE();
    case 30:
      if (lookahead == 'd') ADVANCE(32);
      END_STATE();
    case 31:
      if (lookahead == 'p') ADVANCE(238);
      END_STATE();
    case 32:
      if (lookahead == 'p') ADVANCE(239);
      END_STATE();
    case 33:
      if (lookahead == 't') ADVANCE(29);
      if (lookahead == 'u') ADVANCE(30);
      END_STATE();
    case 34:
      ADVANCE_MAP(
        'u', 130,
        '"', 278,
        '/', 278,
        '\\', 278,
        'b', 278,
        'f', 278,
        'n', 278,
        'r', 278,
        't', 278,
      );
      END_STATE();
    case 35:
      if (lookahead == '{') ADVANCE(229);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(232);
      END_STATE();
    case 36:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(42);
      END_STATE();
    case 37:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(66);
      END_STATE();
    case 38:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(74);
      END_STATE();
    case 39:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(77);
      END_STATE();
    case 40:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(67);
      END_STATE();
    case 41:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(120);
      END_STATE();
    case 42:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(58);
      END_STATE();
    case 43:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(122);
      END_STATE();
    case 44:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(72);
      END_STATE();
    case 45:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(64);
      END_STATE();
    case 46:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(47);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(60);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(168);
      END_STATE();
    case 47:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(175);
      END_STATE();
    case 48:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(170);
      END_STATE();
    case 49:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(190);
      END_STATE();
    case 50:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(221);
      END_STATE();
    case 51:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(68);
      END_STATE();
    case 52:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(173);
      END_STATE();
    case 53:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(178);
      END_STATE();
    case 54:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(44);
      END_STATE();
    case 55:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(38);
      END_STATE();
    case 56:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(81);
      END_STATE();
    case 57:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(104);
      END_STATE();
    case 58:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(75);
      END_STATE();
    case 59:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(106);
      END_STATE();
    case 60:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(187);
      END_STATE();
    case 61:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(90);
      END_STATE();
    case 62:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(45);
      END_STATE();
    case 63:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(56);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(96);
      END_STATE();
    case 64:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(54);
      END_STATE();
    case 65:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(61);
      END_STATE();
    case 66:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(88);
      END_STATE();
    case 67:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(91);
      END_STATE();
    case 68:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(105);
      END_STATE();
    case 69:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(89);
      END_STATE();
    case 70:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(79);
      END_STATE();
    case 71:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(80);
      END_STATE();
    case 72:
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(194);
      END_STATE();
    case 73:
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(51);
      END_STATE();
    case 74:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(115);
      END_STATE();
    case 75:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(172);
      END_STATE();
    case 76:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(217);
      END_STATE();
    case 77:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(191);
      END_STATE();
    case 78:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(121);
      END_STATE();
    case 79:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(49);
      END_STATE();
    case 80:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(50);
      END_STATE();
    case 81:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(76);
      END_STATE();
    case 82:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(167);
      END_STATE();
    case 83:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(48);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(99);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(95);
      END_STATE();
    case 84:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(53);
      END_STATE();
    case 85:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(117);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(101);
      END_STATE();
    case 86:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(169);
      END_STATE();
    case 87:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(41);
      END_STATE();
    case 88:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(118);
      END_STATE();
    case 89:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(116);
      END_STATE();
    case 90:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(39);
      END_STATE();
    case 91:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(59);
      END_STATE();
    case 92:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(78);
      END_STATE();
    case 93:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(82);
      END_STATE();
    case 94:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(103);
      END_STATE();
    case 95:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(112);
      END_STATE();
    case 96:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(100);
      END_STATE();
    case 97:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(114);
      END_STATE();
    case 98:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(69);
      END_STATE();
    case 99:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(123);
      END_STATE();
    case 100:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(113);
      END_STATE();
    case 101:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(97);
      END_STATE();
    case 102:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(98);
      END_STATE();
    case 103:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(73);
      END_STATE();
    case 104:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(179);
      END_STATE();
    case 105:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(186);
      END_STATE();
    case 106:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(218);
      END_STATE();
    case 107:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(124);
      END_STATE();
    case 108:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(93);
      END_STATE();
    case 109:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(168);
      END_STATE();
    case 110:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(28);
      END_STATE();
    case 111:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(57);
      END_STATE();
    case 112:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(110);
      END_STATE();
    case 113:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(65);
      END_STATE();
    case 114:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(52);
      END_STATE();
    case 115:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(62);
      END_STATE();
    case 116:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(177);
      END_STATE();
    case 117:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(107);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(174);
      END_STATE();
    case 118:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(40);
      END_STATE();
    case 119:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(86);
      END_STATE();
    case 120:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(70);
      END_STATE();
    case 121:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(84);
      END_STATE();
    case 122:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(71);
      END_STATE();
    case 123:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(176);
      END_STATE();
    case 124:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(102);
      END_STATE();
    case 125:
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(125);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(188);
      END_STATE();
    case 126:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(237);
      END_STATE();
    case 127:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(278);
      END_STATE();
    case 128:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(127);
      END_STATE();
    case 129:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(128);
      END_STATE();
    case 130:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(129);
      END_STATE();
    case 131:
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '$' &&
          lookahead != '-' &&
          lookahead != '<') ADVANCE(225);
      END_STATE();
    case 132:
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != ',' &&
          lookahead != '=') ADVANCE(258);
      END_STATE();
    case 133:
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(255);
      END_STATE();
    case 134:
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != ' ') ADVANCE(205);
      END_STATE();
    case 135:
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != ' ') ADVANCE(206);
      END_STATE();
    case 136:
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != ' ') ADVANCE(207);
      END_STATE();
    case 137:
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != ' ') ADVANCE(208);
      END_STATE();
    case 138:
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != ' ') ADVANCE(262);
      END_STATE();
    case 139:
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(266);
      END_STATE();
    case 140:
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(13);
      END_STATE();
    case 141:
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(26);
      END_STATE();
    case 142:
      if (lookahead != 0 &&
          lookahead != '<') ADVANCE(213);
      END_STATE();
    case 143:
      if (lookahead != 0 &&
          lookahead != '}') ADVANCE(230);
      END_STATE();
    case 144:
      if (eof) ADVANCE(166);
      ADVANCE_MAP(
        '\n', 222,
        '"', 263,
        '$', 226,
        '\'', 268,
        ',', 273,
        '-', 200,
        '<', 241,
        '=', 199,
        '[', 272,
        '\\', 282,
        ']', 274,
        '`', 285,
        ':', 223,
        '@', 223,
      );
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(144);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(223);
      END_STATE();
    case 145:
      if (eof) ADVANCE(166);
      ADVANCE_MAP(
        '"', 263,
        '$', 226,
        ',', 273,
        '/', 33,
        ':', 180,
        '@', 246,
        '[', 272,
        ']', 274,
        '}', 231,
        'A', 46,
        'a', 46,
        'C', 83,
        'c', 83,
        'E', 85,
        'e', 85,
        'F', 108,
        'f', 108,
        'H', 55,
        'h', 55,
        'L', 36,
        'l', 36,
        'M', 37,
        'm', 37,
        'O', 87,
        'o', 87,
        'R', 119,
        'r', 119,
        'S', 63,
        's', 63,
        'U', 111,
        'u', 111,
        'V', 92,
        'v', 92,
        'W', 94,
        'w', 94,
      );
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(146);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(236);
      END_STATE();
    case 146:
      if (eof) ADVANCE(166);
      ADVANCE_MAP(
        '"', 263,
        '$', 226,
        ',', 273,
        '/', 33,
        '[', 272,
        ']', 274,
        'A', 46,
        'a', 46,
        'C', 83,
        'c', 83,
        'E', 85,
        'e', 85,
        'F', 108,
        'f', 108,
        'H', 55,
        'h', 55,
        'L', 36,
        'l', 36,
        'M', 37,
        'm', 37,
        'O', 87,
        'o', 87,
        'R', 119,
        'r', 119,
        'S', 63,
        's', 63,
        'U', 111,
        'u', 111,
        'V', 92,
        'v', 92,
        'W', 94,
        'w', 94,
      );
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(146);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(236);
      END_STATE();
    case 147:
      if (eof) ADVANCE(166);
      if (lookahead == '"') ADVANCE(263);
      if (lookahead == '$') ADVANCE(227);
      if (lookahead == '\'') ADVANCE(268);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(298);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(1);
      if (lookahead == '\f' ||
          lookahead == '\r') SKIP(151);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(292);
      END_STATE();
    case 148:
      if (eof) ADVANCE(166);
      if (lookahead == '"') ADVANCE(263);
      if (lookahead == '$') ADVANCE(227);
      if (lookahead == '\'') ADVANCE(268);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(1);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(150);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(290);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(292);
      END_STATE();
    case 149:
      if (eof) ADVANCE(166);
      if (lookahead == '"') ADVANCE(263);
      if (lookahead == '$') ADVANCE(227);
      if (lookahead == '\'') ADVANCE(268);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(1);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(152);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(291);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(292);
      END_STATE();
    case 150:
      if (eof) ADVANCE(166);
      if (lookahead == '"') ADVANCE(263);
      if (lookahead == '\'') ADVANCE(268);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(150);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(235);
      END_STATE();
    case 151:
      if (eof) ADVANCE(166);
      if (lookahead == '"') ADVANCE(263);
      if (lookahead == '\'') ADVANCE(268);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(151);
      END_STATE();
    case 152:
      if (eof) ADVANCE(166);
      if (lookahead == '"') ADVANCE(263);
      if (lookahead == '\'') ADVANCE(268);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(152);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(240);
      END_STATE();
    case 153:
      if (eof) ADVANCE(166);
      ADVANCE_MAP(
        '"', 264,
        '\'', 269,
        '-', 200,
        '<', 204,
        'N', 203,
        ',', 199,
        '=', 199,
        'C', 210,
        'c', 210,
        '\\', 212,
        '`', 212,
      );
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(153);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '"' &&
          lookahead != '#' &&
          lookahead != '[' &&
          lookahead != '\\') ADVANCE(205);
      END_STATE();
    case 154:
      if (eof) ADVANCE(166);
      ADVANCE_MAP(
        '"', 264,
        '\'', 269,
        '<', 261,
        '[', 272,
        '\\', 212,
        '`', 212,
        ',', 199,
        '-', 199,
        '=', 199,
      );
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(154);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '"' &&
          lookahead != '#') ADVANCE(262);
      END_STATE();
    case 155:
      if (eof) ADVANCE(166);
      if (lookahead == '$') ADVANCE(227);
      if (lookahead == ':') ADVANCE(180);
      if (lookahead == '=') ADVANCE(189);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(234);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(184);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 156:
      if (eof) ADVANCE(166);
      if (lookahead == '$') ADVANCE(227);
      if (lookahead == ':') ADVANCE(180);
      if (lookahead == '@') ADVANCE(246);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(242);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(164);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(243);
      END_STATE();
    case 157:
      if (eof) ADVANCE(166);
      if (lookahead == '$') ADVANCE(227);
      if (lookahead == '=') ADVANCE(189);
      if (lookahead == 'm') ADVANCE(251);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(298);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(254);
      if (lookahead == '+' ||
          lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z')) ADVANCE(193);
      END_STATE();
    case 158:
      if (eof) ADVANCE(166);
      if (lookahead == '$') ADVANCE(227);
      if (lookahead == '@') ADVANCE(246);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(244);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(164);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(245);
      END_STATE();
    case 159:
      if (eof) ADVANCE(166);
      if (lookahead == '$') ADVANCE(227);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(298);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(1);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(295);
      END_STATE();
    case 160:
      if (eof) ADVANCE(166);
      if (lookahead == '$') ADVANCE(227);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(298);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(224);
      END_STATE();
    case 161:
      if (eof) ADVANCE(166);
      if (lookahead == '$') ADVANCE(227);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(247);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(164);
      if (('0' <= lookahead && lookahead <= ':') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(248);
      END_STATE();
    case 162:
      if (eof) ADVANCE(166);
      if (lookahead == '$') ADVANCE(227);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(260);
      END_STATE();
    case 163:
      if (eof) ADVANCE(166);
      if (lookahead == '\'') ADVANCE(233);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(165);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(235);
      END_STATE();
    case 164:
      if (eof) ADVANCE(166);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(109);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(164);
      END_STATE();
    case 165:
      if (eof) ADVANCE(166);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(165);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(235);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(aux_sym_from_instruction_token1);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(aux_sym_from_instruction_token2);
      END_STATE();
    case 169:
      ACCEPT_TOKEN(aux_sym_run_instruction_token1);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(aux_sym_cmd_instruction_token1);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(aux_sym_cmd_instruction_token1);
      if (((!eof && set_contains(aux_sym__shell_fragment_atom_token1_character_set_1, 9, lookahead)) ||
          lookahead == '#' ||
          lookahead == ',' ||
          lookahead == '-' ||
          lookahead == '=' ||
          lookahead == '[') &&
          lookahead != '<') ADVANCE(208);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(aux_sym_label_instruction_token1);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(aux_sym_expose_instruction_token1);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(aux_sym_env_instruction_token1);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(aux_sym_add_instruction_token1);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(aux_sym_copy_instruction_token1);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(aux_sym_entrypoint_instruction_token1);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(aux_sym_volume_instruction_token1);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(aux_sym_user_instruction_token1);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(aux_sym__user_name_or_group_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(181);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(aux_sym__user_name_or_group_token1);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(182);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(aux_sym__immediate_user_name_or_group_fragment_token1);
      if (lookahead == '-') ADVANCE(185);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(183);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(aux_sym__immediate_user_name_or_group_fragment_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(184);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(aux_sym__immediate_user_name_or_group_fragment_token1);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(aux_sym_workdir_instruction_token1);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(aux_sym_arg_instruction_token1);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(aux_sym_arg_pair_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(188);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(aux_sym_onbuild_instruction_token1);
      END_STATE();
    case 191:
      ACCEPT_TOKEN(aux_sym_stopsignal_instruction_token1);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(aux_sym__stopsignal_value_token1);
      if (lookahead == '+' ||
          lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z')) ADVANCE(192);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(aux_sym__stopsignal_value_token2);
      if (lookahead == '+' ||
          lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z')) ADVANCE(193);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(aux_sym_healthcheck_instruction_token1);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_NONE);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(134);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\'' &&
          lookahead != '<') ADVANCE(205);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token1);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(196);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token2);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token3);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token4);
      END_STATE();
    case 200:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token4);
      if (lookahead == '-') ADVANCE(249);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token5);
      if (lookahead == 'E') ADVANCE(195);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(134);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\'' &&
          lookahead != '<') ADVANCE(205);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token5);
      if (lookahead == 'N') ADVANCE(201);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(134);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\'' &&
          lookahead != '<') ADVANCE(205);
      END_STATE();
    case 203:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token5);
      if (lookahead == 'O') ADVANCE(202);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(134);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\'' &&
          lookahead != '<') ADVANCE(205);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token5);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(215);
      if (lookahead == '\n' ||
          lookahead == '"' ||
          lookahead == '\'') ADVANCE(213);
      if (lookahead != 0 &&
          lookahead != '<') ADVANCE(205);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token5);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(134);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\'' &&
          lookahead != '<') ADVANCE(205);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token6);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(135);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\'' &&
          lookahead != '<') ADVANCE(206);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token7);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(136);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\'' &&
          lookahead != '<') ADVANCE(207);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token8);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(137);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\'' &&
          lookahead != '<') ADVANCE(208);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token9);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(171);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(136);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\'' &&
          lookahead != '<') ADVANCE(207);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token9);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(209);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(135);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\'' &&
          lookahead != '<') ADVANCE(206);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token10);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token11);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != ',' &&
          lookahead != '-' &&
          lookahead != '=') ADVANCE(211);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token12);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token12);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(138);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\'' &&
          lookahead != '<') ADVANCE(262);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token12);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != ' ') ADVANCE(205);
      END_STATE();
    case 216:
      ACCEPT_TOKEN(aux_sym__healthcheck_raw_fragment_atom_token12);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != ' ') ADVANCE(262);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(aux_sym_shell_instruction_token1);
      END_STATE();
    case 218:
      ACCEPT_TOKEN(aux_sym_maintainer_instruction_token1);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(aux_sym_maintainer_instruction_token2);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') ADVANCE(219);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n') ADVANCE(220);
      END_STATE();
    case 220:
      ACCEPT_TOKEN(aux_sym_maintainer_instruction_token2);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(220);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(aux_sym_cross_build_instruction_token1);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(221);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(anon_sym_LF);
      END_STATE();
    case 223:
      ACCEPT_TOKEN(aux_sym_path_token1);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(aux_sym_path_token2);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '$') ADVANCE(224);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(aux_sym_path_with_heredoc_token1);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(anon_sym_DOLLAR2);
      END_STATE();
    case 228:
      ACCEPT_TOKEN(anon_sym_DOLLAR2);
      if (lookahead == '(') ADVANCE(281);
      END_STATE();
    case 229:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(aux_sym__expansion_body_token1);
      if (lookahead != 0 &&
          lookahead != '}') ADVANCE(230);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 232:
      ACCEPT_TOKEN(sym_variable);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(232);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(sym__env_trailing_single_quotes);
      if (lookahead == '\'') ADVANCE(233);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(aux_sym__spaced_env_pair_token1);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(234);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(aux_sym__env_key_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(235);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(aux_sym_expose_port_token1);
      if (lookahead == '-') ADVANCE(126);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(236);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(aux_sym_expose_port_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(237);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(anon_sym_SLASHtcp);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(anon_sym_SLASHudp);
      END_STATE();
    case 240:
      ACCEPT_TOKEN(aux_sym__label_key_token1);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(240);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(aux_sym_image_name_token1);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(aux_sym_image_name_token2);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(243);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '$' &&
          lookahead != ':' &&
          lookahead != '@') ADVANCE(243);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(aux_sym_image_name_token2);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '$' &&
          lookahead != ':' &&
          lookahead != '@') ADVANCE(243);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(aux_sym_image_tag_token1);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(245);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '$' &&
          lookahead != '@') ADVANCE(245);
      END_STATE();
    case 245:
      ACCEPT_TOKEN(aux_sym_image_tag_token1);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '$' &&
          lookahead != '@') ADVANCE(245);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 247:
      ACCEPT_TOKEN(aux_sym_image_digest_token1);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(248);
      if (('0' <= lookahead && lookahead <= ':') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(248);
      END_STATE();
    case 248:
      ACCEPT_TOKEN(aux_sym_image_digest_token1);
      if (('0' <= lookahead && lookahead <= ':') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(248);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(aux_sym_param_token1);
      if (lookahead == 'n') ADVANCE(252);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(254);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(aux_sym_param_token1);
      if (lookahead == 'o') ADVANCE(253);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(254);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(aux_sym_param_token1);
      if (lookahead == 't') ADVANCE(256);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(254);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(aux_sym_param_token1);
      if (lookahead == 'u') ADVANCE(250);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(254);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(aux_sym_param_token1);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(254);
      END_STATE();
    case 255:
      ACCEPT_TOKEN(aux_sym_param_token2);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(255);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(anon_sym_mount);
      if (lookahead == '-' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(254);
      END_STATE();
    case 257:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(aux_sym_mount_param_param_token1);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != ',' &&
          lookahead != '=') ADVANCE(258);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(aux_sym_image_alias_token1);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(259);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(aux_sym_image_alias_token2);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(260);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(aux_sym__shell_fragment_atom_token1);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(216);
      if (lookahead == '\n' ||
          lookahead == '"' ||
          lookahead == '\'') ADVANCE(213);
      if (lookahead != 0 &&
          lookahead != '<') ADVANCE(214);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(aux_sym__shell_fragment_atom_token1);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(138);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\'' &&
          lookahead != '<') ADVANCE(262);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 264:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      if (lookahead == '"') ADVANCE(197);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(140);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(13);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(aux_sym__shell_double_quoted_fragment_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\' &&
          lookahead != '`') ADVANCE(265);
      END_STATE();
    case 266:
      ACCEPT_TOKEN(aux_sym__shell_double_quoted_fragment_token2);
      END_STATE();
    case 267:
      ACCEPT_TOKEN(anon_sym_DQUOTE2);
      END_STATE();
    case 268:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 269:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      if (lookahead == '\'') ADVANCE(198);
      if (lookahead == '\\' ||
          lookahead == '`') ADVANCE(141);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(26);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(aux_sym__shell_single_quoted_fragment_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\'' &&
          lookahead != '\\' &&
          lookahead != '`') ADVANCE(270);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(anon_sym_SQUOTE2);
      END_STATE();
    case 272:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(anon_sym_COMMA2);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(aux_sym_json_string_token1);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') ADVANCE(275);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\' &&
          lookahead != '`') ADVANCE(276);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(aux_sym_json_string_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\' &&
          lookahead != '`') ADVANCE(276);
      END_STATE();
    case 277:
      ACCEPT_TOKEN(anon_sym_BQUOTE);
      END_STATE();
    case 278:
      ACCEPT_TOKEN(sym_json_escape_sequence);
      END_STATE();
    case 279:
      ACCEPT_TOKEN(aux_sym_double_quoted_string_token1);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') ADVANCE(279);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '\\' &&
          lookahead != '`') ADVANCE(280);
      END_STATE();
    case 280:
      ACCEPT_TOKEN(aux_sym_double_quoted_string_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '\\' &&
          lookahead != '`') ADVANCE(280);
      END_STATE();
    case 281:
      ACCEPT_TOKEN(aux_sym_double_quoted_string_token2);
      END_STATE();
    case 282:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      END_STATE();
    case 283:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      if (lookahead == '"' ||
          lookahead == '\\' ||
          lookahead == '`') ADVANCE(296);
      END_STATE();
    case 284:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      if (lookahead == '\'' ||
          lookahead == '\\' ||
          lookahead == '`') ADVANCE(297);
      END_STATE();
    case 285:
      ACCEPT_TOKEN(anon_sym_BQUOTE2);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(anon_sym_BQUOTE2);
      if (lookahead == '"' ||
          lookahead == '\\' ||
          lookahead == '`') ADVANCE(296);
      END_STATE();
    case 287:
      ACCEPT_TOKEN(anon_sym_BQUOTE2);
      if (lookahead == '\'' ||
          lookahead == '\\' ||
          lookahead == '`') ADVANCE(297);
      END_STATE();
    case 288:
      ACCEPT_TOKEN(aux_sym_single_quoted_string_token1);
      if (lookahead == '\t' ||
          lookahead == '\f' ||
          lookahead == '\r' ||
          lookahead == ' ') ADVANCE(288);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\'' &&
          lookahead != '\\' &&
          lookahead != '`') ADVANCE(289);
      END_STATE();
    case 289:
      ACCEPT_TOKEN(aux_sym_single_quoted_string_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\'' &&
          lookahead != '\\' &&
          lookahead != '`') ADVANCE(289);
      END_STATE();
    case 290:
      ACCEPT_TOKEN(aux_sym_unquoted_string_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(290);
      if (((!eof && set_contains(aux_sym__shell_fragment_atom_token1_character_set_1, 9, lookahead)) ||
          lookahead == '#' ||
          lookahead == ',' ||
          lookahead == '-' ||
          lookahead == '=' ||
          lookahead == '[') &&
          lookahead != '$') ADVANCE(292);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(aux_sym_unquoted_string_token1);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(291);
      if (((!eof && set_contains(aux_sym__shell_fragment_atom_token1_character_set_1, 9, lookahead)) ||
          lookahead == '#' ||
          lookahead == ',' ||
          lookahead == '-' ||
          lookahead == '=' ||
          lookahead == '[') &&
          lookahead != '$') ADVANCE(292);
      END_STATE();
    case 292:
      ACCEPT_TOKEN(aux_sym_unquoted_string_token1);
      if (((!eof && set_contains(aux_sym__shell_fragment_atom_token1_character_set_1, 9, lookahead)) ||
          lookahead == '#' ||
          lookahead == ',' ||
          lookahead == '-' ||
          lookahead == '=' ||
          lookahead == '[') &&
          lookahead != '$') ADVANCE(292);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(aux_sym_unquoted_string_token2);
      END_STATE();
    case 294:
      ACCEPT_TOKEN(aux_sym_unquoted_string_token3);
      END_STATE();
    case 295:
      ACCEPT_TOKEN(aux_sym__spaced_label_value_fragment_token1);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '$' &&
          lookahead != '\\' &&
          lookahead != '`') ADVANCE(295);
      END_STATE();
    case 296:
      ACCEPT_TOKEN(sym_double_quoted_escape_sequence);
      END_STATE();
    case 297:
      ACCEPT_TOKEN(sym_single_quoted_escape_sequence);
      END_STATE();
    case 298:
      ACCEPT_TOKEN(sym__non_newline_whitespace);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(298);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 145, .external_lex_state = 2},
  [2] = {.lex_state = 145, .external_lex_state = 2},
  [3] = {.lex_state = 145, .external_lex_state = 2},
  [4] = {.lex_state = 145, .external_lex_state = 2},
  [5] = {.lex_state = 145, .external_lex_state = 2},
  [6] = {.lex_state = 145, .external_lex_state = 2},
  [7] = {.lex_state = 145, .external_lex_state = 3},
  [8] = {.lex_state = 153, .external_lex_state = 4},
  [9] = {.lex_state = 11, .external_lex_state = 5},
  [10] = {.lex_state = 12, .external_lex_state = 6},
  [11] = {.lex_state = 12, .external_lex_state = 6},
  [12] = {.lex_state = 145, .external_lex_state = 2},
  [13] = {.lex_state = 154, .external_lex_state = 6},
  [14] = {.lex_state = 154, .external_lex_state = 7},
  [15] = {.lex_state = 154, .external_lex_state = 6},
  [16] = {.lex_state = 154, .external_lex_state = 4},
  [17] = {.lex_state = 154, .external_lex_state = 7},
  [18] = {.lex_state = 11, .external_lex_state = 5},
  [19] = {.lex_state = 154, .external_lex_state = 8},
  [20] = {.lex_state = 154, .external_lex_state = 7},
  [21] = {.lex_state = 154, .external_lex_state = 7},
  [22] = {.lex_state = 154, .external_lex_state = 8},
  [23] = {.lex_state = 148, .external_lex_state = 2},
  [24] = {.lex_state = 154, .external_lex_state = 8},
  [25] = {.lex_state = 154, .external_lex_state = 8},
  [26] = {.lex_state = 154, .external_lex_state = 8},
  [27] = {.lex_state = 154, .external_lex_state = 8},
  [28] = {.lex_state = 154, .external_lex_state = 8},
  [29] = {.lex_state = 9, .external_lex_state = 5},
  [30] = {.lex_state = 11, .external_lex_state = 5},
  [31] = {.lex_state = 154, .external_lex_state = 4},
  [32] = {.lex_state = 154, .external_lex_state = 4},
  [33] = {.lex_state = 154, .external_lex_state = 4},
  [34] = {.lex_state = 154, .external_lex_state = 4},
  [35] = {.lex_state = 12, .external_lex_state = 6},
  [36] = {.lex_state = 147, .external_lex_state = 3},
  [37] = {.lex_state = 8, .external_lex_state = 6},
  [38] = {.lex_state = 8, .external_lex_state = 6},
  [39] = {.lex_state = 7, .external_lex_state = 6},
  [40] = {.lex_state = 8, .external_lex_state = 6},
  [41] = {.lex_state = 159, .external_lex_state = 9},
  [42] = {.lex_state = 147, .external_lex_state = 9},
  [43] = {.lex_state = 154, .external_lex_state = 7},
  [44] = {.lex_state = 154, .external_lex_state = 7},
  [45] = {.lex_state = 154, .external_lex_state = 7},
  [46] = {.lex_state = 8, .external_lex_state = 6},
  [47] = {.lex_state = 8, .external_lex_state = 6},
  [48] = {.lex_state = 10, .external_lex_state = 6},
  [49] = {.lex_state = 154, .external_lex_state = 7},
  [50] = {.lex_state = 154, .external_lex_state = 7},
  [51] = {.lex_state = 147, .external_lex_state = 9},
  [52] = {.lex_state = 154, .external_lex_state = 8},
  [53] = {.lex_state = 159, .external_lex_state = 9},
  [54] = {.lex_state = 159, .external_lex_state = 9},
  [55] = {.lex_state = 154, .external_lex_state = 8},
  [56] = {.lex_state = 147, .external_lex_state = 9},
  [57] = {.lex_state = 154, .external_lex_state = 8},
  [58] = {.lex_state = 154, .external_lex_state = 8},
  [59] = {.lex_state = 154, .external_lex_state = 8},
  [60] = {.lex_state = 154, .external_lex_state = 8},
  [61] = {.lex_state = 12, .external_lex_state = 6},
  [62] = {.lex_state = 148, .external_lex_state = 3},
  [63] = {.lex_state = 15, .external_lex_state = 5},
  [64] = {.lex_state = 15, .external_lex_state = 5},
  [65] = {.lex_state = 149, .external_lex_state = 2},
  [66] = {.lex_state = 148, .external_lex_state = 3},
  [67] = {.lex_state = 15, .external_lex_state = 5},
  [68] = {.lex_state = 15, .external_lex_state = 5},
  [69] = {.lex_state = 149, .external_lex_state = 2},
  [70] = {.lex_state = 154, .external_lex_state = 4},
  [71] = {.lex_state = 159, .external_lex_state = 3},
  [72] = {.lex_state = 3, .external_lex_state = 3},
  [73] = {.lex_state = 159, .external_lex_state = 3},
  [74] = {.lex_state = 3, .external_lex_state = 3},
  [75] = {.lex_state = 3, .external_lex_state = 3},
  [76] = {.lex_state = 156, .external_lex_state = 2},
  [77] = {.lex_state = 156, .external_lex_state = 2},
  [78] = {.lex_state = 3, .external_lex_state = 3},
  [79] = {.lex_state = 152, .external_lex_state = 2},
  [80] = {.lex_state = 3, .external_lex_state = 3},
  [81] = {.lex_state = 148, .external_lex_state = 2},
  [82] = {.lex_state = 152, .external_lex_state = 2},
  [83] = {.lex_state = 147, .external_lex_state = 3},
  [84] = {.lex_state = 148, .external_lex_state = 2},
  [85] = {.lex_state = 147, .external_lex_state = 2},
  [86] = {.lex_state = 147, .external_lex_state = 2},
  [87] = {.lex_state = 156, .external_lex_state = 2},
  [88] = {.lex_state = 3, .external_lex_state = 3},
  [89] = {.lex_state = 3, .external_lex_state = 3},
  [90] = {.lex_state = 3, .external_lex_state = 3},
  [91] = {.lex_state = 3, .external_lex_state = 3},
  [92] = {.lex_state = 3, .external_lex_state = 3},
  [93] = {.lex_state = 3, .external_lex_state = 3},
  [94] = {.lex_state = 160, .external_lex_state = 10},
  [95] = {.lex_state = 155, .external_lex_state = 2},
  [96] = {.lex_state = 152, .external_lex_state = 3},
  [97] = {.lex_state = 160, .external_lex_state = 10},
  [98] = {.lex_state = 160, .external_lex_state = 10},
  [99] = {.lex_state = 149, .external_lex_state = 2},
  [100] = {.lex_state = 149, .external_lex_state = 2},
  [101] = {.lex_state = 158, .external_lex_state = 2},
  [102] = {.lex_state = 158, .external_lex_state = 2},
  [103] = {.lex_state = 155, .external_lex_state = 2},
  [104] = {.lex_state = 155, .external_lex_state = 2},
  [105] = {.lex_state = 160, .external_lex_state = 2},
  [106] = {.lex_state = 155, .external_lex_state = 2},
  [107] = {.lex_state = 155, .external_lex_state = 2},
  [108] = {.lex_state = 15, .external_lex_state = 5},
  [109] = {.lex_state = 160, .external_lex_state = 2},
  [110] = {.lex_state = 159, .external_lex_state = 9},
  [111] = {.lex_state = 147, .external_lex_state = 9},
  [112] = {.lex_state = 159, .external_lex_state = 9},
  [113] = {.lex_state = 147, .external_lex_state = 9},
  [114] = {.lex_state = 161, .external_lex_state = 2},
  [115] = {.lex_state = 161, .external_lex_state = 2},
  [116] = {.lex_state = 160, .external_lex_state = 2},
  [117] = {.lex_state = 16, .external_lex_state = 3},
  [118] = {.lex_state = 145, .external_lex_state = 2},
  [119] = {.lex_state = 157, .external_lex_state = 2},
  [120] = {.lex_state = 17, .external_lex_state = 5},
  [121] = {.lex_state = 17, .external_lex_state = 5},
  [122] = {.lex_state = 156, .external_lex_state = 2},
  [123] = {.lex_state = 162, .external_lex_state = 2},
  [124] = {.lex_state = 17, .external_lex_state = 5},
  [125] = {.lex_state = 17, .external_lex_state = 5},
  [126] = {.lex_state = 17, .external_lex_state = 5},
  [127] = {.lex_state = 157, .external_lex_state = 2},
  [128] = {.lex_state = 162, .external_lex_state = 2},
  [129] = {.lex_state = 157, .external_lex_state = 2},
  [130] = {.lex_state = 18, .external_lex_state = 3},
  [131] = {.lex_state = 156, .external_lex_state = 2},
  [132] = {.lex_state = 156, .external_lex_state = 2},
  [133] = {.lex_state = 162, .external_lex_state = 2},
  [134] = {.lex_state = 3, .external_lex_state = 3},
  [135] = {.lex_state = 145, .external_lex_state = 2},
  [136] = {.lex_state = 14, .external_lex_state = 5},
  [137] = {.lex_state = 3, .external_lex_state = 3},
  [138] = {.lex_state = 148, .external_lex_state = 2},
  [139] = {.lex_state = 145, .external_lex_state = 2},
  [140] = {.lex_state = 147, .external_lex_state = 2},
  [141] = {.lex_state = 148, .external_lex_state = 2},
  [142] = {.lex_state = 147, .external_lex_state = 2},
  [143] = {.lex_state = 25, .external_lex_state = 3},
  [144] = {.lex_state = 160, .external_lex_state = 3},
  [145] = {.lex_state = 147, .external_lex_state = 10},
  [146] = {.lex_state = 145, .external_lex_state = 2},
  [147] = {.lex_state = 15, .external_lex_state = 5},
  [148] = {.lex_state = 160, .external_lex_state = 10},
  [149] = {.lex_state = 158, .external_lex_state = 2},
  [150] = {.lex_state = 4, .external_lex_state = 3},
  [151] = {.lex_state = 160, .external_lex_state = 10},
  [152] = {.lex_state = 25, .external_lex_state = 3},
  [153] = {.lex_state = 158, .external_lex_state = 2},
  [154] = {.lex_state = 25, .external_lex_state = 3},
  [155] = {.lex_state = 160, .external_lex_state = 10},
  [156] = {.lex_state = 155, .external_lex_state = 3},
  [157] = {.lex_state = 147, .external_lex_state = 10},
  [158] = {.lex_state = 160, .external_lex_state = 3},
  [159] = {.lex_state = 163, .external_lex_state = 2},
  [160] = {.lex_state = 25, .external_lex_state = 3},
  [161] = {.lex_state = 25, .external_lex_state = 3},
  [162] = {.lex_state = 25, .external_lex_state = 3},
  [163] = {.lex_state = 163, .external_lex_state = 2},
  [164] = {.lex_state = 25, .external_lex_state = 3},
  [165] = {.lex_state = 25, .external_lex_state = 3},
  [166] = {.lex_state = 160, .external_lex_state = 3},
  [167] = {.lex_state = 25, .external_lex_state = 3},
  [168] = {.lex_state = 147, .external_lex_state = 10},
  [169] = {.lex_state = 25, .external_lex_state = 3},
  [170] = {.lex_state = 25, .external_lex_state = 3},
  [171] = {.lex_state = 147, .external_lex_state = 10},
  [172] = {.lex_state = 155, .external_lex_state = 11},
  [173] = {.lex_state = 19, .external_lex_state = 3},
  [174] = {.lex_state = 152, .external_lex_state = 2},
  [175] = {.lex_state = 152, .external_lex_state = 2},
  [176] = {.lex_state = 0, .external_lex_state = 12},
  [177] = {.lex_state = 24, .external_lex_state = 11},
  [178] = {.lex_state = 0, .external_lex_state = 9},
  [179] = {.lex_state = 0, .external_lex_state = 12},
  [180] = {.lex_state = 160, .external_lex_state = 2},
  [181] = {.lex_state = 0, .external_lex_state = 10},
  [182] = {.lex_state = 160, .external_lex_state = 2},
  [183] = {.lex_state = 160, .external_lex_state = 2},
  [184] = {.lex_state = 2, .external_lex_state = 11},
  [185] = {.lex_state = 155, .external_lex_state = 2},
  [186] = {.lex_state = 155, .external_lex_state = 2},
  [187] = {.lex_state = 155, .external_lex_state = 2},
  [188] = {.lex_state = 19, .external_lex_state = 3},
  [189] = {.lex_state = 145, .external_lex_state = 3},
  [190] = {.lex_state = 0, .external_lex_state = 9},
  [191] = {.lex_state = 161, .external_lex_state = 2},
  [192] = {.lex_state = 6, .external_lex_state = 3},
  [193] = {.lex_state = 0, .external_lex_state = 10},
  [194] = {.lex_state = 0, .external_lex_state = 10},
  [195] = {.lex_state = 163, .external_lex_state = 3},
  [196] = {.lex_state = 0, .external_lex_state = 10},
  [197] = {.lex_state = 161, .external_lex_state = 2},
  [198] = {.lex_state = 16, .external_lex_state = 3},
  [199] = {.lex_state = 152, .external_lex_state = 2},
  [200] = {.lex_state = 6, .external_lex_state = 3},
  [201] = {.lex_state = 0, .external_lex_state = 9},
  [202] = {.lex_state = 0, .external_lex_state = 9},
  [203] = {.lex_state = 0, .external_lex_state = 9},
  [204] = {.lex_state = 0, .external_lex_state = 9},
  [205] = {.lex_state = 152, .external_lex_state = 2},
  [206] = {.lex_state = 2, .external_lex_state = 11},
  [207] = {.lex_state = 24, .external_lex_state = 11},
  [208] = {.lex_state = 0, .external_lex_state = 9},
  [209] = {.lex_state = 2, .external_lex_state = 11},
  [210] = {.lex_state = 2, .external_lex_state = 11},
  [211] = {.lex_state = 24, .external_lex_state = 11},
  [212] = {.lex_state = 158, .external_lex_state = 3},
  [213] = {.lex_state = 0, .external_lex_state = 10},
  [214] = {.lex_state = 0, .external_lex_state = 10},
  [215] = {.lex_state = 24, .external_lex_state = 11},
  [216] = {.lex_state = 0, .external_lex_state = 9},
  [217] = {.lex_state = 161, .external_lex_state = 3},
  [218] = {.lex_state = 145, .external_lex_state = 2},
  [219] = {.lex_state = 152, .external_lex_state = 2},
  [220] = {.lex_state = 2, .external_lex_state = 11},
  [221] = {.lex_state = 155, .external_lex_state = 11},
  [222] = {.lex_state = 0, .external_lex_state = 10},
  [223] = {.lex_state = 24, .external_lex_state = 11},
  [224] = {.lex_state = 6, .external_lex_state = 3},
  [225] = {.lex_state = 0, .external_lex_state = 10},
  [226] = {.lex_state = 0, .external_lex_state = 10},
  [227] = {.lex_state = 147, .external_lex_state = 2},
  [228] = {.lex_state = 17, .external_lex_state = 5},
  [229] = {.lex_state = 0, .external_lex_state = 2},
  [230] = {.lex_state = 0, .external_lex_state = 2},
  [231] = {.lex_state = 147, .external_lex_state = 2},
  [232] = {.lex_state = 157, .external_lex_state = 2},
  [233] = {.lex_state = 147, .external_lex_state = 2},
  [234] = {.lex_state = 21, .external_lex_state = 3},
  [235] = {.lex_state = 0, .external_lex_state = 9},
  [236] = {.lex_state = 147, .external_lex_state = 10},
  [237] = {.lex_state = 145, .external_lex_state = 2},
  [238] = {.lex_state = 0, .external_lex_state = 2},
  [239] = {.lex_state = 0, .external_lex_state = 2},
  [240] = {.lex_state = 0, .external_lex_state = 9},
  [241] = {.lex_state = 21, .external_lex_state = 3},
  [242] = {.lex_state = 0, .external_lex_state = 9},
  [243] = {.lex_state = 0, .external_lex_state = 2},
  [244] = {.lex_state = 145, .external_lex_state = 2},
  [245] = {.lex_state = 163, .external_lex_state = 2},
  [246] = {.lex_state = 163, .external_lex_state = 2},
  [247] = {.lex_state = 147, .external_lex_state = 2},
  [248] = {.lex_state = 157, .external_lex_state = 2},
  [249] = {.lex_state = 157, .external_lex_state = 2},
  [250] = {.lex_state = 157, .external_lex_state = 2},
  [251] = {.lex_state = 22, .external_lex_state = 3},
  [252] = {.lex_state = 162, .external_lex_state = 2},
  [253] = {.lex_state = 162, .external_lex_state = 2},
  [254] = {.lex_state = 163, .external_lex_state = 2},
  [255] = {.lex_state = 147, .external_lex_state = 2},
  [256] = {.lex_state = 162, .external_lex_state = 2},
  [257] = {.lex_state = 145, .external_lex_state = 2},
  [258] = {.lex_state = 155, .external_lex_state = 2},
  [259] = {.lex_state = 0, .external_lex_state = 9},
  [260] = {.lex_state = 35, .external_lex_state = 3},
  [261] = {.lex_state = 0, .external_lex_state = 10},
  [262] = {.lex_state = 160, .external_lex_state = 3},
  [263] = {.lex_state = 0, .external_lex_state = 13},
  [264] = {.lex_state = 0, .external_lex_state = 10},
  [265] = {.lex_state = 145, .external_lex_state = 3},
  [266] = {.lex_state = 0, .external_lex_state = 13},
  [267] = {.lex_state = 163, .external_lex_state = 2},
  [268] = {.lex_state = 35, .external_lex_state = 3},
  [269] = {.lex_state = 35, .external_lex_state = 3},
  [270] = {.lex_state = 163, .external_lex_state = 2},
  [271] = {.lex_state = 147, .external_lex_state = 2},
  [272] = {.lex_state = 155, .external_lex_state = 11},
  [273] = {.lex_state = 147, .external_lex_state = 2},
  [274] = {.lex_state = 147, .external_lex_state = 2},
  [275] = {.lex_state = 155, .external_lex_state = 11},
  [276] = {.lex_state = 147, .external_lex_state = 2},
  [277] = {.lex_state = 147, .external_lex_state = 2},
  [278] = {.lex_state = 0, .external_lex_state = 13},
  [279] = {.lex_state = 145, .external_lex_state = 2},
  [280] = {.lex_state = 20, .external_lex_state = 3},
  [281] = {.lex_state = 145, .external_lex_state = 2},
  [282] = {.lex_state = 145, .external_lex_state = 2},
  [283] = {.lex_state = 155, .external_lex_state = 11},
  [284] = {.lex_state = 35, .external_lex_state = 3},
  [285] = {.lex_state = 145, .external_lex_state = 2},
  [286] = {.lex_state = 155, .external_lex_state = 11},
  [287] = {.lex_state = 0, .external_lex_state = 10},
  [288] = {.lex_state = 0, .external_lex_state = 10},
  [289] = {.lex_state = 145, .external_lex_state = 3},
  [290] = {.lex_state = 163, .external_lex_state = 2},
  [291] = {.lex_state = 147, .external_lex_state = 2},
  [292] = {.lex_state = 147, .external_lex_state = 2},
  [293] = {.lex_state = 35, .external_lex_state = 3},
  [294] = {.lex_state = 160, .external_lex_state = 3},
  [295] = {.lex_state = 145, .external_lex_state = 3},
  [296] = {.lex_state = 0, .external_lex_state = 10},
  [297] = {.lex_state = 160, .external_lex_state = 3},
  [298] = {.lex_state = 145, .external_lex_state = 3},
  [299] = {.lex_state = 145, .external_lex_state = 2},
  [300] = {.lex_state = 145, .external_lex_state = 3},
  [301] = {.lex_state = 35, .external_lex_state = 3},
  [302] = {.lex_state = 35, .external_lex_state = 3},
  [303] = {.lex_state = 35, .external_lex_state = 3},
  [304] = {.lex_state = 35, .external_lex_state = 3},
  [305] = {.lex_state = 145, .external_lex_state = 3},
  [306] = {.lex_state = 35, .external_lex_state = 3},
  [307] = {.lex_state = 35, .external_lex_state = 3},
  [308] = {.lex_state = 145, .external_lex_state = 3},
  [309] = {.lex_state = 35, .external_lex_state = 3},
  [310] = {.lex_state = 35, .external_lex_state = 3},
  [311] = {.lex_state = 35, .external_lex_state = 3},
  [312] = {.lex_state = 35, .external_lex_state = 3},
  [313] = {.lex_state = 35, .external_lex_state = 3},
  [314] = {.lex_state = 35, .external_lex_state = 3},
  [315] = {.lex_state = 35, .external_lex_state = 3},
  [316] = {.lex_state = 35, .external_lex_state = 3},
  [317] = {.lex_state = 35, .external_lex_state = 3},
  [318] = {.lex_state = 35, .external_lex_state = 3},
  [319] = {.lex_state = 35, .external_lex_state = 3},
  [320] = {.lex_state = 35, .external_lex_state = 3},
  [321] = {.lex_state = 163, .external_lex_state = 2},
  [322] = {.lex_state = 157, .external_lex_state = 3},
  [323] = {.lex_state = 145, .external_lex_state = 3},
  [324] = {.lex_state = 0, .external_lex_state = 2},
  [325] = {.lex_state = 0, .external_lex_state = 2},
  [326] = {.lex_state = 125, .external_lex_state = 3},
  [327] = {.lex_state = 0, .external_lex_state = 2},
  [328] = {.lex_state = 0, .external_lex_state = 2},
  [329] = {.lex_state = 0, .external_lex_state = 2},
  [330] = {.lex_state = 0, .external_lex_state = 2},
  [331] = {.lex_state = 0, .external_lex_state = 2},
  [332] = {.lex_state = 0, .external_lex_state = 2},
  [333] = {.lex_state = 0, .external_lex_state = 2},
  [334] = {.lex_state = 0, .external_lex_state = 2},
  [335] = {.lex_state = 0, .external_lex_state = 2},
  [336] = {.lex_state = 0, .external_lex_state = 2},
  [337] = {.lex_state = 0, .external_lex_state = 2},
  [338] = {.lex_state = 145, .external_lex_state = 3},
  [339] = {.lex_state = 16, .external_lex_state = 3},
  [340] = {.lex_state = 0, .external_lex_state = 2},
  [341] = {.lex_state = 0, .external_lex_state = 2},
  [342] = {.lex_state = 0, .external_lex_state = 2},
  [343] = {.lex_state = 0, .external_lex_state = 2},
  [344] = {.lex_state = 0, .external_lex_state = 2},
  [345] = {.lex_state = 0, .external_lex_state = 2},
  [346] = {.lex_state = 0, .external_lex_state = 2},
  [347] = {.lex_state = 0, .external_lex_state = 13},
  [348] = {.lex_state = 0, .external_lex_state = 2},
  [349] = {.lex_state = 145, .external_lex_state = 3},
  [350] = {.lex_state = 145, .external_lex_state = 3},
  [351] = {.lex_state = 132, .external_lex_state = 3},
  [352] = {.lex_state = 0, .external_lex_state = 2},
  [353] = {.lex_state = 0, .external_lex_state = 2},
  [354] = {.lex_state = 0, .external_lex_state = 2},
  [355] = {.lex_state = 0, .external_lex_state = 2},
  [356] = {.lex_state = 0, .external_lex_state = 2},
  [357] = {.lex_state = 0, .external_lex_state = 2},
  [358] = {.lex_state = 145, .external_lex_state = 3},
  [359] = {.lex_state = 147, .external_lex_state = 3},
  [360] = {.lex_state = 0, .external_lex_state = 2},
  [361] = {.lex_state = 132, .external_lex_state = 3},
  [362] = {.lex_state = 0, .external_lex_state = 3},
  [363] = {.lex_state = 219, .external_lex_state = 3},
  [364] = {.lex_state = 0, .external_lex_state = 3},
  [365] = {.lex_state = 196, .external_lex_state = 3},
  [366] = {.lex_state = 143, .external_lex_state = 3},
  [367] = {.lex_state = 0, .external_lex_state = 3},
  [368] = {.lex_state = 145, .external_lex_state = 3},
  [369] = {.lex_state = 145, .external_lex_state = 3},
  [370] = {.lex_state = 133, .external_lex_state = 3},
  [371] = {.lex_state = 4, .external_lex_state = 3},
  [372] = {.lex_state = 133, .external_lex_state = 3},
  [373] = {.lex_state = 0, .external_lex_state = 3},
  [374] = {.lex_state = 0, .external_lex_state = 3},
  [375] = {.lex_state = 219, .external_lex_state = 3},
  [376] = {.lex_state = 147, .external_lex_state = 3},
  [377] = {.lex_state = 145, .external_lex_state = 3},
  [378] = {.lex_state = 145, .external_lex_state = 3},
  [379] = {.lex_state = 133, .external_lex_state = 3},
  [380] = {.lex_state = 145, .external_lex_state = 3},
  [381] = {.lex_state = 0, .external_lex_state = 3},
  [382] = {.lex_state = 0, .external_lex_state = 3},
  [383] = {.lex_state = 145, .external_lex_state = 3},
  [384] = {.lex_state = 147, .external_lex_state = 3},
  [385] = {.lex_state = 196, .external_lex_state = 3},
  [386] = {.lex_state = 133, .external_lex_state = 3},
  [387] = {.lex_state = 132, .external_lex_state = 3},
  [388] = {.lex_state = 145, .external_lex_state = 3},
  [389] = {.lex_state = 145, .external_lex_state = 3},
  [390] = {.lex_state = 145, .external_lex_state = 3},
  [391] = {.lex_state = 4, .external_lex_state = 3},
  [392] = {.lex_state = 145, .external_lex_state = 3},
  [393] = {.lex_state = 0, .external_lex_state = 3},
  [394] = {.lex_state = 145, .external_lex_state = 3},
  [395] = {.lex_state = 0, .external_lex_state = 3},
  [396] = {.lex_state = 145, .external_lex_state = 3},
  [397] = {.lex_state = 4, .external_lex_state = 3},
  [398] = {.lex_state = 145, .external_lex_state = 3},
  [399] = {.lex_state = 196, .external_lex_state = 3},
  [400] = {.lex_state = 145, .external_lex_state = 3},
  [401] = {.lex_state = 145, .external_lex_state = 3},
  [402] = {.lex_state = 143, .external_lex_state = 3},
  [403] = {.lex_state = 143, .external_lex_state = 3},
  [404] = {.lex_state = 143, .external_lex_state = 3},
  [405] = {.lex_state = 143, .external_lex_state = 3},
  [406] = {.lex_state = 143, .external_lex_state = 3},
  [407] = {.lex_state = 143, .external_lex_state = 3},
  [408] = {.lex_state = 143, .external_lex_state = 3},
  [409] = {.lex_state = 143, .external_lex_state = 3},
  [410] = {.lex_state = 143, .external_lex_state = 3},
  [411] = {.lex_state = 143, .external_lex_state = 3},
  [412] = {.lex_state = 143, .external_lex_state = 3},
  [413] = {.lex_state = 143, .external_lex_state = 3},
  [414] = {.lex_state = 143, .external_lex_state = 3},
  [415] = {.lex_state = 143, .external_lex_state = 3},
  [416] = {.lex_state = 143, .external_lex_state = 3},
  [417] = {.lex_state = 145, .external_lex_state = 3},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [aux_sym__immediate_user_name_or_group_fragment_token1] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [aux_sym__healthcheck_raw_fragment_atom_token4] = ACTIONS(1),
    [aux_sym__healthcheck_raw_fragment_atom_token11] = ACTIONS(1),
    [anon_sym_LF] = ACTIONS(1),
    [aux_sym_path_token1] = ACTIONS(1),
    [anon_sym_DOLLAR] = ACTIONS(1),
    [anon_sym_DOLLAR2] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [sym_variable] = ACTIONS(1),
    [aux_sym_image_name_token1] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [anon_sym_DASH_DASH] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [anon_sym_DQUOTE2] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_SQUOTE2] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_COMMA2] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_BQUOTE] = ACTIONS(1),
    [aux_sym_double_quoted_string_token2] = ACTIONS(1),
    [anon_sym_BSLASH] = ACTIONS(1),
    [anon_sym_BQUOTE2] = ACTIONS(1),
    [sym_comment] = ACTIONS(3),
    [sym_line_continuation] = ACTIONS(3),
    [sym_required_line_continuation] = ACTIONS(1),
    [sym__newline] = ACTIONS(1),
    [sym__invalid_json_array_shell_command] = ACTIONS(1),
    [sym_heredoc_marker] = ACTIONS(1),
    [sym__heredoc_line] = ACTIONS(1),
    [sym_heredoc_end] = ACTIONS(1),
    [sym__heredoc_nl] = ACTIONS(1),
    [sym_error_sentinel] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_source_file] = STATE(364),
    [sym__instruction] = STATE(243),
    [sym_from_instruction] = STATE(243),
    [sym_run_instruction] = STATE(243),
    [sym_cmd_instruction] = STATE(243),
    [sym_label_instruction] = STATE(243),
    [sym_expose_instruction] = STATE(243),
    [sym_env_instruction] = STATE(243),
    [sym_add_instruction] = STATE(243),
    [sym_copy_instruction] = STATE(243),
    [sym_entrypoint_instruction] = STATE(243),
    [sym_volume_instruction] = STATE(243),
    [sym_user_instruction] = STATE(243),
    [sym_workdir_instruction] = STATE(243),
    [sym_arg_instruction] = STATE(243),
    [sym_onbuild_instruction] = STATE(243),
    [sym_stopsignal_instruction] = STATE(243),
    [sym_healthcheck_instruction] = STATE(243),
    [sym_shell_instruction] = STATE(243),
    [sym_maintainer_instruction] = STATE(243),
    [sym_cross_build_instruction] = STATE(243),
    [aux_sym_source_file_repeat1] = STATE(3),
    [ts_builtin_sym_end] = ACTIONS(5),
    [aux_sym_from_instruction_token1] = ACTIONS(7),
    [aux_sym_run_instruction_token1] = ACTIONS(9),
    [aux_sym_cmd_instruction_token1] = ACTIONS(11),
    [aux_sym_label_instruction_token1] = ACTIONS(13),
    [aux_sym_expose_instruction_token1] = ACTIONS(15),
    [aux_sym_env_instruction_token1] = ACTIONS(17),
    [aux_sym_add_instruction_token1] = ACTIONS(19),
    [aux_sym_copy_instruction_token1] = ACTIONS(21),
    [aux_sym_entrypoint_instruction_token1] = ACTIONS(23),
    [aux_sym_volume_instruction_token1] = ACTIONS(25),
    [aux_sym_user_instruction_token1] = ACTIONS(27),
    [aux_sym_workdir_instruction_token1] = ACTIONS(29),
    [aux_sym_arg_instruction_token1] = ACTIONS(31),
    [aux_sym_onbuild_instruction_token1] = ACTIONS(33),
    [aux_sym_stopsignal_instruction_token1] = ACTIONS(35),
    [aux_sym_healthcheck_instruction_token1] = ACTIONS(37),
    [aux_sym_shell_instruction_token1] = ACTIONS(39),
    [aux_sym_maintainer_instruction_token1] = ACTIONS(41),
    [aux_sym_cross_build_instruction_token1] = ACTIONS(43),
    [sym_comment] = ACTIONS(3),
    [sym_line_continuation] = ACTIONS(3),
    [sym__newline] = ACTIONS(45),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 24,
    ACTIONS(7), 1,
      aux_sym_from_instruction_token1,
    ACTIONS(9), 1,
      aux_sym_run_instruction_token1,
    ACTIONS(11), 1,
      aux_sym_cmd_instruction_token1,
    ACTIONS(13), 1,
      aux_sym_label_instruction_token1,
    ACTIONS(15), 1,
      aux_sym_expose_instruction_token1,
    ACTIONS(17), 1,
      aux_sym_env_instruction_token1,
    ACTIONS(19), 1,
      aux_sym_add_instruction_token1,
    ACTIONS(21), 1,
      aux_sym_copy_instruction_token1,
    ACTIONS(23), 1,
      aux_sym_entrypoint_instruction_token1,
    ACTIONS(25), 1,
      aux_sym_volume_instruction_token1,
    ACTIONS(27), 1,
      aux_sym_user_instruction_token1,
    ACTIONS(29), 1,
      aux_sym_workdir_instruction_token1,
    ACTIONS(31), 1,
      aux_sym_arg_instruction_token1,
    ACTIONS(33), 1,
      aux_sym_onbuild_instruction_token1,
    ACTIONS(35), 1,
      aux_sym_stopsignal_instruction_token1,
    ACTIONS(37), 1,
      aux_sym_healthcheck_instruction_token1,
    ACTIONS(39), 1,
      aux_sym_shell_instruction_token1,
    ACTIONS(41), 1,
      aux_sym_maintainer_instruction_token1,
    ACTIONS(43), 1,
      aux_sym_cross_build_instruction_token1,
    ACTIONS(47), 1,
      ts_builtin_sym_end,
    ACTIONS(49), 1,
      sym__newline,
    STATE(12), 1,
      aux_sym_source_file_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(354), 20,
      sym__instruction,
      sym_from_instruction,
      sym_run_instruction,
      sym_cmd_instruction,
      sym_label_instruction,
      sym_expose_instruction,
      sym_env_instruction,
      sym_add_instruction,
      sym_copy_instruction,
      sym_entrypoint_instruction,
      sym_volume_instruction,
      sym_user_instruction,
      sym_workdir_instruction,
      sym_arg_instruction,
      sym_onbuild_instruction,
      sym_stopsignal_instruction,
      sym_healthcheck_instruction,
      sym_shell_instruction,
      sym_maintainer_instruction,
      sym_cross_build_instruction,
  [93] = 24,
    ACTIONS(7), 1,
      aux_sym_from_instruction_token1,
    ACTIONS(9), 1,
      aux_sym_run_instruction_token1,
    ACTIONS(11), 1,
      aux_sym_cmd_instruction_token1,
    ACTIONS(13), 1,
      aux_sym_label_instruction_token1,
    ACTIONS(15), 1,
      aux_sym_expose_instruction_token1,
    ACTIONS(17), 1,
      aux_sym_env_instruction_token1,
    ACTIONS(19), 1,
      aux_sym_add_instruction_token1,
    ACTIONS(21), 1,
      aux_sym_copy_instruction_token1,
    ACTIONS(23), 1,
      aux_sym_entrypoint_instruction_token1,
    ACTIONS(25), 1,
      aux_sym_volume_instruction_token1,
    ACTIONS(27), 1,
      aux_sym_user_instruction_token1,
    ACTIONS(29), 1,
      aux_sym_workdir_instruction_token1,
    ACTIONS(31), 1,
      aux_sym_arg_instruction_token1,
    ACTIONS(33), 1,
      aux_sym_onbuild_instruction_token1,
    ACTIONS(35), 1,
      aux_sym_stopsignal_instruction_token1,
    ACTIONS(37), 1,
      aux_sym_healthcheck_instruction_token1,
    ACTIONS(39), 1,
      aux_sym_shell_instruction_token1,
    ACTIONS(41), 1,
      aux_sym_maintainer_instruction_token1,
    ACTIONS(43), 1,
      aux_sym_cross_build_instruction_token1,
    ACTIONS(49), 1,
      sym__newline,
    ACTIONS(51), 1,
      ts_builtin_sym_end,
    STATE(12), 1,
      aux_sym_source_file_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(230), 20,
      sym__instruction,
      sym_from_instruction,
      sym_run_instruction,
      sym_cmd_instruction,
      sym_label_instruction,
      sym_expose_instruction,
      sym_env_instruction,
      sym_add_instruction,
      sym_copy_instruction,
      sym_entrypoint_instruction,
      sym_volume_instruction,
      sym_user_instruction,
      sym_workdir_instruction,
      sym_arg_instruction,
      sym_onbuild_instruction,
      sym_stopsignal_instruction,
      sym_healthcheck_instruction,
      sym_shell_instruction,
      sym_maintainer_instruction,
      sym_cross_build_instruction,
  [186] = 24,
    ACTIONS(7), 1,
      aux_sym_from_instruction_token1,
    ACTIONS(9), 1,
      aux_sym_run_instruction_token1,
    ACTIONS(11), 1,
      aux_sym_cmd_instruction_token1,
    ACTIONS(13), 1,
      aux_sym_label_instruction_token1,
    ACTIONS(15), 1,
      aux_sym_expose_instruction_token1,
    ACTIONS(17), 1,
      aux_sym_env_instruction_token1,
    ACTIONS(19), 1,
      aux_sym_add_instruction_token1,
    ACTIONS(21), 1,
      aux_sym_copy_instruction_token1,
    ACTIONS(23), 1,
      aux_sym_entrypoint_instruction_token1,
    ACTIONS(25), 1,
      aux_sym_volume_instruction_token1,
    ACTIONS(27), 1,
      aux_sym_user_instruction_token1,
    ACTIONS(29), 1,
      aux_sym_workdir_instruction_token1,
    ACTIONS(31), 1,
      aux_sym_arg_instruction_token1,
    ACTIONS(33), 1,
      aux_sym_onbuild_instruction_token1,
    ACTIONS(35), 1,
      aux_sym_stopsignal_instruction_token1,
    ACTIONS(37), 1,
      aux_sym_healthcheck_instruction_token1,
    ACTIONS(39), 1,
      aux_sym_shell_instruction_token1,
    ACTIONS(41), 1,
      aux_sym_maintainer_instruction_token1,
    ACTIONS(43), 1,
      aux_sym_cross_build_instruction_token1,
    ACTIONS(49), 1,
      sym__newline,
    ACTIONS(53), 1,
      ts_builtin_sym_end,
    STATE(12), 1,
      aux_sym_source_file_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(354), 20,
      sym__instruction,
      sym_from_instruction,
      sym_run_instruction,
      sym_cmd_instruction,
      sym_label_instruction,
      sym_expose_instruction,
      sym_env_instruction,
      sym_add_instruction,
      sym_copy_instruction,
      sym_entrypoint_instruction,
      sym_volume_instruction,
      sym_user_instruction,
      sym_workdir_instruction,
      sym_arg_instruction,
      sym_onbuild_instruction,
      sym_stopsignal_instruction,
      sym_healthcheck_instruction,
      sym_shell_instruction,
      sym_maintainer_instruction,
      sym_cross_build_instruction,
  [279] = 24,
    ACTIONS(7), 1,
      aux_sym_from_instruction_token1,
    ACTIONS(9), 1,
      aux_sym_run_instruction_token1,
    ACTIONS(11), 1,
      aux_sym_cmd_instruction_token1,
    ACTIONS(13), 1,
      aux_sym_label_instruction_token1,
    ACTIONS(15), 1,
      aux_sym_expose_instruction_token1,
    ACTIONS(17), 1,
      aux_sym_env_instruction_token1,
    ACTIONS(19), 1,
      aux_sym_add_instruction_token1,
    ACTIONS(21), 1,
      aux_sym_copy_instruction_token1,
    ACTIONS(23), 1,
      aux_sym_entrypoint_instruction_token1,
    ACTIONS(25), 1,
      aux_sym_volume_instruction_token1,
    ACTIONS(27), 1,
      aux_sym_user_instruction_token1,
    ACTIONS(29), 1,
      aux_sym_workdir_instruction_token1,
    ACTIONS(31), 1,
      aux_sym_arg_instruction_token1,
    ACTIONS(33), 1,
      aux_sym_onbuild_instruction_token1,
    ACTIONS(35), 1,
      aux_sym_stopsignal_instruction_token1,
    ACTIONS(37), 1,
      aux_sym_healthcheck_instruction_token1,
    ACTIONS(39), 1,
      aux_sym_shell_instruction_token1,
    ACTIONS(41), 1,
      aux_sym_maintainer_instruction_token1,
    ACTIONS(43), 1,
      aux_sym_cross_build_instruction_token1,
    ACTIONS(49), 1,
      sym__newline,
    ACTIONS(55), 1,
      ts_builtin_sym_end,
    STATE(12), 1,
      aux_sym_source_file_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(354), 20,
      sym__instruction,
      sym_from_instruction,
      sym_run_instruction,
      sym_cmd_instruction,
      sym_label_instruction,
      sym_expose_instruction,
      sym_env_instruction,
      sym_add_instruction,
      sym_copy_instruction,
      sym_entrypoint_instruction,
      sym_volume_instruction,
      sym_user_instruction,
      sym_workdir_instruction,
      sym_arg_instruction,
      sym_onbuild_instruction,
      sym_stopsignal_instruction,
      sym_healthcheck_instruction,
      sym_shell_instruction,
      sym_maintainer_instruction,
      sym_cross_build_instruction,
  [372] = 23,
    ACTIONS(7), 1,
      aux_sym_from_instruction_token1,
    ACTIONS(9), 1,
      aux_sym_run_instruction_token1,
    ACTIONS(11), 1,
      aux_sym_cmd_instruction_token1,
    ACTIONS(13), 1,
      aux_sym_label_instruction_token1,
    ACTIONS(15), 1,
      aux_sym_expose_instruction_token1,
    ACTIONS(17), 1,
      aux_sym_env_instruction_token1,
    ACTIONS(19), 1,
      aux_sym_add_instruction_token1,
    ACTIONS(21), 1,
      aux_sym_copy_instruction_token1,
    ACTIONS(23), 1,
      aux_sym_entrypoint_instruction_token1,
    ACTIONS(25), 1,
      aux_sym_volume_instruction_token1,
    ACTIONS(27), 1,
      aux_sym_user_instruction_token1,
    ACTIONS(29), 1,
      aux_sym_workdir_instruction_token1,
    ACTIONS(31), 1,
      aux_sym_arg_instruction_token1,
    ACTIONS(33), 1,
      aux_sym_onbuild_instruction_token1,
    ACTIONS(35), 1,
      aux_sym_stopsignal_instruction_token1,
    ACTIONS(37), 1,
      aux_sym_healthcheck_instruction_token1,
    ACTIONS(39), 1,
      aux_sym_shell_instruction_token1,
    ACTIONS(41), 1,
      aux_sym_maintainer_instruction_token1,
    ACTIONS(43), 1,
      aux_sym_cross_build_instruction_token1,
    ACTIONS(49), 1,
      sym__newline,
    STATE(12), 1,
      aux_sym_source_file_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(354), 20,
      sym__instruction,
      sym_from_instruction,
      sym_run_instruction,
      sym_cmd_instruction,
      sym_label_instruction,
      sym_expose_instruction,
      sym_env_instruction,
      sym_add_instruction,
      sym_copy_instruction,
      sym_entrypoint_instruction,
      sym_volume_instruction,
      sym_user_instruction,
      sym_workdir_instruction,
      sym_arg_instruction,
      sym_onbuild_instruction,
      sym_stopsignal_instruction,
      sym_healthcheck_instruction,
      sym_shell_instruction,
      sym_maintainer_instruction,
      sym_cross_build_instruction,
  [462] = 21,
    ACTIONS(7), 1,
      aux_sym_from_instruction_token1,
    ACTIONS(9), 1,
      aux_sym_run_instruction_token1,
    ACTIONS(11), 1,
      aux_sym_cmd_instruction_token1,
    ACTIONS(13), 1,
      aux_sym_label_instruction_token1,
    ACTIONS(15), 1,
      aux_sym_expose_instruction_token1,
    ACTIONS(17), 1,
      aux_sym_env_instruction_token1,
    ACTIONS(19), 1,
      aux_sym_add_instruction_token1,
    ACTIONS(21), 1,
      aux_sym_copy_instruction_token1,
    ACTIONS(23), 1,
      aux_sym_entrypoint_instruction_token1,
    ACTIONS(25), 1,
      aux_sym_volume_instruction_token1,
    ACTIONS(27), 1,
      aux_sym_user_instruction_token1,
    ACTIONS(29), 1,
      aux_sym_workdir_instruction_token1,
    ACTIONS(31), 1,
      aux_sym_arg_instruction_token1,
    ACTIONS(33), 1,
      aux_sym_onbuild_instruction_token1,
    ACTIONS(35), 1,
      aux_sym_stopsignal_instruction_token1,
    ACTIONS(37), 1,
      aux_sym_healthcheck_instruction_token1,
    ACTIONS(39), 1,
      aux_sym_shell_instruction_token1,
    ACTIONS(41), 1,
      aux_sym_maintainer_instruction_token1,
    ACTIONS(43), 1,
      aux_sym_cross_build_instruction_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(327), 20,
      sym__instruction,
      sym_from_instruction,
      sym_run_instruction,
      sym_cmd_instruction,
      sym_label_instruction,
      sym_expose_instruction,
      sym_env_instruction,
      sym_add_instruction,
      sym_copy_instruction,
      sym_entrypoint_instruction,
      sym_volume_instruction,
      sym_user_instruction,
      sym_workdir_instruction,
      sym_arg_instruction,
      sym_onbuild_instruction,
      sym_stopsignal_instruction,
      sym_healthcheck_instruction,
      sym_shell_instruction,
      sym_maintainer_instruction,
      sym_cross_build_instruction,
  [546] = 14,
    ACTIONS(59), 1,
      aux_sym_cmd_instruction_token1,
    ACTIONS(61), 1,
      anon_sym_NONE,
    ACTIONS(67), 1,
      anon_sym_DASH_DASH,
    ACTIONS(69), 1,
      anon_sym_DQUOTE,
    ACTIONS(71), 1,
      anon_sym_SQUOTE,
    ACTIONS(73), 1,
      sym_heredoc_marker,
    STATE(259), 1,
      sym__healthcheck_raw_fragment,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(57), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(9), 2,
      sym_param,
      aux_sym_add_instruction_repeat1,
    STATE(334), 2,
      sym__healthcheck_cmd_instruction,
      sym__healthcheck_raw_command,
    STATE(22), 3,
      sym__healthcheck_raw_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
    ACTIONS(65), 5,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token5,
      aux_sym__healthcheck_raw_fragment_atom_token9,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
    ACTIONS(63), 6,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token6,
      aux_sym__healthcheck_raw_fragment_atom_token7,
      aux_sym__healthcheck_raw_fragment_atom_token8,
      aux_sym__healthcheck_raw_fragment_atom_token10,
  [604] = 12,
    ACTIONS(59), 1,
      aux_sym_cmd_instruction_token1,
    ACTIONS(67), 1,
      anon_sym_DASH_DASH,
    ACTIONS(69), 1,
      anon_sym_DQUOTE,
    ACTIONS(71), 1,
      anon_sym_SQUOTE,
    ACTIONS(73), 1,
      sym_heredoc_marker,
    STATE(259), 1,
      sym__healthcheck_raw_fragment,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(18), 2,
      sym_param,
      aux_sym_add_instruction_repeat1,
    STATE(352), 2,
      sym__healthcheck_cmd_instruction,
      sym__healthcheck_raw_command,
    STATE(22), 3,
      sym__healthcheck_raw_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
    ACTIONS(65), 5,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token5,
      aux_sym__healthcheck_raw_fragment_atom_token9,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
    ACTIONS(63), 6,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token6,
      aux_sym__healthcheck_raw_fragment_atom_token7,
      aux_sym__healthcheck_raw_fragment_atom_token8,
      aux_sym__healthcheck_raw_fragment_atom_token10,
  [655] = 12,
    ACTIONS(79), 1,
      anon_sym_DASH_DASH,
    ACTIONS(81), 1,
      anon_sym_DQUOTE,
    ACTIONS(83), 1,
      anon_sym_SQUOTE,
    ACTIONS(85), 1,
      anon_sym_LBRACK,
    ACTIONS(87), 1,
      sym__invalid_json_array_shell_command,
    ACTIONS(89), 1,
      sym_heredoc_marker,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(75), 3,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(11), 3,
      sym_param,
      sym_mount_param,
      aux_sym_run_instruction_repeat1,
    STATE(193), 3,
      sym__json_or_shell_command,
      sym_shell_command,
      sym_json_string_array,
    ACTIONS(77), 4,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    STATE(14), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [705] = 12,
    ACTIONS(79), 1,
      anon_sym_DASH_DASH,
    ACTIONS(81), 1,
      anon_sym_DQUOTE,
    ACTIONS(83), 1,
      anon_sym_SQUOTE,
    ACTIONS(85), 1,
      anon_sym_LBRACK,
    ACTIONS(89), 1,
      sym_heredoc_marker,
    ACTIONS(91), 1,
      sym__invalid_json_array_shell_command,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(75), 3,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(35), 3,
      sym_param,
      sym_mount_param,
      aux_sym_run_instruction_repeat1,
    STATE(226), 3,
      sym__json_or_shell_command,
      sym_shell_command,
      sym_json_string_array,
    ACTIONS(77), 4,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    STATE(14), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [755] = 4,
    ACTIONS(95), 1,
      sym__newline,
    STATE(12), 1,
      aux_sym_source_file_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(93), 20,
      ts_builtin_sym_end,
      aux_sym_from_instruction_token1,
      aux_sym_run_instruction_token1,
      aux_sym_cmd_instruction_token1,
      aux_sym_label_instruction_token1,
      aux_sym_expose_instruction_token1,
      aux_sym_env_instruction_token1,
      aux_sym_add_instruction_token1,
      aux_sym_copy_instruction_token1,
      aux_sym_entrypoint_instruction_token1,
      aux_sym_volume_instruction_token1,
      aux_sym_user_instruction_token1,
      aux_sym_workdir_instruction_token1,
      aux_sym_arg_instruction_token1,
      aux_sym_onbuild_instruction_token1,
      aux_sym_stopsignal_instruction_token1,
      aux_sym_healthcheck_instruction_token1,
      aux_sym_shell_instruction_token1,
      aux_sym_maintainer_instruction_token1,
      aux_sym_cross_build_instruction_token1,
  [788] = 10,
    ACTIONS(69), 1,
      anon_sym_DQUOTE,
    ACTIONS(71), 1,
      anon_sym_SQUOTE,
    ACTIONS(102), 1,
      anon_sym_LBRACK,
    ACTIONS(104), 1,
      sym__invalid_json_array_shell_command,
    ACTIONS(106), 1,
      sym_heredoc_marker,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(100), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    STATE(341), 3,
      sym__json_or_shell_command,
      sym_shell_command,
      sym_json_string_array,
    ACTIONS(98), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(19), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [830] = 10,
    ACTIONS(81), 1,
      anon_sym_DQUOTE,
    ACTIONS(83), 1,
      anon_sym_SQUOTE,
    ACTIONS(89), 1,
      sym_heredoc_marker,
    ACTIONS(114), 1,
      sym_required_line_continuation,
    STATE(176), 1,
      aux_sym__healthcheck_raw_command_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(108), 3,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
    ACTIONS(112), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(110), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(17), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [872] = 10,
    ACTIONS(69), 1,
      anon_sym_DQUOTE,
    ACTIONS(71), 1,
      anon_sym_SQUOTE,
    ACTIONS(102), 1,
      anon_sym_LBRACK,
    ACTIONS(106), 1,
      sym_heredoc_marker,
    ACTIONS(116), 1,
      sym__invalid_json_array_shell_command,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(100), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    STATE(346), 3,
      sym__json_or_shell_command,
      sym_shell_command,
      sym_json_string_array,
    ACTIONS(98), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(19), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [914] = 10,
    ACTIONS(69), 1,
      anon_sym_DQUOTE,
    ACTIONS(71), 1,
      anon_sym_SQUOTE,
    ACTIONS(102), 1,
      anon_sym_LBRACK,
    ACTIONS(106), 1,
      sym_heredoc_marker,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(118), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(333), 2,
      sym_shell_command,
      sym_json_string_array,
    ACTIONS(100), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(98), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(19), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [956] = 8,
    ACTIONS(128), 1,
      anon_sym_DQUOTE,
    ACTIONS(131), 1,
      anon_sym_SQUOTE,
    ACTIONS(134), 1,
      sym_heredoc_marker,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(125), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(120), 4,
      sym_required_line_continuation,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
    ACTIONS(122), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(17), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [993] = 5,
    ACTIONS(141), 1,
      anon_sym_DASH_DASH,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(18), 2,
      sym_param,
      aux_sym_add_instruction_repeat1,
    ACTIONS(139), 7,
      sym_heredoc_marker,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token6,
      aux_sym__healthcheck_raw_fragment_atom_token7,
      aux_sym__healthcheck_raw_fragment_atom_token8,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    ACTIONS(137), 8,
      aux_sym_cmd_instruction_token1,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token5,
      aux_sym__healthcheck_raw_fragment_atom_token9,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
  [1024] = 10,
    ACTIONS(69), 1,
      anon_sym_DQUOTE,
    ACTIONS(71), 1,
      anon_sym_SQUOTE,
    ACTIONS(106), 1,
      sym_heredoc_marker,
    ACTIONS(148), 1,
      sym_required_line_continuation,
    STATE(240), 1,
      aux_sym__healthcheck_raw_command_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(108), 2,
      sym__newline,
      ts_builtin_sym_end,
    ACTIONS(146), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(144), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(25), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [1065] = 8,
    ACTIONS(81), 1,
      anon_sym_DQUOTE,
    ACTIONS(83), 1,
      anon_sym_SQUOTE,
    ACTIONS(89), 1,
      sym_heredoc_marker,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(112), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(110), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    ACTIONS(150), 4,
      sym_required_line_continuation,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
    STATE(17), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [1102] = 8,
    ACTIONS(81), 1,
      anon_sym_DQUOTE,
    ACTIONS(83), 1,
      anon_sym_SQUOTE,
    ACTIONS(89), 1,
      sym_heredoc_marker,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(112), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(110), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    ACTIONS(152), 4,
      sym_required_line_continuation,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
    STATE(17), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [1139] = 8,
    ACTIONS(69), 1,
      anon_sym_DQUOTE,
    ACTIONS(71), 1,
      anon_sym_SQUOTE,
    ACTIONS(106), 1,
      sym_heredoc_marker,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(154), 3,
      sym_required_line_continuation,
      sym__newline,
      ts_builtin_sym_end,
    ACTIONS(158), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(156), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(24), 4,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
      aux_sym__healthcheck_raw_fragment_repeat1,
  [1175] = 10,
    ACTIONS(162), 1,
      anon_sym_DOLLAR2,
    ACTIONS(164), 1,
      aux_sym__env_key_token1,
    ACTIONS(166), 1,
      anon_sym_DQUOTE,
    ACTIONS(168), 1,
      anon_sym_SQUOTE,
    STATE(254), 1,
      sym_single_quoted_string,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(160), 2,
      sym__newline,
      ts_builtin_sym_end,
    ACTIONS(170), 3,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
    STATE(81), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_unquoted_string_repeat1,
    STATE(321), 4,
      sym__env_assignment_value,
      sym__env_single_quoted_string_with_trailing_quote,
      sym_double_quoted_string,
      sym_unquoted_string,
  [1215] = 8,
    ACTIONS(69), 1,
      anon_sym_DQUOTE,
    ACTIONS(71), 1,
      anon_sym_SQUOTE,
    ACTIONS(106), 1,
      sym_heredoc_marker,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(172), 3,
      sym_required_line_continuation,
      sym__newline,
      ts_builtin_sym_end,
    ACTIONS(176), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(174), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(26), 4,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
      aux_sym__healthcheck_raw_fragment_repeat1,
  [1251] = 8,
    ACTIONS(184), 1,
      anon_sym_DQUOTE,
    ACTIONS(187), 1,
      anon_sym_SQUOTE,
    ACTIONS(190), 1,
      sym_heredoc_marker,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(120), 3,
      sym_required_line_continuation,
      sym__newline,
      ts_builtin_sym_end,
    ACTIONS(181), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(178), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(25), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [1287] = 8,
    ACTIONS(201), 1,
      anon_sym_DQUOTE,
    ACTIONS(204), 1,
      anon_sym_SQUOTE,
    ACTIONS(207), 1,
      sym_heredoc_marker,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(193), 3,
      sym_required_line_continuation,
      sym__newline,
      ts_builtin_sym_end,
    ACTIONS(198), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(195), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(26), 4,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
      aux_sym__healthcheck_raw_fragment_repeat1,
  [1323] = 8,
    ACTIONS(69), 1,
      anon_sym_DQUOTE,
    ACTIONS(71), 1,
      anon_sym_SQUOTE,
    ACTIONS(106), 1,
      sym_heredoc_marker,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(146), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(150), 3,
      sym_required_line_continuation,
      sym__newline,
      ts_builtin_sym_end,
    ACTIONS(144), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(25), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [1359] = 8,
    ACTIONS(69), 1,
      anon_sym_DQUOTE,
    ACTIONS(71), 1,
      anon_sym_SQUOTE,
    ACTIONS(106), 1,
      sym_heredoc_marker,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(146), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(152), 3,
      sym_required_line_continuation,
      sym__newline,
      ts_builtin_sym_end,
    ACTIONS(144), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(25), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [1395] = 4,
    ACTIONS(212), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(210), 8,
      aux_sym_cmd_instruction_token1,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token5,
      aux_sym__healthcheck_raw_fragment_atom_token9,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(214), 8,
      sym_heredoc_marker,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token6,
      aux_sym__healthcheck_raw_fragment_atom_token7,
      aux_sym__healthcheck_raw_fragment_atom_token8,
      aux_sym__healthcheck_raw_fragment_atom_token10,
      anon_sym_DASH_DASH,
  [1423] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(216), 8,
      aux_sym_cmd_instruction_token1,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token5,
      aux_sym__healthcheck_raw_fragment_atom_token9,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(218), 8,
      sym_heredoc_marker,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token6,
      aux_sym__healthcheck_raw_fragment_atom_token7,
      aux_sym__healthcheck_raw_fragment_atom_token8,
      aux_sym__healthcheck_raw_fragment_atom_token10,
      anon_sym_DASH_DASH,
  [1448] = 9,
    ACTIONS(81), 1,
      anon_sym_DQUOTE,
    ACTIONS(83), 1,
      anon_sym_SQUOTE,
    ACTIONS(89), 1,
      sym_heredoc_marker,
    ACTIONS(224), 1,
      sym__newline,
    STATE(34), 1,
      aux_sym_source_file_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(222), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(220), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(20), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [1485] = 9,
    ACTIONS(69), 1,
      anon_sym_DQUOTE,
    ACTIONS(71), 1,
      anon_sym_SQUOTE,
    ACTIONS(106), 1,
      sym_heredoc_marker,
    ACTIONS(230), 1,
      sym__newline,
    STATE(33), 1,
      aux_sym_source_file_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(228), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(226), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(27), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [1522] = 9,
    ACTIONS(69), 1,
      anon_sym_DQUOTE,
    ACTIONS(71), 1,
      anon_sym_SQUOTE,
    ACTIONS(106), 1,
      sym_heredoc_marker,
    ACTIONS(236), 1,
      sym__newline,
    STATE(70), 1,
      aux_sym_source_file_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(234), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(232), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(28), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [1559] = 9,
    ACTIONS(81), 1,
      anon_sym_DQUOTE,
    ACTIONS(83), 1,
      anon_sym_SQUOTE,
    ACTIONS(89), 1,
      sym_heredoc_marker,
    ACTIONS(236), 1,
      sym__newline,
    STATE(70), 1,
      aux_sym_source_file_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(240), 3,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
    ACTIONS(238), 4,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    STATE(21), 4,
      aux_sym__shell_fragment,
      sym__shell_fragment_atom,
      sym__shell_double_quoted_fragment,
      sym__shell_single_quoted_fragment,
  [1596] = 5,
    ACTIONS(246), 1,
      anon_sym_DASH_DASH,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(35), 3,
      sym_param,
      sym_mount_param,
      aux_sym_run_instruction_repeat1,
    ACTIONS(242), 6,
      sym__invalid_json_array_shell_command,
      sym_heredoc_marker,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token10,
      anon_sym_LBRACK,
    ACTIONS(244), 6,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
  [1625] = 7,
    ACTIONS(249), 1,
      anon_sym_DOLLAR2,
    ACTIONS(251), 1,
      anon_sym_DQUOTE,
    ACTIONS(253), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(325), 3,
      sym_double_quoted_string,
      sym_single_quoted_string,
      sym__spaced_env_value,
    ACTIONS(255), 4,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      sym__non_newline_whitespace,
    STATE(42), 5,
      sym__immediate_expansion,
      sym__imm_expansion,
      sym__spaced_env_value_atom,
      sym__spaced_env_value_fragment,
      aux_sym__spaced_env_value_repeat1,
  [1657] = 5,
    ACTIONS(261), 1,
      anon_sym_COMMA,
    STATE(38), 1,
      aux_sym_mount_param_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(259), 6,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(257), 7,
      sym__invalid_json_array_shell_command,
      sym_heredoc_marker,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token10,
      anon_sym_DASH_DASH,
      anon_sym_LBRACK,
  [1685] = 5,
    ACTIONS(261), 1,
      anon_sym_COMMA,
    STATE(40), 1,
      aux_sym_mount_param_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(265), 6,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(263), 7,
      sym__invalid_json_array_shell_command,
      sym_heredoc_marker,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token10,
      anon_sym_DASH_DASH,
      anon_sym_LBRACK,
  [1713] = 4,
    ACTIONS(267), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(271), 6,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(269), 8,
      sym__invalid_json_array_shell_command,
      sym_heredoc_marker,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token10,
      anon_sym_DASH_DASH,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [1739] = 5,
    ACTIONS(277), 1,
      anon_sym_COMMA,
    STATE(40), 1,
      aux_sym_mount_param_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(275), 6,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(273), 7,
      sym__invalid_json_array_shell_command,
      sym_heredoc_marker,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token10,
      anon_sym_DASH_DASH,
      anon_sym_LBRACK,
  [1767] = 7,
    ACTIONS(282), 1,
      anon_sym_DOLLAR2,
    ACTIONS(286), 1,
      sym_required_line_continuation,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(280), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(178), 2,
      sym__spaced_label_value_continuation,
      aux_sym__spaced_label_value_repeat2,
    ACTIONS(284), 4,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      aux_sym__spaced_label_value_fragment_token1,
      sym__non_newline_whitespace,
    STATE(54), 5,
      sym__immediate_expansion,
      sym__imm_expansion,
      sym__spaced_label_value_atom,
      sym__spaced_label_value_fragment,
      aux_sym__spaced_label_value_repeat1,
  [1799] = 7,
    ACTIONS(249), 1,
      anon_sym_DOLLAR2,
    ACTIONS(292), 1,
      sym_required_line_continuation,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(288), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(203), 2,
      sym__spaced_env_value_continuation,
      aux_sym__spaced_env_value_repeat2,
    ACTIONS(290), 4,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      sym__non_newline_whitespace,
    STATE(56), 5,
      sym__immediate_expansion,
      sym__imm_expansion,
      sym__spaced_env_value_atom,
      sym__spaced_env_value_fragment,
      aux_sym__spaced_env_value_repeat1,
  [1831] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(296), 5,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(294), 9,
      sym_required_line_continuation,
      sym__newline,
      sym_heredoc_marker,
      sym__heredoc_nl,
      ts_builtin_sym_end,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
  [1854] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(300), 5,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(298), 9,
      sym_required_line_continuation,
      sym__newline,
      sym_heredoc_marker,
      sym__heredoc_nl,
      ts_builtin_sym_end,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
  [1877] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(304), 5,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(302), 9,
      sym_required_line_continuation,
      sym__newline,
      sym_heredoc_marker,
      sym__heredoc_nl,
      ts_builtin_sym_end,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
  [1900] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(308), 6,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(306), 8,
      sym__invalid_json_array_shell_command,
      sym_heredoc_marker,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token10,
      anon_sym_DASH_DASH,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [1923] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(275), 6,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(273), 8,
      sym__invalid_json_array_shell_command,
      sym_heredoc_marker,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token10,
      anon_sym_DASH_DASH,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [1946] = 4,
    ACTIONS(310), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(210), 6,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(214), 7,
      sym__invalid_json_array_shell_command,
      sym_heredoc_marker,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token10,
      anon_sym_DASH_DASH,
      anon_sym_LBRACK,
  [1971] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(314), 5,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(312), 9,
      sym_required_line_continuation,
      sym__newline,
      sym_heredoc_marker,
      sym__heredoc_nl,
      ts_builtin_sym_end,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
  [1994] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(318), 5,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(316), 9,
      sym_required_line_continuation,
      sym__newline,
      sym_heredoc_marker,
      sym__heredoc_nl,
      ts_builtin_sym_end,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
  [2017] = 5,
    ACTIONS(249), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(320), 3,
      sym_required_line_continuation,
      sym__newline,
      ts_builtin_sym_end,
    ACTIONS(290), 4,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      sym__non_newline_whitespace,
    STATE(56), 5,
      sym__immediate_expansion,
      sym__imm_expansion,
      sym__spaced_env_value_atom,
      sym__spaced_env_value_fragment,
      aux_sym__spaced_env_value_repeat1,
  [2043] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(324), 5,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(322), 8,
      sym_required_line_continuation,
      sym__newline,
      sym_heredoc_marker,
      ts_builtin_sym_end,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
  [2065] = 5,
    ACTIONS(282), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(326), 3,
      sym_required_line_continuation,
      sym__newline,
      ts_builtin_sym_end,
    ACTIONS(284), 4,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      aux_sym__spaced_label_value_fragment_token1,
      sym__non_newline_whitespace,
    STATE(54), 5,
      sym__immediate_expansion,
      sym__imm_expansion,
      sym__spaced_label_value_atom,
      sym__spaced_label_value_fragment,
      aux_sym__spaced_label_value_repeat1,
  [2091] = 5,
    ACTIONS(330), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(328), 3,
      sym_required_line_continuation,
      sym__newline,
      ts_builtin_sym_end,
    ACTIONS(333), 4,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      aux_sym__spaced_label_value_fragment_token1,
      sym__non_newline_whitespace,
    STATE(54), 5,
      sym__immediate_expansion,
      sym__imm_expansion,
      sym__spaced_label_value_atom,
      sym__spaced_label_value_fragment,
      aux_sym__spaced_label_value_repeat1,
  [2117] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(300), 5,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(298), 8,
      sym_required_line_continuation,
      sym__newline,
      sym_heredoc_marker,
      ts_builtin_sym_end,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
  [2139] = 5,
    ACTIONS(338), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(336), 3,
      sym_required_line_continuation,
      sym__newline,
      ts_builtin_sym_end,
    ACTIONS(341), 4,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      sym__non_newline_whitespace,
    STATE(56), 5,
      sym__immediate_expansion,
      sym__imm_expansion,
      sym__spaced_env_value_atom,
      sym__spaced_env_value_fragment,
      aux_sym__spaced_env_value_repeat1,
  [2165] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(314), 5,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(312), 8,
      sym_required_line_continuation,
      sym__newline,
      sym_heredoc_marker,
      ts_builtin_sym_end,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
  [2187] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(318), 5,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(316), 8,
      sym_required_line_continuation,
      sym__newline,
      sym_heredoc_marker,
      ts_builtin_sym_end,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
  [2209] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(296), 5,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(294), 8,
      sym_required_line_continuation,
      sym__newline,
      sym_heredoc_marker,
      ts_builtin_sym_end,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
  [2231] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(304), 5,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(302), 8,
      sym_required_line_continuation,
      sym__newline,
      sym_heredoc_marker,
      ts_builtin_sym_end,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
  [2253] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(216), 6,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(218), 7,
      sym__invalid_json_array_shell_command,
      sym_heredoc_marker,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token10,
      anon_sym_DASH_DASH,
      anon_sym_LBRACK,
  [2275] = 7,
    ACTIONS(344), 1,
      anon_sym_DOLLAR2,
    ACTIONS(346), 1,
      anon_sym_DQUOTE,
    ACTIONS(348), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(350), 3,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
    STATE(65), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_unquoted_string_repeat1,
    STATE(199), 3,
      sym_double_quoted_string,
      sym_single_quoted_string,
      sym_unquoted_string,
  [2304] = 12,
    ACTIONS(102), 1,
      anon_sym_LBRACK,
    ACTIONS(352), 1,
      aux_sym_path_token1,
    ACTIONS(354), 1,
      aux_sym_path_with_heredoc_token1,
    ACTIONS(356), 1,
      anon_sym_DOLLAR,
    ACTIONS(358), 1,
      anon_sym_DASH_DASH,
    ACTIONS(360), 1,
      sym_heredoc_marker,
    STATE(121), 1,
      aux_sym_add_instruction_repeat2,
    STATE(144), 1,
      sym_expansion,
    STATE(353), 1,
      sym_json_string_array,
    STATE(384), 1,
      sym_path_with_heredoc,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(67), 2,
      sym_param,
      aux_sym_add_instruction_repeat1,
  [2343] = 12,
    ACTIONS(102), 1,
      anon_sym_LBRACK,
    ACTIONS(352), 1,
      aux_sym_path_token1,
    ACTIONS(354), 1,
      aux_sym_path_with_heredoc_token1,
    ACTIONS(356), 1,
      anon_sym_DOLLAR,
    ACTIONS(358), 1,
      anon_sym_DASH_DASH,
    ACTIONS(360), 1,
      sym_heredoc_marker,
    STATE(126), 1,
      aux_sym_add_instruction_repeat2,
    STATE(144), 1,
      sym_expansion,
    STATE(355), 1,
      sym_json_string_array,
    STATE(384), 1,
      sym_path_with_heredoc,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(68), 2,
      sym_param,
      aux_sym_add_instruction_repeat1,
  [2382] = 6,
    ACTIONS(344), 1,
      anon_sym_DOLLAR2,
    ACTIONS(364), 1,
      aux_sym__label_key_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(366), 3,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
    STATE(69), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_unquoted_string_repeat1,
    ACTIONS(362), 4,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
  [2409] = 7,
    ACTIONS(368), 1,
      anon_sym_DOLLAR2,
    ACTIONS(370), 1,
      anon_sym_DQUOTE,
    ACTIONS(372), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(374), 3,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
    STATE(85), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_unquoted_string_repeat1,
    STATE(271), 3,
      sym_double_quoted_string,
      sym_single_quoted_string,
      sym_unquoted_string,
  [2438] = 12,
    ACTIONS(102), 1,
      anon_sym_LBRACK,
    ACTIONS(352), 1,
      aux_sym_path_token1,
    ACTIONS(354), 1,
      aux_sym_path_with_heredoc_token1,
    ACTIONS(356), 1,
      anon_sym_DOLLAR,
    ACTIONS(358), 1,
      anon_sym_DASH_DASH,
    ACTIONS(360), 1,
      sym_heredoc_marker,
    STATE(120), 1,
      aux_sym_add_instruction_repeat2,
    STATE(144), 1,
      sym_expansion,
    STATE(343), 1,
      sym_json_string_array,
    STATE(384), 1,
      sym_path_with_heredoc,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(108), 2,
      sym_param,
      aux_sym_add_instruction_repeat1,
  [2477] = 12,
    ACTIONS(102), 1,
      anon_sym_LBRACK,
    ACTIONS(352), 1,
      aux_sym_path_token1,
    ACTIONS(354), 1,
      aux_sym_path_with_heredoc_token1,
    ACTIONS(356), 1,
      anon_sym_DOLLAR,
    ACTIONS(358), 1,
      anon_sym_DASH_DASH,
    ACTIONS(360), 1,
      sym_heredoc_marker,
    STATE(125), 1,
      aux_sym_add_instruction_repeat2,
    STATE(144), 1,
      sym_expansion,
    STATE(336), 1,
      sym_json_string_array,
    STATE(384), 1,
      sym_path_with_heredoc,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(108), 2,
      sym_param,
      aux_sym_add_instruction_repeat1,
  [2516] = 6,
    ACTIONS(378), 1,
      anon_sym_DOLLAR2,
    ACTIONS(381), 1,
      aux_sym__label_key_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(383), 3,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
    STATE(69), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_unquoted_string_repeat1,
    ACTIONS(376), 4,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
  [2543] = 5,
    ACTIONS(388), 1,
      sym__newline,
    STATE(70), 1,
      aux_sym_source_file_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(93), 5,
      sym_heredoc_marker,
      aux_sym__healthcheck_raw_fragment_atom_token2,
      aux_sym__healthcheck_raw_fragment_atom_token3,
      aux_sym__healthcheck_raw_fragment_atom_token4,
      aux_sym__healthcheck_raw_fragment_atom_token10,
    ACTIONS(386), 5,
      aux_sym__healthcheck_raw_fragment_atom_token11,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym__shell_fragment_atom_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
  [2568] = 5,
    ACTIONS(282), 1,
      anon_sym_DOLLAR2,
    STATE(360), 1,
      sym__spaced_label_value,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(391), 4,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      aux_sym__spaced_label_value_fragment_token1,
      sym__non_newline_whitespace,
    STATE(41), 5,
      sym__immediate_expansion,
      sym__imm_expansion,
      sym__spaced_label_value_atom,
      sym__spaced_label_value_fragment,
      aux_sym__spaced_label_value_repeat1,
  [2592] = 6,
    ACTIONS(393), 1,
      anon_sym_DOLLAR2,
    ACTIONS(395), 1,
      anon_sym_DQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(399), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(397), 3,
      aux_sym_double_quoted_string_token1,
      aux_sym_double_quoted_string_token2,
      sym_double_quoted_escape_sequence,
    STATE(74), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_double_quoted_string_repeat1,
  [2617] = 4,
    ACTIONS(282), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(401), 4,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      aux_sym__spaced_label_value_fragment_token1,
      sym__non_newline_whitespace,
    STATE(53), 5,
      sym__immediate_expansion,
      sym__imm_expansion,
      sym__spaced_label_value_atom,
      sym__spaced_label_value_fragment,
      aux_sym__spaced_label_value_repeat1,
  [2638] = 6,
    ACTIONS(393), 1,
      anon_sym_DOLLAR2,
    ACTIONS(403), 1,
      anon_sym_DQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(407), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(405), 3,
      aux_sym_double_quoted_string_token1,
      aux_sym_double_quoted_string_token2,
      sym_double_quoted_escape_sequence,
    STATE(80), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_double_quoted_string_repeat1,
  [2663] = 6,
    ACTIONS(393), 1,
      anon_sym_DOLLAR2,
    ACTIONS(409), 1,
      anon_sym_DQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(413), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(411), 3,
      aux_sym_double_quoted_string_token1,
      aux_sym_double_quoted_string_token2,
      sym_double_quoted_escape_sequence,
    STATE(78), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_double_quoted_string_repeat1,
  [2688] = 6,
    ACTIONS(417), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(419), 1,
      anon_sym_DOLLAR2,
    ACTIONS(422), 1,
      aux_sym_image_name_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(76), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_image_name_repeat1,
    ACTIONS(415), 4,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_COLON,
      anon_sym_AT,
  [2713] = 6,
    ACTIONS(427), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(429), 1,
      anon_sym_DOLLAR2,
    ACTIONS(431), 1,
      aux_sym_image_name_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(76), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_image_name_repeat1,
    ACTIONS(425), 4,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_COLON,
      anon_sym_AT,
  [2738] = 6,
    ACTIONS(393), 1,
      anon_sym_DOLLAR2,
    ACTIONS(433), 1,
      anon_sym_DQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(407), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(405), 3,
      aux_sym_double_quoted_string_token1,
      aux_sym_double_quoted_string_token2,
      sym_double_quoted_escape_sequence,
    STATE(80), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_double_quoted_string_repeat1,
  [2763] = 7,
    ACTIONS(437), 1,
      aux_sym__label_key_token1,
    ACTIONS(439), 1,
      anon_sym_DQUOTE,
    ACTIONS(441), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(435), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(82), 2,
      sym_label_pair,
      aux_sym_label_instruction_repeat1,
    STATE(374), 3,
      sym__label_key,
      sym_double_quoted_string,
      sym_single_quoted_string,
  [2790] = 6,
    ACTIONS(443), 1,
      anon_sym_DOLLAR2,
    ACTIONS(446), 1,
      anon_sym_DQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(451), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(448), 3,
      aux_sym_double_quoted_string_token1,
      aux_sym_double_quoted_string_token2,
      sym_double_quoted_escape_sequence,
    STATE(80), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_double_quoted_string_repeat1,
  [2815] = 6,
    ACTIONS(162), 1,
      anon_sym_DOLLAR2,
    ACTIONS(364), 1,
      aux_sym__env_key_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(362), 2,
      sym__newline,
      ts_builtin_sym_end,
    ACTIONS(454), 3,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
    STATE(84), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_unquoted_string_repeat1,
  [2840] = 7,
    ACTIONS(458), 1,
      aux_sym__label_key_token1,
    ACTIONS(461), 1,
      anon_sym_DQUOTE,
    ACTIONS(464), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(456), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(82), 2,
      sym_label_pair,
      aux_sym_label_instruction_repeat1,
    STATE(374), 3,
      sym__label_key,
      sym_double_quoted_string,
      sym_single_quoted_string,
  [2867] = 4,
    ACTIONS(249), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(467), 4,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      sym__non_newline_whitespace,
    STATE(51), 5,
      sym__immediate_expansion,
      sym__imm_expansion,
      sym__spaced_env_value_atom,
      sym__spaced_env_value_fragment,
      aux_sym__spaced_env_value_repeat1,
  [2888] = 6,
    ACTIONS(381), 1,
      aux_sym__env_key_token1,
    ACTIONS(469), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(376), 2,
      sym__newline,
      ts_builtin_sym_end,
    ACTIONS(472), 3,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
    STATE(84), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_unquoted_string_repeat1,
  [2913] = 5,
    ACTIONS(368), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(362), 3,
      sym__newline,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
    ACTIONS(475), 3,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
    STATE(86), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_unquoted_string_repeat1,
  [2936] = 5,
    ACTIONS(477), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(376), 3,
      sym__newline,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
    ACTIONS(480), 3,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
    STATE(86), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_unquoted_string_repeat1,
  [2959] = 6,
    ACTIONS(429), 1,
      anon_sym_DOLLAR2,
    ACTIONS(485), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(487), 1,
      aux_sym_image_name_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(77), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_image_name_repeat1,
    ACTIONS(483), 4,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_COLON,
      anon_sym_AT,
  [2984] = 6,
    ACTIONS(393), 1,
      anon_sym_DOLLAR2,
    ACTIONS(489), 1,
      anon_sym_DQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(493), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(491), 3,
      aux_sym_double_quoted_string_token1,
      aux_sym_double_quoted_string_token2,
      sym_double_quoted_escape_sequence,
    STATE(89), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_double_quoted_string_repeat1,
  [3009] = 6,
    ACTIONS(393), 1,
      anon_sym_DOLLAR2,
    ACTIONS(495), 1,
      anon_sym_DQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(407), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(405), 3,
      aux_sym_double_quoted_string_token1,
      aux_sym_double_quoted_string_token2,
      sym_double_quoted_escape_sequence,
    STATE(80), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_double_quoted_string_repeat1,
  [3034] = 6,
    ACTIONS(393), 1,
      anon_sym_DOLLAR2,
    ACTIONS(497), 1,
      anon_sym_DQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(501), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(499), 3,
      aux_sym_double_quoted_string_token1,
      aux_sym_double_quoted_string_token2,
      sym_double_quoted_escape_sequence,
    STATE(91), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_double_quoted_string_repeat1,
  [3059] = 6,
    ACTIONS(393), 1,
      anon_sym_DOLLAR2,
    ACTIONS(503), 1,
      anon_sym_DQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(407), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(405), 3,
      aux_sym_double_quoted_string_token1,
      aux_sym_double_quoted_string_token2,
      sym_double_quoted_escape_sequence,
    STATE(80), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_double_quoted_string_repeat1,
  [3084] = 6,
    ACTIONS(393), 1,
      anon_sym_DOLLAR2,
    ACTIONS(505), 1,
      anon_sym_DQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(509), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(507), 3,
      aux_sym_double_quoted_string_token1,
      aux_sym_double_quoted_string_token2,
      sym_double_quoted_escape_sequence,
    STATE(93), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_double_quoted_string_repeat1,
  [3109] = 6,
    ACTIONS(393), 1,
      anon_sym_DOLLAR2,
    ACTIONS(511), 1,
      anon_sym_DQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(407), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(405), 3,
      aux_sym_double_quoted_string_token1,
      aux_sym_double_quoted_string_token2,
      sym_double_quoted_escape_sequence,
    STATE(80), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_double_quoted_string_repeat1,
  [3134] = 5,
    ACTIONS(515), 1,
      aux_sym_path_token2,
    ACTIONS(517), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(98), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_path_repeat1,
    ACTIONS(513), 4,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
  [3156] = 5,
    ACTIONS(521), 1,
      aux_sym__immediate_user_name_or_group_fragment_token1,
    ACTIONS(524), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(519), 3,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_COLON,
    STATE(95), 4,
      sym__immediate_user_name_or_group_fragment,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym__user_name_or_group_repeat1,
  [3178] = 7,
    ACTIONS(527), 1,
      aux_sym__label_key_token1,
    ACTIONS(529), 1,
      anon_sym_DQUOTE,
    ACTIONS(531), 1,
      anon_sym_SQUOTE,
    STATE(332), 1,
      sym__spaced_label_pair,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(79), 2,
      sym_label_pair,
      aux_sym_label_instruction_repeat1,
    STATE(221), 3,
      sym__label_key,
      sym_double_quoted_string,
      sym_single_quoted_string,
  [3204] = 5,
    ACTIONS(517), 1,
      anon_sym_DOLLAR2,
    ACTIONS(535), 1,
      aux_sym_path_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(94), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_path_repeat1,
    ACTIONS(533), 4,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
  [3226] = 5,
    ACTIONS(539), 1,
      aux_sym_path_token2,
    ACTIONS(542), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(98), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_path_repeat1,
    ACTIONS(537), 4,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
  [3248] = 3,
    ACTIONS(547), 1,
      aux_sym__label_key_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 8,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
  [3266] = 3,
    ACTIONS(551), 1,
      aux_sym__label_key_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 8,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
  [3284] = 6,
    ACTIONS(555), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(557), 1,
      anon_sym_DOLLAR2,
    ACTIONS(559), 1,
      aux_sym_image_tag_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(553), 3,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_AT,
    STATE(102), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_image_tag_repeat1,
  [3308] = 6,
    ACTIONS(563), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(565), 1,
      anon_sym_DOLLAR2,
    ACTIONS(568), 1,
      aux_sym_image_tag_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(561), 3,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_AT,
    STATE(102), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_image_tag_repeat1,
  [3332] = 5,
    ACTIONS(573), 1,
      aux_sym__immediate_user_name_or_group_fragment_token1,
    ACTIONS(575), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(571), 3,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_COLON,
    STATE(104), 4,
      sym__immediate_user_name_or_group_fragment,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym__user_name_or_group_repeat1,
  [3354] = 5,
    ACTIONS(575), 1,
      anon_sym_DOLLAR2,
    ACTIONS(579), 1,
      aux_sym__immediate_user_name_or_group_fragment_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(577), 3,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_COLON,
    STATE(95), 4,
      sym__immediate_user_name_or_group_fragment,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym__user_name_or_group_repeat1,
  [3376] = 5,
    ACTIONS(581), 1,
      aux_sym_path_token2,
    ACTIONS(584), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(537), 3,
      sym__newline,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
    STATE(105), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_path_repeat1,
  [3397] = 6,
    ACTIONS(589), 1,
      aux_sym__immediate_user_name_or_group_fragment_token1,
    ACTIONS(592), 1,
      anon_sym_DOLLAR2,
    STATE(106), 1,
      aux_sym__immediate_user_name_or_group,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(587), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(258), 3,
      sym__immediate_user_name_or_group_fragment,
      sym__immediate_expansion,
      sym__imm_expansion,
  [3420] = 6,
    ACTIONS(575), 1,
      anon_sym_DOLLAR2,
    ACTIONS(597), 1,
      aux_sym__immediate_user_name_or_group_fragment_token1,
    STATE(106), 1,
      aux_sym__immediate_user_name_or_group,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(595), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(258), 3,
      sym__immediate_user_name_or_group_fragment,
      sym__immediate_expansion,
      sym__imm_expansion,
  [3443] = 5,
    ACTIONS(137), 1,
      aux_sym_path_token1,
    ACTIONS(599), 1,
      anon_sym_DASH_DASH,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(108), 2,
      sym_param,
      aux_sym_add_instruction_repeat1,
    ACTIONS(139), 4,
      sym_heredoc_marker,
      aux_sym_path_with_heredoc_token1,
      anon_sym_DOLLAR,
      anon_sym_LBRACK,
  [3464] = 5,
    ACTIONS(604), 1,
      aux_sym_path_token2,
    ACTIONS(606), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(602), 3,
      sym__newline,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
    STATE(105), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_path_repeat1,
  [3485] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 8,
      sym_required_line_continuation,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      aux_sym__spaced_label_value_fragment_token1,
      sym__non_newline_whitespace,
  [3500] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 8,
      sym_required_line_continuation,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      sym__non_newline_whitespace,
  [3515] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 8,
      sym_required_line_continuation,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      aux_sym__spaced_label_value_fragment_token1,
      sym__non_newline_whitespace,
  [3530] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 8,
      sym_required_line_continuation,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      sym__non_newline_whitespace,
  [3545] = 6,
    ACTIONS(610), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(612), 1,
      anon_sym_DOLLAR2,
    ACTIONS(614), 1,
      aux_sym_image_digest_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(608), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(115), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_image_digest_repeat1,
  [3568] = 6,
    ACTIONS(618), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(620), 1,
      anon_sym_DOLLAR2,
    ACTIONS(623), 1,
      aux_sym_image_digest_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(616), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(115), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_image_digest_repeat1,
  [3591] = 5,
    ACTIONS(606), 1,
      anon_sym_DOLLAR2,
    ACTIONS(628), 1,
      aux_sym_path_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(626), 3,
      sym__newline,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
    STATE(109), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_path_repeat1,
  [3612] = 8,
    ACTIONS(630), 1,
      anon_sym_DOLLAR,
    ACTIONS(632), 1,
      aux_sym_image_name_token1,
    ACTIONS(634), 1,
      anon_sym_DASH_DASH,
    STATE(87), 1,
      sym_expansion,
    STATE(118), 1,
      sym_image_name,
    STATE(198), 1,
      sym_param,
    STATE(299), 1,
      sym_image_spec,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [3638] = 6,
    ACTIONS(638), 1,
      anon_sym_COLON,
    ACTIONS(640), 1,
      anon_sym_AT,
    STATE(218), 1,
      sym_image_tag,
    STATE(279), 1,
      sym_image_digest,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(636), 3,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym_from_instruction_token2,
  [3660] = 5,
    ACTIONS(644), 1,
      aux_sym__stopsignal_value_token2,
    ACTIONS(647), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(642), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(119), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym__stopsignal_value_repeat1,
  [3680] = 7,
    ACTIONS(652), 1,
      anon_sym_DOLLAR,
    ACTIONS(654), 1,
      sym_heredoc_marker,
    STATE(97), 1,
      sym_expansion,
    STATE(124), 1,
      aux_sym_add_instruction_repeat2,
    STATE(168), 1,
      sym_path_with_heredoc,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(650), 2,
      aux_sym_path_token1,
      aux_sym_path_with_heredoc_token1,
  [3704] = 7,
    ACTIONS(652), 1,
      anon_sym_DOLLAR,
    ACTIONS(654), 1,
      sym_heredoc_marker,
    STATE(97), 1,
      sym_expansion,
    STATE(124), 1,
      aux_sym_add_instruction_repeat2,
    STATE(157), 1,
      sym_path_with_heredoc,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(650), 2,
      aux_sym_path_token1,
      aux_sym_path_with_heredoc_token1,
  [3728] = 3,
    ACTIONS(551), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 6,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_COLON,
      anon_sym_DOLLAR2,
      aux_sym_image_name_token2,
      anon_sym_AT,
  [3744] = 5,
    ACTIONS(658), 1,
      anon_sym_DOLLAR2,
    ACTIONS(660), 1,
      aux_sym_image_alias_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(656), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(128), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_image_alias_repeat1,
  [3764] = 7,
    ACTIONS(665), 1,
      anon_sym_DOLLAR,
    ACTIONS(668), 1,
      sym_heredoc_marker,
    STATE(124), 1,
      aux_sym_add_instruction_repeat2,
    STATE(144), 1,
      sym_expansion,
    STATE(384), 1,
      sym_path_with_heredoc,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(662), 2,
      aux_sym_path_token1,
      aux_sym_path_with_heredoc_token1,
  [3788] = 7,
    ACTIONS(652), 1,
      anon_sym_DOLLAR,
    ACTIONS(654), 1,
      sym_heredoc_marker,
    STATE(97), 1,
      sym_expansion,
    STATE(124), 1,
      aux_sym_add_instruction_repeat2,
    STATE(171), 1,
      sym_path_with_heredoc,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(650), 2,
      aux_sym_path_token1,
      aux_sym_path_with_heredoc_token1,
  [3812] = 7,
    ACTIONS(652), 1,
      anon_sym_DOLLAR,
    ACTIONS(654), 1,
      sym_heredoc_marker,
    STATE(97), 1,
      sym_expansion,
    STATE(124), 1,
      aux_sym_add_instruction_repeat2,
    STATE(145), 1,
      sym_path_with_heredoc,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(650), 2,
      aux_sym_path_token1,
      aux_sym_path_with_heredoc_token1,
  [3836] = 5,
    ACTIONS(673), 1,
      aux_sym__stopsignal_value_token2,
    ACTIONS(675), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(671), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(119), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym__stopsignal_value_repeat1,
  [3856] = 5,
    ACTIONS(679), 1,
      anon_sym_DOLLAR2,
    ACTIONS(682), 1,
      aux_sym_image_alias_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(677), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(128), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_image_alias_repeat1,
  [3876] = 5,
    ACTIONS(675), 1,
      anon_sym_DOLLAR2,
    ACTIONS(687), 1,
      aux_sym__stopsignal_value_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(685), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(127), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym__stopsignal_value_repeat1,
  [3896] = 8,
    ACTIONS(102), 1,
      anon_sym_LBRACK,
    ACTIONS(689), 1,
      aux_sym__healthcheck_raw_fragment_atom_token12,
    ACTIONS(691), 1,
      aux_sym_path_token1,
    ACTIONS(693), 1,
      anon_sym_DOLLAR,
    STATE(116), 1,
      sym_expansion,
    STATE(247), 1,
      sym_path,
    STATE(356), 1,
      sym_json_string_array,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [3922] = 3,
    ACTIONS(547), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 6,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_COLON,
      anon_sym_DOLLAR2,
      aux_sym_image_name_token2,
      anon_sym_AT,
  [3938] = 3,
    ACTIONS(697), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(695), 6,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_COLON,
      anon_sym_DOLLAR2,
      aux_sym_image_name_token2,
      anon_sym_AT,
  [3954] = 5,
    ACTIONS(658), 1,
      anon_sym_DOLLAR2,
    ACTIONS(701), 1,
      aux_sym_image_alias_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(699), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(123), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_image_alias_repeat1,
  [3974] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 3,
      aux_sym_double_quoted_string_token1,
      aux_sym_double_quoted_string_token2,
      sym_double_quoted_escape_sequence,
    ACTIONS(547), 4,
      anon_sym_DOLLAR2,
      anon_sym_DQUOTE,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
  [3990] = 5,
    ACTIONS(705), 1,
      anon_sym_DOLLAR,
    ACTIONS(708), 1,
      aux_sym_expose_port_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(703), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(135), 3,
      sym_expansion,
      sym_expose_port,
      aux_sym_expose_instruction_repeat1,
  [4010] = 4,
    ACTIONS(210), 1,
      aux_sym_path_token1,
    ACTIONS(711), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(214), 5,
      sym_heredoc_marker,
      aux_sym_path_with_heredoc_token1,
      anon_sym_DOLLAR,
      anon_sym_DASH_DASH,
      anon_sym_LBRACK,
  [4028] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 3,
      aux_sym_double_quoted_string_token1,
      aux_sym_double_quoted_string_token2,
      sym_double_quoted_escape_sequence,
    ACTIONS(551), 4,
      anon_sym_DOLLAR2,
      anon_sym_DQUOTE,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
  [4044] = 3,
    ACTIONS(547), 1,
      aux_sym__env_key_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 6,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
  [4060] = 5,
    ACTIONS(715), 1,
      anon_sym_DOLLAR,
    ACTIONS(717), 1,
      aux_sym_expose_port_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(713), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(135), 3,
      sym_expansion,
      sym_expose_port,
      aux_sym_expose_instruction_repeat1,
  [4080] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 7,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      sym__non_newline_whitespace,
  [4094] = 3,
    ACTIONS(551), 1,
      aux_sym__env_key_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 6,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
  [4110] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 7,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_unquoted_string_token1,
      aux_sym_unquoted_string_token2,
      aux_sym_unquoted_string_token3,
      sym__non_newline_whitespace,
  [4124] = 5,
    ACTIONS(719), 1,
      anon_sym_SQUOTE,
    STATE(162), 1,
      aux_sym_single_quoted_string_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(721), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(723), 2,
      aux_sym_single_quoted_string_token1,
      sym_single_quoted_escape_sequence,
  [4143] = 5,
    ACTIONS(533), 1,
      sym__non_newline_whitespace,
    ACTIONS(725), 1,
      aux_sym_path_token2,
    ACTIONS(727), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(158), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_path_repeat1,
  [4162] = 5,
    ACTIONS(731), 1,
      sym__non_newline_whitespace,
    ACTIONS(733), 1,
      sym__heredoc_nl,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(729), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(196), 2,
      sym_heredoc_block,
      aux_sym_run_instruction_repeat2,
  [4181] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(737), 2,
      anon_sym_SLASHtcp,
      anon_sym_SLASHudp,
    ACTIONS(735), 4,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR,
      aux_sym_expose_port_token1,
  [4196] = 3,
    ACTIONS(216), 1,
      aux_sym_path_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(218), 5,
      sym_heredoc_marker,
      aux_sym_path_with_heredoc_token1,
      anon_sym_DOLLAR,
      anon_sym_DASH_DASH,
      anon_sym_LBRACK,
  [4211] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(695), 6,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
      aux_sym_path_token2,
      anon_sym_DOLLAR2,
      sym__non_newline_whitespace,
  [4224] = 3,
    ACTIONS(547), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 5,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_image_tag_token1,
      anon_sym_AT,
  [4239] = 6,
    ACTIONS(346), 1,
      anon_sym_DQUOTE,
    ACTIONS(739), 1,
      aux_sym__stopsignal_value_token1,
    ACTIONS(741), 1,
      anon_sym_DOLLAR,
    STATE(129), 1,
      sym_expansion,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(329), 2,
      sym__stopsignal_value,
      sym_double_quoted_string,
  [4260] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 6,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
      aux_sym_path_token2,
      anon_sym_DOLLAR2,
      sym__non_newline_whitespace,
  [4273] = 5,
    ACTIONS(743), 1,
      anon_sym_SQUOTE,
    STATE(152), 1,
      aux_sym_single_quoted_string_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(745), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(748), 2,
      aux_sym_single_quoted_string_token1,
      sym_single_quoted_escape_sequence,
  [4292] = 3,
    ACTIONS(551), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 5,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_image_tag_token1,
      anon_sym_AT,
  [4307] = 5,
    ACTIONS(751), 1,
      anon_sym_SQUOTE,
    STATE(152), 1,
      aux_sym_single_quoted_string_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(753), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(755), 2,
      aux_sym_single_quoted_string_token1,
      sym_single_quoted_escape_sequence,
  [4326] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 6,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
      aux_sym_path_token2,
      anon_sym_DOLLAR2,
      sym__non_newline_whitespace,
  [4339] = 5,
    ACTIONS(575), 1,
      anon_sym_DOLLAR2,
    ACTIONS(597), 1,
      aux_sym__immediate_user_name_or_group_fragment_token1,
    STATE(107), 1,
      aux_sym__immediate_user_name_or_group,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(258), 3,
      sym__immediate_user_name_or_group_fragment,
      sym__immediate_expansion,
      sym__imm_expansion,
  [4358] = 5,
    ACTIONS(731), 1,
      sym__non_newline_whitespace,
    ACTIONS(733), 1,
      sym__heredoc_nl,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(757), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(222), 2,
      sym_heredoc_block,
      aux_sym_run_instruction_repeat2,
  [4377] = 5,
    ACTIONS(513), 1,
      sym__non_newline_whitespace,
    ACTIONS(727), 1,
      anon_sym_DOLLAR2,
    ACTIONS(759), 1,
      aux_sym_path_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(166), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_path_repeat1,
  [4396] = 5,
    ACTIONS(763), 1,
      aux_sym__env_key_token1,
    STATE(393), 1,
      sym__env_key,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(761), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(159), 2,
      sym_env_pair,
      aux_sym_env_instruction_repeat1,
  [4415] = 5,
    ACTIONS(766), 1,
      anon_sym_SQUOTE,
    STATE(152), 1,
      aux_sym_single_quoted_string_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(753), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(755), 2,
      aux_sym_single_quoted_string_token1,
      sym_single_quoted_escape_sequence,
  [4434] = 5,
    ACTIONS(768), 1,
      anon_sym_SQUOTE,
    STATE(154), 1,
      aux_sym_single_quoted_string_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(770), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(772), 2,
      aux_sym_single_quoted_string_token1,
      sym_single_quoted_escape_sequence,
  [4453] = 5,
    ACTIONS(774), 1,
      anon_sym_SQUOTE,
    STATE(152), 1,
      aux_sym_single_quoted_string_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(753), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(755), 2,
      aux_sym_single_quoted_string_token1,
      sym_single_quoted_escape_sequence,
  [4472] = 5,
    ACTIONS(778), 1,
      aux_sym__env_key_token1,
    STATE(393), 1,
      sym__env_key,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(776), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(159), 2,
      sym_env_pair,
      aux_sym_env_instruction_repeat1,
  [4491] = 5,
    ACTIONS(780), 1,
      anon_sym_SQUOTE,
    STATE(165), 1,
      aux_sym_single_quoted_string_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(782), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(784), 2,
      aux_sym_single_quoted_string_token1,
      sym_single_quoted_escape_sequence,
  [4510] = 5,
    ACTIONS(786), 1,
      anon_sym_SQUOTE,
    STATE(152), 1,
      aux_sym_single_quoted_string_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(753), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(755), 2,
      aux_sym_single_quoted_string_token1,
      sym_single_quoted_escape_sequence,
  [4529] = 5,
    ACTIONS(537), 1,
      sym__non_newline_whitespace,
    ACTIONS(788), 1,
      aux_sym_path_token2,
    ACTIONS(791), 1,
      anon_sym_DOLLAR2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(166), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_path_repeat1,
  [4548] = 5,
    ACTIONS(794), 1,
      anon_sym_SQUOTE,
    STATE(169), 1,
      aux_sym_single_quoted_string_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(796), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(798), 2,
      aux_sym_single_quoted_string_token1,
      sym_single_quoted_escape_sequence,
  [4567] = 5,
    ACTIONS(731), 1,
      sym__non_newline_whitespace,
    ACTIONS(733), 1,
      sym__heredoc_nl,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(800), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(213), 2,
      sym_heredoc_block,
      aux_sym_run_instruction_repeat2,
  [4586] = 5,
    ACTIONS(802), 1,
      anon_sym_SQUOTE,
    STATE(152), 1,
      aux_sym_single_quoted_string_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(753), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(755), 2,
      aux_sym_single_quoted_string_token1,
      sym_single_quoted_escape_sequence,
  [4605] = 5,
    ACTIONS(804), 1,
      anon_sym_SQUOTE,
    STATE(160), 1,
      aux_sym_single_quoted_string_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(806), 2,
      anon_sym_BSLASH,
      anon_sym_BQUOTE2,
    ACTIONS(808), 2,
      aux_sym_single_quoted_string_token1,
      sym_single_quoted_escape_sequence,
  [4624] = 5,
    ACTIONS(731), 1,
      sym__non_newline_whitespace,
    ACTIONS(733), 1,
      sym__heredoc_nl,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(810), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(214), 2,
      sym_heredoc_block,
      aux_sym_run_instruction_repeat2,
  [4643] = 6,
    ACTIONS(292), 1,
      sym_required_line_continuation,
    ACTIONS(812), 1,
      anon_sym_EQ,
    ACTIONS(814), 1,
      aux_sym__spaced_env_pair_token1,
    STATE(201), 1,
      sym__spaced_env_value_continuation,
    STATE(335), 1,
      sym__continued_spaced_env_value,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [4663] = 5,
    ACTIONS(693), 1,
      anon_sym_DOLLAR,
    STATE(116), 1,
      sym_expansion,
    STATE(292), 1,
      sym_path,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(689), 2,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym_path_token1,
  [4681] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(816), 5,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym__label_key_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
  [4693] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(818), 5,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym__label_key_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
  [4705] = 4,
    ACTIONS(114), 1,
      sym_required_line_continuation,
    STATE(179), 1,
      aux_sym__healthcheck_raw_command_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(820), 3,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
  [4721] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(824), 1,
      anon_sym_SQUOTE2,
    STATE(215), 1,
      aux_sym__shell_single_quoted_fragment_repeat1,
    ACTIONS(822), 4,
      sym_line_continuation,
      sym_required_line_continuation,
      aux_sym__shell_double_quoted_fragment_token2,
      aux_sym__shell_single_quoted_fragment_token1,
  [4737] = 4,
    ACTIONS(286), 1,
      sym_required_line_continuation,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(826), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(190), 2,
      sym__spaced_label_value_continuation,
      aux_sym__spaced_label_value_repeat2,
  [4753] = 4,
    ACTIONS(828), 1,
      sym_required_line_continuation,
    STATE(179), 1,
      aux_sym__healthcheck_raw_command_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(150), 3,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
  [4769] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(695), 5,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym_path_token2,
      anon_sym_DOLLAR2,
      sym__non_newline_whitespace,
  [4781] = 4,
    ACTIONS(733), 1,
      sym__heredoc_nl,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(831), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(225), 2,
      sym_heredoc_block,
      aux_sym_run_instruction_repeat2,
  [4797] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 5,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym_path_token2,
      anon_sym_DOLLAR2,
      sym__non_newline_whitespace,
  [4809] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 5,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym_path_token2,
      anon_sym_DOLLAR2,
      sym__non_newline_whitespace,
  [4821] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(835), 1,
      anon_sym_DQUOTE2,
    STATE(220), 1,
      aux_sym__shell_double_quoted_fragment_repeat1,
    ACTIONS(833), 4,
      sym_line_continuation,
      sym_required_line_continuation,
      aux_sym__shell_double_quoted_fragment_token1,
      aux_sym__shell_double_quoted_fragment_token2,
  [4837] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(695), 5,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_COLON,
      aux_sym__immediate_user_name_or_group_fragment_token1,
      anon_sym_DOLLAR2,
  [4849] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 5,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_COLON,
      aux_sym__immediate_user_name_or_group_fragment_token1,
      anon_sym_DOLLAR2,
  [4861] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 5,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_COLON,
      aux_sym__immediate_user_name_or_group_fragment_token1,
      anon_sym_DOLLAR2,
  [4873] = 5,
    ACTIONS(693), 1,
      anon_sym_DOLLAR,
    STATE(116), 1,
      sym_expansion,
    STATE(324), 1,
      sym_path,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(689), 2,
      aux_sym__healthcheck_raw_fragment_atom_token12,
      aux_sym_path_token1,
  [4891] = 4,
    ACTIONS(715), 1,
      anon_sym_DOLLAR,
    ACTIONS(717), 1,
      aux_sym_expose_port_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(139), 3,
      sym_expansion,
      sym_expose_port,
      aux_sym_expose_instruction_repeat1,
  [4907] = 4,
    ACTIONS(839), 1,
      sym_required_line_continuation,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(837), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(190), 2,
      sym__spaced_label_value_continuation,
      aux_sym__spaced_label_value_repeat2,
  [4923] = 3,
    ACTIONS(547), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 4,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_image_digest_token1,
  [4937] = 4,
    ACTIONS(842), 1,
      anon_sym_DQUOTE,
    STATE(224), 1,
      aux_sym_json_string_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(844), 3,
      aux_sym_json_string_token1,
      anon_sym_BQUOTE,
      sym_json_escape_sequence,
  [4953] = 4,
    ACTIONS(733), 1,
      sym__heredoc_nl,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(846), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(194), 2,
      sym_heredoc_block,
      aux_sym_run_instruction_repeat2,
  [4969] = 4,
    ACTIONS(733), 1,
      sym__heredoc_nl,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(848), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(225), 2,
      sym_heredoc_block,
      aux_sym_run_instruction_repeat2,
  [4985] = 5,
    ACTIONS(850), 1,
      aux_sym__env_key_token1,
    STATE(172), 1,
      sym__env_key,
    STATE(331), 1,
      sym__spaced_env_pair,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(163), 2,
      sym_env_pair,
      aux_sym_env_instruction_repeat1,
  [5003] = 4,
    ACTIONS(733), 1,
      sym__heredoc_nl,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(810), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(225), 2,
      sym_heredoc_block,
      aux_sym_run_instruction_repeat2,
  [5019] = 3,
    ACTIONS(551), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 4,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_image_digest_token1,
  [5033] = 6,
    ACTIONS(630), 1,
      anon_sym_DOLLAR,
    ACTIONS(632), 1,
      aux_sym_image_name_token1,
    STATE(87), 1,
      sym_expansion,
    STATE(118), 1,
      sym_image_name,
    STATE(281), 1,
      sym_image_spec,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [5053] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(852), 5,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym__label_key_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
  [5065] = 4,
    ACTIONS(854), 1,
      anon_sym_DQUOTE,
    STATE(200), 1,
      aux_sym_json_string_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(856), 3,
      aux_sym_json_string_token1,
      anon_sym_BQUOTE,
      sym_json_escape_sequence,
  [5081] = 4,
    ACTIONS(292), 1,
      sym_required_line_continuation,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(859), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(216), 2,
      sym__spaced_env_value_continuation,
      aux_sym__spaced_env_value_repeat2,
  [5097] = 4,
    ACTIONS(286), 1,
      sym_required_line_continuation,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(861), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(208), 2,
      sym__spaced_label_value_continuation,
      aux_sym__spaced_label_value_repeat2,
  [5113] = 4,
    ACTIONS(292), 1,
      sym_required_line_continuation,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(863), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(204), 2,
      sym__spaced_env_value_continuation,
      aux_sym__spaced_env_value_repeat2,
  [5129] = 4,
    ACTIONS(867), 1,
      sym_required_line_continuation,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(865), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(204), 2,
      sym__spaced_env_value_continuation,
      aux_sym__spaced_env_value_repeat2,
  [5145] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(870), 5,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym__label_key_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
  [5157] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(874), 1,
      anon_sym_DQUOTE2,
    STATE(210), 1,
      aux_sym__shell_double_quoted_fragment_repeat1,
    ACTIONS(872), 4,
      sym_line_continuation,
      sym_required_line_continuation,
      aux_sym__shell_double_quoted_fragment_token1,
      aux_sym__shell_double_quoted_fragment_token2,
  [5173] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(878), 1,
      anon_sym_SQUOTE2,
    STATE(211), 1,
      aux_sym__shell_single_quoted_fragment_repeat1,
    ACTIONS(876), 4,
      sym_line_continuation,
      sym_required_line_continuation,
      aux_sym__shell_double_quoted_fragment_token2,
      aux_sym__shell_single_quoted_fragment_token1,
  [5189] = 4,
    ACTIONS(286), 1,
      sym_required_line_continuation,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(880), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(190), 2,
      sym__spaced_label_value_continuation,
      aux_sym__spaced_label_value_repeat2,
  [5205] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(884), 1,
      anon_sym_DQUOTE2,
    STATE(184), 1,
      aux_sym__shell_double_quoted_fragment_repeat1,
    ACTIONS(882), 4,
      sym_line_continuation,
      sym_required_line_continuation,
      aux_sym__shell_double_quoted_fragment_token1,
      aux_sym__shell_double_quoted_fragment_token2,
  [5221] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(886), 1,
      anon_sym_DQUOTE2,
    STATE(220), 1,
      aux_sym__shell_double_quoted_fragment_repeat1,
    ACTIONS(833), 4,
      sym_line_continuation,
      sym_required_line_continuation,
      aux_sym__shell_double_quoted_fragment_token1,
      aux_sym__shell_double_quoted_fragment_token2,
  [5237] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(890), 1,
      anon_sym_SQUOTE2,
    STATE(223), 1,
      aux_sym__shell_single_quoted_fragment_repeat1,
    ACTIONS(888), 4,
      sym_line_continuation,
      sym_required_line_continuation,
      aux_sym__shell_double_quoted_fragment_token2,
      aux_sym__shell_single_quoted_fragment_token1,
  [5253] = 4,
    ACTIONS(557), 1,
      anon_sym_DOLLAR2,
    ACTIONS(892), 1,
      aux_sym_image_tag_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(101), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_image_tag_repeat1,
  [5269] = 4,
    ACTIONS(733), 1,
      sym__heredoc_nl,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(894), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(225), 2,
      sym_heredoc_block,
      aux_sym_run_instruction_repeat2,
  [5285] = 4,
    ACTIONS(733), 1,
      sym__heredoc_nl,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(896), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(225), 2,
      sym_heredoc_block,
      aux_sym_run_instruction_repeat2,
  [5301] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(898), 1,
      anon_sym_SQUOTE2,
    STATE(223), 1,
      aux_sym__shell_single_quoted_fragment_repeat1,
    ACTIONS(888), 4,
      sym_line_continuation,
      sym_required_line_continuation,
      aux_sym__shell_double_quoted_fragment_token2,
      aux_sym__shell_single_quoted_fragment_token1,
  [5317] = 4,
    ACTIONS(292), 1,
      sym_required_line_continuation,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(900), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(204), 2,
      sym__spaced_env_value_continuation,
      aux_sym__spaced_env_value_repeat2,
  [5333] = 4,
    ACTIONS(612), 1,
      anon_sym_DOLLAR2,
    ACTIONS(902), 1,
      aux_sym_image_digest_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    STATE(114), 3,
      sym__immediate_expansion,
      sym__imm_expansion,
      aux_sym_image_digest_repeat1,
  [5349] = 4,
    ACTIONS(640), 1,
      anon_sym_AT,
    STATE(282), 1,
      sym_image_digest,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(904), 3,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym_from_instruction_token2,
  [5365] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(906), 5,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym__label_key_token1,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
  [5377] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(911), 1,
      anon_sym_DQUOTE2,
    STATE(220), 1,
      aux_sym__shell_double_quoted_fragment_repeat1,
    ACTIONS(908), 4,
      sym_line_continuation,
      sym_required_line_continuation,
      aux_sym__shell_double_quoted_fragment_token1,
      aux_sym__shell_double_quoted_fragment_token2,
  [5393] = 6,
    ACTIONS(286), 1,
      sym_required_line_continuation,
    ACTIONS(913), 1,
      anon_sym_EQ,
    ACTIONS(915), 1,
      aux_sym__spaced_env_pair_token1,
    STATE(202), 1,
      sym__spaced_label_value_continuation,
    STATE(328), 1,
      sym__continued_spaced_label_value,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [5413] = 4,
    ACTIONS(733), 1,
      sym__heredoc_nl,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(800), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(225), 2,
      sym_heredoc_block,
      aux_sym_run_instruction_repeat2,
  [5429] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(920), 1,
      anon_sym_SQUOTE2,
    STATE(223), 1,
      aux_sym__shell_single_quoted_fragment_repeat1,
    ACTIONS(917), 4,
      sym_line_continuation,
      sym_required_line_continuation,
      aux_sym__shell_double_quoted_fragment_token2,
      aux_sym__shell_single_quoted_fragment_token1,
  [5445] = 4,
    ACTIONS(922), 1,
      anon_sym_DQUOTE,
    STATE(200), 1,
      aux_sym_json_string_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(924), 3,
      aux_sym_json_string_token1,
      anon_sym_BQUOTE,
      sym_json_escape_sequence,
  [5461] = 4,
    ACTIONS(928), 1,
      sym__heredoc_nl,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(926), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(225), 2,
      sym_heredoc_block,
      aux_sym_run_instruction_repeat2,
  [5477] = 4,
    ACTIONS(733), 1,
      sym__heredoc_nl,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(848), 2,
      sym__newline,
      ts_builtin_sym_end,
    STATE(181), 2,
      sym_heredoc_block,
      aux_sym_run_instruction_repeat2,
  [5493] = 4,
    ACTIONS(933), 1,
      sym__non_newline_whitespace,
    STATE(227), 1,
      aux_sym_volume_instruction_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(931), 2,
      sym__newline,
      ts_builtin_sym_end,
  [5508] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(936), 4,
      sym_heredoc_marker,
      aux_sym_path_token1,
      aux_sym_path_with_heredoc_token1,
      anon_sym_DOLLAR,
  [5519] = 5,
    ACTIONS(47), 1,
      ts_builtin_sym_end,
    ACTIONS(938), 1,
      sym__newline,
    STATE(4), 1,
      aux_sym_source_file_repeat1,
    STATE(238), 1,
      aux_sym_source_file_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [5536] = 5,
    ACTIONS(47), 1,
      ts_builtin_sym_end,
    ACTIONS(938), 1,
      sym__newline,
    STATE(4), 1,
      aux_sym_source_file_repeat1,
    STATE(239), 1,
      aux_sym_source_file_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [5553] = 4,
    ACTIONS(942), 1,
      sym__non_newline_whitespace,
    STATE(227), 1,
      aux_sym_volume_instruction_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(940), 2,
      sym__newline,
      ts_builtin_sym_end,
  [5568] = 3,
    ACTIONS(946), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(944), 3,
      sym__newline,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
  [5581] = 4,
    ACTIONS(950), 1,
      sym__non_newline_whitespace,
    STATE(233), 1,
      aux_sym_arg_instruction_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(948), 2,
      sym__newline,
      ts_builtin_sym_end,
  [5596] = 5,
    ACTIONS(953), 1,
      anon_sym_DOLLAR,
    ACTIONS(955), 1,
      aux_sym_image_alias_token1,
    STATE(133), 1,
      sym_expansion,
    STATE(337), 1,
      sym_image_alias,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [5613] = 4,
    ACTIONS(148), 1,
      sym_required_line_continuation,
    STATE(242), 1,
      aux_sym__healthcheck_raw_command_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(957), 2,
      sym__newline,
      ts_builtin_sym_end,
  [5628] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(533), 4,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
  [5639] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(695), 4,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR,
      aux_sym_expose_port_token1,
  [5650] = 5,
    ACTIONS(959), 1,
      ts_builtin_sym_end,
    ACTIONS(961), 1,
      sym__newline,
    STATE(6), 1,
      aux_sym_source_file_repeat1,
    STATE(238), 1,
      aux_sym_source_file_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [5667] = 5,
    ACTIONS(53), 1,
      ts_builtin_sym_end,
    ACTIONS(964), 1,
      sym__newline,
    STATE(5), 1,
      aux_sym_source_file_repeat1,
    STATE(238), 1,
      aux_sym_source_file_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [5684] = 4,
    ACTIONS(148), 1,
      sym_required_line_continuation,
    STATE(242), 1,
      aux_sym__healthcheck_raw_command_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(820), 2,
      sym__newline,
      ts_builtin_sym_end,
  [5699] = 5,
    ACTIONS(953), 1,
      anon_sym_DOLLAR,
    ACTIONS(955), 1,
      aux_sym_image_alias_token1,
    STATE(133), 1,
      sym_expansion,
    STATE(344), 1,
      sym_image_alias,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [5716] = 4,
    ACTIONS(966), 1,
      sym_required_line_continuation,
    STATE(242), 1,
      aux_sym__healthcheck_raw_command_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(150), 2,
      sym__newline,
      ts_builtin_sym_end,
  [5731] = 5,
    ACTIONS(51), 1,
      ts_builtin_sym_end,
    ACTIONS(969), 1,
      sym__newline,
    STATE(2), 1,
      aux_sym_source_file_repeat1,
    STATE(229), 1,
      aux_sym_source_file_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [5748] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 4,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR,
      aux_sym_expose_port_token1,
  [5759] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(816), 4,
      sym__newline,
      ts_builtin_sym_end,
      sym__env_trailing_single_quotes,
      aux_sym__env_key_token1,
  [5770] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(818), 4,
      sym__newline,
      ts_builtin_sym_end,
      sym__env_trailing_single_quotes,
      aux_sym__env_key_token1,
  [5781] = 4,
    ACTIONS(942), 1,
      sym__non_newline_whitespace,
    STATE(231), 1,
      aux_sym_volume_instruction_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(971), 2,
      sym__newline,
      ts_builtin_sym_end,
  [5796] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(695), 4,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym__stopsignal_value_token2,
      anon_sym_DOLLAR2,
  [5807] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 4,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym__stopsignal_value_token2,
      anon_sym_DOLLAR2,
  [5818] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 4,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym__stopsignal_value_token2,
      anon_sym_DOLLAR2,
  [5829] = 5,
    ACTIONS(973), 1,
      aux_sym__user_name_or_group_token1,
    ACTIONS(975), 1,
      anon_sym_DOLLAR,
    STATE(103), 1,
      sym_expansion,
    STATE(285), 1,
      sym__user_name_or_group,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [5846] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(695), 4,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_image_alias_token2,
  [5857] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 4,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_image_alias_token2,
  [5868] = 3,
    ACTIONS(979), 1,
      sym__env_trailing_single_quotes,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(977), 3,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym__env_key_token1,
  [5881] = 4,
    ACTIONS(983), 1,
      sym__non_newline_whitespace,
    STATE(233), 1,
      aux_sym_arg_instruction_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(981), 2,
      sym__newline,
      ts_builtin_sym_end,
  [5896] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 4,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR2,
      aux_sym_image_alias_token2,
  [5907] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(985), 4,
      sym__newline,
      ts_builtin_sym_end,
      anon_sym_DOLLAR,
      aux_sym_expose_port_token1,
  [5918] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(987), 4,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym__immediate_user_name_or_group_fragment_token1,
      anon_sym_DOLLAR2,
  [5929] = 4,
    ACTIONS(148), 1,
      sym_required_line_continuation,
    STATE(235), 1,
      aux_sym__healthcheck_raw_command_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(989), 2,
      sym__newline,
      ts_builtin_sym_end,
  [5944] = 4,
    ACTIONS(991), 1,
      anon_sym_LBRACE,
    ACTIONS(993), 1,
      sym_variable,
    STATE(249), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [5958] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(995), 3,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
  [5968] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(549), 3,
      aux_sym_path_token2,
      anon_sym_DOLLAR2,
      sym__non_newline_whitespace,
  [5978] = 4,
    ACTIONS(997), 1,
      sym__heredoc_line,
    ACTIONS(999), 1,
      sym_heredoc_end,
    STATE(278), 1,
      aux_sym_heredoc_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [5992] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1001), 3,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
  [6002] = 4,
    ACTIONS(1003), 1,
      anon_sym_COMMA2,
    ACTIONS(1005), 1,
      anon_sym_RBRACK,
    STATE(305), 1,
      aux_sym_json_string_array_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6016] = 4,
    ACTIONS(997), 1,
      sym__heredoc_line,
    ACTIONS(1007), 1,
      sym_heredoc_end,
    STATE(263), 1,
      aux_sym_heredoc_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6030] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(870), 3,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym__env_key_token1,
  [6040] = 4,
    ACTIONS(1009), 1,
      anon_sym_LBRACE,
    ACTIONS(1011), 1,
      sym_variable,
    STATE(134), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6054] = 4,
    ACTIONS(1013), 1,
      anon_sym_LBRACE,
    ACTIONS(1015), 1,
      sym_variable,
    STATE(132), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6068] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(906), 3,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym__env_key_token1,
  [6078] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1017), 3,
      sym__newline,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
  [6088] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(906), 3,
      sym_required_line_continuation,
      anon_sym_EQ,
      aux_sym__spaced_env_pair_token1,
  [6098] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(870), 3,
      sym__newline,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
  [6108] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(816), 3,
      sym__newline,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
  [6118] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(818), 3,
      sym_required_line_continuation,
      anon_sym_EQ,
      aux_sym__spaced_env_pair_token1,
  [6128] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(906), 3,
      sym__newline,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
  [6138] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(818), 3,
      sym__newline,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
  [6148] = 4,
    ACTIONS(1019), 1,
      sym__heredoc_line,
    ACTIONS(1022), 1,
      sym_heredoc_end,
    STATE(278), 1,
      aux_sym_heredoc_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6162] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1024), 3,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym_from_instruction_token2,
  [6172] = 4,
    ACTIONS(210), 1,
      aux_sym_image_name_token1,
    ACTIONS(214), 1,
      anon_sym_DOLLAR,
    ACTIONS(1026), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6186] = 3,
    ACTIONS(1030), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1028), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6198] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1032), 3,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym_from_instruction_token2,
  [6208] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(870), 3,
      sym_required_line_continuation,
      anon_sym_EQ,
      aux_sym__spaced_env_pair_token1,
  [6218] = 4,
    ACTIONS(1013), 1,
      anon_sym_LBRACE,
    ACTIONS(1034), 1,
      sym_variable,
    STATE(131), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6232] = 3,
    ACTIONS(1038), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1036), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6244] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(816), 3,
      sym_required_line_continuation,
      anon_sym_EQ,
      aux_sym__spaced_env_pair_token1,
  [6254] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1040), 3,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
  [6264] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1042), 3,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
  [6274] = 4,
    ACTIONS(1044), 1,
      anon_sym_DQUOTE,
    ACTIONS(1046), 1,
      anon_sym_RBRACK,
    STATE(265), 1,
      sym_json_string,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6288] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1048), 3,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym__env_key_token1,
  [6298] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(948), 3,
      sym__newline,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
  [6308] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(931), 3,
      sym__newline,
      ts_builtin_sym_end,
      sym__non_newline_whitespace,
  [6318] = 4,
    ACTIONS(1050), 1,
      anon_sym_LBRACE,
    ACTIONS(1052), 1,
      sym_variable,
    STATE(237), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6332] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(695), 3,
      aux_sym_path_token2,
      anon_sym_DOLLAR2,
      sym__non_newline_whitespace,
  [6342] = 4,
    ACTIONS(1044), 1,
      anon_sym_DQUOTE,
    ACTIONS(1054), 1,
      anon_sym_RBRACK,
    STATE(298), 1,
      sym_json_string,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6356] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1056), 3,
      sym__newline,
      sym__heredoc_nl,
      ts_builtin_sym_end,
  [6366] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(545), 3,
      aux_sym_path_token2,
      anon_sym_DOLLAR2,
      sym__non_newline_whitespace,
  [6376] = 4,
    ACTIONS(1003), 1,
      anon_sym_COMMA2,
    ACTIONS(1058), 1,
      anon_sym_RBRACK,
    STATE(300), 1,
      aux_sym_json_string_array_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6390] = 3,
    ACTIONS(1062), 1,
      aux_sym_from_instruction_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1060), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6402] = 4,
    ACTIONS(1003), 1,
      anon_sym_COMMA2,
    ACTIONS(1064), 1,
      anon_sym_RBRACK,
    STATE(308), 1,
      aux_sym_json_string_array_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6416] = 4,
    ACTIONS(1066), 1,
      anon_sym_LBRACE,
    ACTIONS(1068), 1,
      sym_variable,
    STATE(297), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6430] = 4,
    ACTIONS(1066), 1,
      anon_sym_LBRACE,
    ACTIONS(1070), 1,
      sym_variable,
    STATE(294), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6444] = 4,
    ACTIONS(1072), 1,
      anon_sym_LBRACE,
    ACTIONS(1074), 1,
      sym_variable,
    STATE(182), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6458] = 4,
    ACTIONS(1072), 1,
      anon_sym_LBRACE,
    ACTIONS(1076), 1,
      sym_variable,
    STATE(180), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6472] = 4,
    ACTIONS(1003), 1,
      anon_sym_COMMA2,
    ACTIONS(1078), 1,
      anon_sym_RBRACK,
    STATE(308), 1,
      aux_sym_json_string_array_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6486] = 4,
    ACTIONS(1080), 1,
      anon_sym_LBRACE,
    ACTIONS(1082), 1,
      sym_variable,
    STATE(186), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6500] = 4,
    ACTIONS(1080), 1,
      anon_sym_LBRACE,
    ACTIONS(1084), 1,
      sym_variable,
    STATE(185), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6514] = 4,
    ACTIONS(1086), 1,
      anon_sym_COMMA2,
    ACTIONS(1089), 1,
      anon_sym_RBRACK,
    STATE(308), 1,
      aux_sym_json_string_array_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6528] = 4,
    ACTIONS(991), 1,
      anon_sym_LBRACE,
    ACTIONS(1091), 1,
      sym_variable,
    STATE(248), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6542] = 4,
    ACTIONS(1093), 1,
      anon_sym_LBRACE,
    ACTIONS(1095), 1,
      sym_variable,
    STATE(149), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6556] = 4,
    ACTIONS(1097), 1,
      anon_sym_LBRACE,
    ACTIONS(1099), 1,
      sym_variable,
    STATE(148), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6570] = 4,
    ACTIONS(1101), 1,
      anon_sym_LBRACE,
    ACTIONS(1103), 1,
      sym_variable,
    STATE(191), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6584] = 4,
    ACTIONS(1105), 1,
      anon_sym_LBRACE,
    ACTIONS(1107), 1,
      sym_variable,
    STATE(252), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6598] = 4,
    ACTIONS(1109), 1,
      anon_sym_LBRACE,
    ACTIONS(1111), 1,
      sym_variable,
    STATE(110), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6612] = 4,
    ACTIONS(1113), 1,
      anon_sym_LBRACE,
    ACTIONS(1115), 1,
      sym_variable,
    STATE(99), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6626] = 4,
    ACTIONS(1117), 1,
      anon_sym_LBRACE,
    ACTIONS(1119), 1,
      sym_variable,
    STATE(111), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6640] = 4,
    ACTIONS(1121), 1,
      anon_sym_LBRACE,
    ACTIONS(1123), 1,
      sym_variable,
    STATE(138), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6654] = 4,
    ACTIONS(1097), 1,
      anon_sym_LBRACE,
    ACTIONS(1125), 1,
      sym_variable,
    STATE(155), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6668] = 4,
    ACTIONS(1105), 1,
      anon_sym_LBRACE,
    ACTIONS(1127), 1,
      sym_variable,
    STATE(256), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6682] = 4,
    ACTIONS(1129), 1,
      anon_sym_LBRACE,
    ACTIONS(1131), 1,
      sym_variable,
    STATE(140), 1,
      sym__expansion_body,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6696] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1133), 3,
      sym__newline,
      ts_builtin_sym_end,
      aux_sym__env_key_token1,
  [6706] = 3,
    ACTIONS(1135), 1,
      aux_sym_param_token1,
    ACTIONS(1137), 1,
      anon_sym_mount,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6717] = 3,
    ACTIONS(1044), 1,
      anon_sym_DQUOTE,
    STATE(350), 1,
      sym_json_string,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6728] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1139), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6737] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1141), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6746] = 3,
    ACTIONS(1143), 1,
      aux_sym_arg_pair_token1,
    STATE(291), 1,
      sym_arg_pair,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6757] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1145), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6766] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1147), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6775] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1149), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6784] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1001), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6793] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1151), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6802] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1153), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6811] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1155), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6820] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1157), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6829] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1159), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6838] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(729), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6847] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1161), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6856] = 3,
    ACTIONS(102), 1,
      anon_sym_LBRACK,
    STATE(340), 1,
      sym_json_string_array,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6867] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(218), 2,
      anon_sym_DOLLAR,
      aux_sym_image_name_token1,
  [6876] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1163), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6885] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1165), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6894] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1056), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6903] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(757), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6912] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1167), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6921] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1169), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6930] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1171), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6939] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1022), 2,
      sym__heredoc_line,
      sym_heredoc_end,
  [6948] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1173), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6957] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1175), 2,
      anon_sym_COMMA2,
      anon_sym_RBRACK,
  [6966] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1089), 2,
      anon_sym_COMMA2,
      anon_sym_RBRACK,
  [6975] = 3,
    ACTIONS(1177), 1,
      aux_sym_mount_param_param_token1,
    STATE(37), 1,
      sym_mount_param_param,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [6986] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1179), 2,
      sym__newline,
      ts_builtin_sym_end,
  [6995] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1181), 2,
      sym__newline,
      ts_builtin_sym_end,
  [7004] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(959), 2,
      sym__newline,
      ts_builtin_sym_end,
  [7013] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1183), 2,
      sym__newline,
      ts_builtin_sym_end,
  [7022] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(971), 2,
      sym__newline,
      ts_builtin_sym_end,
  [7031] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1040), 2,
      sym__newline,
      ts_builtin_sym_end,
  [7040] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1185), 2,
      anon_sym_COMMA2,
      anon_sym_RBRACK,
  [7049] = 3,
    ACTIONS(983), 1,
      sym__non_newline_whitespace,
    STATE(255), 1,
      aux_sym_arg_instruction_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7060] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
    ACTIONS(1187), 2,
      sym__newline,
      ts_builtin_sym_end,
  [7069] = 3,
    ACTIONS(1177), 1,
      aux_sym_mount_param_param_token1,
    STATE(47), 1,
      sym_mount_param_param,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7080] = 2,
    ACTIONS(1189), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7088] = 2,
    ACTIONS(1191), 1,
      aux_sym_maintainer_instruction_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7096] = 2,
    ACTIONS(1193), 1,
      ts_builtin_sym_end,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7104] = 2,
    ACTIONS(1195), 1,
      aux_sym__healthcheck_raw_fragment_atom_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7112] = 2,
    ACTIONS(1197), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7120] = 2,
    ACTIONS(870), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7128] = 2,
    ACTIONS(1199), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7136] = 2,
    ACTIONS(1201), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7144] = 2,
    ACTIONS(1203), 1,
      aux_sym_param_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7152] = 2,
    ACTIONS(1205), 1,
      aux_sym_param_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7160] = 2,
    ACTIONS(1207), 1,
      aux_sym_param_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7168] = 2,
    ACTIONS(816), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7176] = 2,
    ACTIONS(913), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7184] = 2,
    ACTIONS(1209), 1,
      aux_sym_maintainer_instruction_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7192] = 2,
    ACTIONS(533), 1,
      sym__non_newline_whitespace,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7200] = 2,
    ACTIONS(1211), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7208] = 2,
    ACTIONS(1213), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7216] = 2,
    ACTIONS(1215), 1,
      aux_sym_param_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7224] = 2,
    ACTIONS(1217), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7232] = 2,
    ACTIONS(1219), 1,
      anon_sym_LF,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7240] = 2,
    ACTIONS(818), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7248] = 2,
    ACTIONS(1221), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7256] = 2,
    ACTIONS(731), 1,
      sym__non_newline_whitespace,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7264] = 2,
    ACTIONS(1223), 1,
      aux_sym__healthcheck_raw_fragment_atom_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7272] = 2,
    ACTIONS(1225), 1,
      aux_sym_param_token2,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7280] = 2,
    ACTIONS(1227), 1,
      aux_sym_mount_param_param_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7288] = 2,
    ACTIONS(1229), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7296] = 2,
    ACTIONS(1231), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7304] = 2,
    ACTIONS(1233), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7312] = 2,
    ACTIONS(1235), 1,
      aux_sym_param_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7320] = 2,
    ACTIONS(1237), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7328] = 2,
    ACTIONS(812), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7336] = 2,
    ACTIONS(1239), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7344] = 2,
    ACTIONS(906), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7352] = 2,
    ACTIONS(1241), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7360] = 2,
    ACTIONS(1243), 1,
      aux_sym_param_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7368] = 2,
    ACTIONS(1245), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7376] = 2,
    ACTIONS(1247), 1,
      aux_sym__healthcheck_raw_fragment_atom_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7384] = 2,
    ACTIONS(1249), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7392] = 2,
    ACTIONS(1251), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7400] = 2,
    ACTIONS(1253), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7408] = 2,
    ACTIONS(1255), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7416] = 2,
    ACTIONS(1257), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7424] = 2,
    ACTIONS(1259), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7432] = 2,
    ACTIONS(1261), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7440] = 2,
    ACTIONS(1263), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7448] = 2,
    ACTIONS(1265), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7456] = 2,
    ACTIONS(1267), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7464] = 2,
    ACTIONS(1269), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7472] = 2,
    ACTIONS(1271), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7480] = 2,
    ACTIONS(1273), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7488] = 2,
    ACTIONS(1275), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7496] = 2,
    ACTIONS(1277), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7504] = 2,
    ACTIONS(1279), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7512] = 2,
    ACTIONS(1281), 1,
      aux_sym__expansion_body_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
  [7520] = 2,
    ACTIONS(1283), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym_line_continuation,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(2)] = 0,
  [SMALL_STATE(3)] = 93,
  [SMALL_STATE(4)] = 186,
  [SMALL_STATE(5)] = 279,
  [SMALL_STATE(6)] = 372,
  [SMALL_STATE(7)] = 462,
  [SMALL_STATE(8)] = 546,
  [SMALL_STATE(9)] = 604,
  [SMALL_STATE(10)] = 655,
  [SMALL_STATE(11)] = 705,
  [SMALL_STATE(12)] = 755,
  [SMALL_STATE(13)] = 788,
  [SMALL_STATE(14)] = 830,
  [SMALL_STATE(15)] = 872,
  [SMALL_STATE(16)] = 914,
  [SMALL_STATE(17)] = 956,
  [SMALL_STATE(18)] = 993,
  [SMALL_STATE(19)] = 1024,
  [SMALL_STATE(20)] = 1065,
  [SMALL_STATE(21)] = 1102,
  [SMALL_STATE(22)] = 1139,
  [SMALL_STATE(23)] = 1175,
  [SMALL_STATE(24)] = 1215,
  [SMALL_STATE(25)] = 1251,
  [SMALL_STATE(26)] = 1287,
  [SMALL_STATE(27)] = 1323,
  [SMALL_STATE(28)] = 1359,
  [SMALL_STATE(29)] = 1395,
  [SMALL_STATE(30)] = 1423,
  [SMALL_STATE(31)] = 1448,
  [SMALL_STATE(32)] = 1485,
  [SMALL_STATE(33)] = 1522,
  [SMALL_STATE(34)] = 1559,
  [SMALL_STATE(35)] = 1596,
  [SMALL_STATE(36)] = 1625,
  [SMALL_STATE(37)] = 1657,
  [SMALL_STATE(38)] = 1685,
  [SMALL_STATE(39)] = 1713,
  [SMALL_STATE(40)] = 1739,
  [SMALL_STATE(41)] = 1767,
  [SMALL_STATE(42)] = 1799,
  [SMALL_STATE(43)] = 1831,
  [SMALL_STATE(44)] = 1854,
  [SMALL_STATE(45)] = 1877,
  [SMALL_STATE(46)] = 1900,
  [SMALL_STATE(47)] = 1923,
  [SMALL_STATE(48)] = 1946,
  [SMALL_STATE(49)] = 1971,
  [SMALL_STATE(50)] = 1994,
  [SMALL_STATE(51)] = 2017,
  [SMALL_STATE(52)] = 2043,
  [SMALL_STATE(53)] = 2065,
  [SMALL_STATE(54)] = 2091,
  [SMALL_STATE(55)] = 2117,
  [SMALL_STATE(56)] = 2139,
  [SMALL_STATE(57)] = 2165,
  [SMALL_STATE(58)] = 2187,
  [SMALL_STATE(59)] = 2209,
  [SMALL_STATE(60)] = 2231,
  [SMALL_STATE(61)] = 2253,
  [SMALL_STATE(62)] = 2275,
  [SMALL_STATE(63)] = 2304,
  [SMALL_STATE(64)] = 2343,
  [SMALL_STATE(65)] = 2382,
  [SMALL_STATE(66)] = 2409,
  [SMALL_STATE(67)] = 2438,
  [SMALL_STATE(68)] = 2477,
  [SMALL_STATE(69)] = 2516,
  [SMALL_STATE(70)] = 2543,
  [SMALL_STATE(71)] = 2568,
  [SMALL_STATE(72)] = 2592,
  [SMALL_STATE(73)] = 2617,
  [SMALL_STATE(74)] = 2638,
  [SMALL_STATE(75)] = 2663,
  [SMALL_STATE(76)] = 2688,
  [SMALL_STATE(77)] = 2713,
  [SMALL_STATE(78)] = 2738,
  [SMALL_STATE(79)] = 2763,
  [SMALL_STATE(80)] = 2790,
  [SMALL_STATE(81)] = 2815,
  [SMALL_STATE(82)] = 2840,
  [SMALL_STATE(83)] = 2867,
  [SMALL_STATE(84)] = 2888,
  [SMALL_STATE(85)] = 2913,
  [SMALL_STATE(86)] = 2936,
  [SMALL_STATE(87)] = 2959,
  [SMALL_STATE(88)] = 2984,
  [SMALL_STATE(89)] = 3009,
  [SMALL_STATE(90)] = 3034,
  [SMALL_STATE(91)] = 3059,
  [SMALL_STATE(92)] = 3084,
  [SMALL_STATE(93)] = 3109,
  [SMALL_STATE(94)] = 3134,
  [SMALL_STATE(95)] = 3156,
  [SMALL_STATE(96)] = 3178,
  [SMALL_STATE(97)] = 3204,
  [SMALL_STATE(98)] = 3226,
  [SMALL_STATE(99)] = 3248,
  [SMALL_STATE(100)] = 3266,
  [SMALL_STATE(101)] = 3284,
  [SMALL_STATE(102)] = 3308,
  [SMALL_STATE(103)] = 3332,
  [SMALL_STATE(104)] = 3354,
  [SMALL_STATE(105)] = 3376,
  [SMALL_STATE(106)] = 3397,
  [SMALL_STATE(107)] = 3420,
  [SMALL_STATE(108)] = 3443,
  [SMALL_STATE(109)] = 3464,
  [SMALL_STATE(110)] = 3485,
  [SMALL_STATE(111)] = 3500,
  [SMALL_STATE(112)] = 3515,
  [SMALL_STATE(113)] = 3530,
  [SMALL_STATE(114)] = 3545,
  [SMALL_STATE(115)] = 3568,
  [SMALL_STATE(116)] = 3591,
  [SMALL_STATE(117)] = 3612,
  [SMALL_STATE(118)] = 3638,
  [SMALL_STATE(119)] = 3660,
  [SMALL_STATE(120)] = 3680,
  [SMALL_STATE(121)] = 3704,
  [SMALL_STATE(122)] = 3728,
  [SMALL_STATE(123)] = 3744,
  [SMALL_STATE(124)] = 3764,
  [SMALL_STATE(125)] = 3788,
  [SMALL_STATE(126)] = 3812,
  [SMALL_STATE(127)] = 3836,
  [SMALL_STATE(128)] = 3856,
  [SMALL_STATE(129)] = 3876,
  [SMALL_STATE(130)] = 3896,
  [SMALL_STATE(131)] = 3922,
  [SMALL_STATE(132)] = 3938,
  [SMALL_STATE(133)] = 3954,
  [SMALL_STATE(134)] = 3974,
  [SMALL_STATE(135)] = 3990,
  [SMALL_STATE(136)] = 4010,
  [SMALL_STATE(137)] = 4028,
  [SMALL_STATE(138)] = 4044,
  [SMALL_STATE(139)] = 4060,
  [SMALL_STATE(140)] = 4080,
  [SMALL_STATE(141)] = 4094,
  [SMALL_STATE(142)] = 4110,
  [SMALL_STATE(143)] = 4124,
  [SMALL_STATE(144)] = 4143,
  [SMALL_STATE(145)] = 4162,
  [SMALL_STATE(146)] = 4181,
  [SMALL_STATE(147)] = 4196,
  [SMALL_STATE(148)] = 4211,
  [SMALL_STATE(149)] = 4224,
  [SMALL_STATE(150)] = 4239,
  [SMALL_STATE(151)] = 4260,
  [SMALL_STATE(152)] = 4273,
  [SMALL_STATE(153)] = 4292,
  [SMALL_STATE(154)] = 4307,
  [SMALL_STATE(155)] = 4326,
  [SMALL_STATE(156)] = 4339,
  [SMALL_STATE(157)] = 4358,
  [SMALL_STATE(158)] = 4377,
  [SMALL_STATE(159)] = 4396,
  [SMALL_STATE(160)] = 4415,
  [SMALL_STATE(161)] = 4434,
  [SMALL_STATE(162)] = 4453,
  [SMALL_STATE(163)] = 4472,
  [SMALL_STATE(164)] = 4491,
  [SMALL_STATE(165)] = 4510,
  [SMALL_STATE(166)] = 4529,
  [SMALL_STATE(167)] = 4548,
  [SMALL_STATE(168)] = 4567,
  [SMALL_STATE(169)] = 4586,
  [SMALL_STATE(170)] = 4605,
  [SMALL_STATE(171)] = 4624,
  [SMALL_STATE(172)] = 4643,
  [SMALL_STATE(173)] = 4663,
  [SMALL_STATE(174)] = 4681,
  [SMALL_STATE(175)] = 4693,
  [SMALL_STATE(176)] = 4705,
  [SMALL_STATE(177)] = 4721,
  [SMALL_STATE(178)] = 4737,
  [SMALL_STATE(179)] = 4753,
  [SMALL_STATE(180)] = 4769,
  [SMALL_STATE(181)] = 4781,
  [SMALL_STATE(182)] = 4797,
  [SMALL_STATE(183)] = 4809,
  [SMALL_STATE(184)] = 4821,
  [SMALL_STATE(185)] = 4837,
  [SMALL_STATE(186)] = 4849,
  [SMALL_STATE(187)] = 4861,
  [SMALL_STATE(188)] = 4873,
  [SMALL_STATE(189)] = 4891,
  [SMALL_STATE(190)] = 4907,
  [SMALL_STATE(191)] = 4923,
  [SMALL_STATE(192)] = 4937,
  [SMALL_STATE(193)] = 4953,
  [SMALL_STATE(194)] = 4969,
  [SMALL_STATE(195)] = 4985,
  [SMALL_STATE(196)] = 5003,
  [SMALL_STATE(197)] = 5019,
  [SMALL_STATE(198)] = 5033,
  [SMALL_STATE(199)] = 5053,
  [SMALL_STATE(200)] = 5065,
  [SMALL_STATE(201)] = 5081,
  [SMALL_STATE(202)] = 5097,
  [SMALL_STATE(203)] = 5113,
  [SMALL_STATE(204)] = 5129,
  [SMALL_STATE(205)] = 5145,
  [SMALL_STATE(206)] = 5157,
  [SMALL_STATE(207)] = 5173,
  [SMALL_STATE(208)] = 5189,
  [SMALL_STATE(209)] = 5205,
  [SMALL_STATE(210)] = 5221,
  [SMALL_STATE(211)] = 5237,
  [SMALL_STATE(212)] = 5253,
  [SMALL_STATE(213)] = 5269,
  [SMALL_STATE(214)] = 5285,
  [SMALL_STATE(215)] = 5301,
  [SMALL_STATE(216)] = 5317,
  [SMALL_STATE(217)] = 5333,
  [SMALL_STATE(218)] = 5349,
  [SMALL_STATE(219)] = 5365,
  [SMALL_STATE(220)] = 5377,
  [SMALL_STATE(221)] = 5393,
  [SMALL_STATE(222)] = 5413,
  [SMALL_STATE(223)] = 5429,
  [SMALL_STATE(224)] = 5445,
  [SMALL_STATE(225)] = 5461,
  [SMALL_STATE(226)] = 5477,
  [SMALL_STATE(227)] = 5493,
  [SMALL_STATE(228)] = 5508,
  [SMALL_STATE(229)] = 5519,
  [SMALL_STATE(230)] = 5536,
  [SMALL_STATE(231)] = 5553,
  [SMALL_STATE(232)] = 5568,
  [SMALL_STATE(233)] = 5581,
  [SMALL_STATE(234)] = 5596,
  [SMALL_STATE(235)] = 5613,
  [SMALL_STATE(236)] = 5628,
  [SMALL_STATE(237)] = 5639,
  [SMALL_STATE(238)] = 5650,
  [SMALL_STATE(239)] = 5667,
  [SMALL_STATE(240)] = 5684,
  [SMALL_STATE(241)] = 5699,
  [SMALL_STATE(242)] = 5716,
  [SMALL_STATE(243)] = 5731,
  [SMALL_STATE(244)] = 5748,
  [SMALL_STATE(245)] = 5759,
  [SMALL_STATE(246)] = 5770,
  [SMALL_STATE(247)] = 5781,
  [SMALL_STATE(248)] = 5796,
  [SMALL_STATE(249)] = 5807,
  [SMALL_STATE(250)] = 5818,
  [SMALL_STATE(251)] = 5829,
  [SMALL_STATE(252)] = 5846,
  [SMALL_STATE(253)] = 5857,
  [SMALL_STATE(254)] = 5868,
  [SMALL_STATE(255)] = 5881,
  [SMALL_STATE(256)] = 5896,
  [SMALL_STATE(257)] = 5907,
  [SMALL_STATE(258)] = 5918,
  [SMALL_STATE(259)] = 5929,
  [SMALL_STATE(260)] = 5944,
  [SMALL_STATE(261)] = 5958,
  [SMALL_STATE(262)] = 5968,
  [SMALL_STATE(263)] = 5978,
  [SMALL_STATE(264)] = 5992,
  [SMALL_STATE(265)] = 6002,
  [SMALL_STATE(266)] = 6016,
  [SMALL_STATE(267)] = 6030,
  [SMALL_STATE(268)] = 6040,
  [SMALL_STATE(269)] = 6054,
  [SMALL_STATE(270)] = 6068,
  [SMALL_STATE(271)] = 6078,
  [SMALL_STATE(272)] = 6088,
  [SMALL_STATE(273)] = 6098,
  [SMALL_STATE(274)] = 6108,
  [SMALL_STATE(275)] = 6118,
  [SMALL_STATE(276)] = 6128,
  [SMALL_STATE(277)] = 6138,
  [SMALL_STATE(278)] = 6148,
  [SMALL_STATE(279)] = 6162,
  [SMALL_STATE(280)] = 6172,
  [SMALL_STATE(281)] = 6186,
  [SMALL_STATE(282)] = 6198,
  [SMALL_STATE(283)] = 6208,
  [SMALL_STATE(284)] = 6218,
  [SMALL_STATE(285)] = 6232,
  [SMALL_STATE(286)] = 6244,
  [SMALL_STATE(287)] = 6254,
  [SMALL_STATE(288)] = 6264,
  [SMALL_STATE(289)] = 6274,
  [SMALL_STATE(290)] = 6288,
  [SMALL_STATE(291)] = 6298,
  [SMALL_STATE(292)] = 6308,
  [SMALL_STATE(293)] = 6318,
  [SMALL_STATE(294)] = 6332,
  [SMALL_STATE(295)] = 6342,
  [SMALL_STATE(296)] = 6356,
  [SMALL_STATE(297)] = 6366,
  [SMALL_STATE(298)] = 6376,
  [SMALL_STATE(299)] = 6390,
  [SMALL_STATE(300)] = 6402,
  [SMALL_STATE(301)] = 6416,
  [SMALL_STATE(302)] = 6430,
  [SMALL_STATE(303)] = 6444,
  [SMALL_STATE(304)] = 6458,
  [SMALL_STATE(305)] = 6472,
  [SMALL_STATE(306)] = 6486,
  [SMALL_STATE(307)] = 6500,
  [SMALL_STATE(308)] = 6514,
  [SMALL_STATE(309)] = 6528,
  [SMALL_STATE(310)] = 6542,
  [SMALL_STATE(311)] = 6556,
  [SMALL_STATE(312)] = 6570,
  [SMALL_STATE(313)] = 6584,
  [SMALL_STATE(314)] = 6598,
  [SMALL_STATE(315)] = 6612,
  [SMALL_STATE(316)] = 6626,
  [SMALL_STATE(317)] = 6640,
  [SMALL_STATE(318)] = 6654,
  [SMALL_STATE(319)] = 6668,
  [SMALL_STATE(320)] = 6682,
  [SMALL_STATE(321)] = 6696,
  [SMALL_STATE(322)] = 6706,
  [SMALL_STATE(323)] = 6717,
  [SMALL_STATE(324)] = 6728,
  [SMALL_STATE(325)] = 6737,
  [SMALL_STATE(326)] = 6746,
  [SMALL_STATE(327)] = 6757,
  [SMALL_STATE(328)] = 6766,
  [SMALL_STATE(329)] = 6775,
  [SMALL_STATE(330)] = 6784,
  [SMALL_STATE(331)] = 6793,
  [SMALL_STATE(332)] = 6802,
  [SMALL_STATE(333)] = 6811,
  [SMALL_STATE(334)] = 6820,
  [SMALL_STATE(335)] = 6829,
  [SMALL_STATE(336)] = 6838,
  [SMALL_STATE(337)] = 6847,
  [SMALL_STATE(338)] = 6856,
  [SMALL_STATE(339)] = 6867,
  [SMALL_STATE(340)] = 6876,
  [SMALL_STATE(341)] = 6885,
  [SMALL_STATE(342)] = 6894,
  [SMALL_STATE(343)] = 6903,
  [SMALL_STATE(344)] = 6912,
  [SMALL_STATE(345)] = 6921,
  [SMALL_STATE(346)] = 6930,
  [SMALL_STATE(347)] = 6939,
  [SMALL_STATE(348)] = 6948,
  [SMALL_STATE(349)] = 6957,
  [SMALL_STATE(350)] = 6966,
  [SMALL_STATE(351)] = 6975,
  [SMALL_STATE(352)] = 6986,
  [SMALL_STATE(353)] = 6995,
  [SMALL_STATE(354)] = 7004,
  [SMALL_STATE(355)] = 7013,
  [SMALL_STATE(356)] = 7022,
  [SMALL_STATE(357)] = 7031,
  [SMALL_STATE(358)] = 7040,
  [SMALL_STATE(359)] = 7049,
  [SMALL_STATE(360)] = 7060,
  [SMALL_STATE(361)] = 7069,
  [SMALL_STATE(362)] = 7080,
  [SMALL_STATE(363)] = 7088,
  [SMALL_STATE(364)] = 7096,
  [SMALL_STATE(365)] = 7104,
  [SMALL_STATE(366)] = 7112,
  [SMALL_STATE(367)] = 7120,
  [SMALL_STATE(368)] = 7128,
  [SMALL_STATE(369)] = 7136,
  [SMALL_STATE(370)] = 7144,
  [SMALL_STATE(371)] = 7152,
  [SMALL_STATE(372)] = 7160,
  [SMALL_STATE(373)] = 7168,
  [SMALL_STATE(374)] = 7176,
  [SMALL_STATE(375)] = 7184,
  [SMALL_STATE(376)] = 7192,
  [SMALL_STATE(377)] = 7200,
  [SMALL_STATE(378)] = 7208,
  [SMALL_STATE(379)] = 7216,
  [SMALL_STATE(380)] = 7224,
  [SMALL_STATE(381)] = 7232,
  [SMALL_STATE(382)] = 7240,
  [SMALL_STATE(383)] = 7248,
  [SMALL_STATE(384)] = 7256,
  [SMALL_STATE(385)] = 7264,
  [SMALL_STATE(386)] = 7272,
  [SMALL_STATE(387)] = 7280,
  [SMALL_STATE(388)] = 7288,
  [SMALL_STATE(389)] = 7296,
  [SMALL_STATE(390)] = 7304,
  [SMALL_STATE(391)] = 7312,
  [SMALL_STATE(392)] = 7320,
  [SMALL_STATE(393)] = 7328,
  [SMALL_STATE(394)] = 7336,
  [SMALL_STATE(395)] = 7344,
  [SMALL_STATE(396)] = 7352,
  [SMALL_STATE(397)] = 7360,
  [SMALL_STATE(398)] = 7368,
  [SMALL_STATE(399)] = 7376,
  [SMALL_STATE(400)] = 7384,
  [SMALL_STATE(401)] = 7392,
  [SMALL_STATE(402)] = 7400,
  [SMALL_STATE(403)] = 7408,
  [SMALL_STATE(404)] = 7416,
  [SMALL_STATE(405)] = 7424,
  [SMALL_STATE(406)] = 7432,
  [SMALL_STATE(407)] = 7440,
  [SMALL_STATE(408)] = 7448,
  [SMALL_STATE(409)] = 7456,
  [SMALL_STATE(410)] = 7464,
  [SMALL_STATE(411)] = 7472,
  [SMALL_STATE(412)] = 7480,
  [SMALL_STATE(413)] = 7488,
  [SMALL_STATE(414)] = 7496,
  [SMALL_STATE(415)] = 7504,
  [SMALL_STATE(416)] = 7512,
  [SMALL_STATE(417)] = 7520,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 0, 0, 0),
  [7] = {.entry = {.count = 1, .reusable = true}}, SHIFT(117),
  [9] = {.entry = {.count = 1, .reusable = true}}, SHIFT(10),
  [11] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [13] = {.entry = {.count = 1, .reusable = true}}, SHIFT(96),
  [15] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [17] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [19] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [21] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [23] = {.entry = {.count = 1, .reusable = true}}, SHIFT(13),
  [25] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [27] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [29] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [31] = {.entry = {.count = 1, .reusable = true}}, SHIFT(359),
  [33] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7),
  [35] = {.entry = {.count = 1, .reusable = true}}, SHIFT(150),
  [37] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8),
  [39] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [41] = {.entry = {.count = 1, .reusable = true}}, SHIFT(375),
  [43] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [45] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3),
  [47] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 2, 0, 0),
  [49] = {.entry = {.count = 1, .reusable = true}}, SHIFT(12),
  [51] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 1, 0, 0),
  [53] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 3, 0, 0),
  [55] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 4, 0, 0),
  [57] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_healthcheck_instruction, 1, 0, 0),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(16),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(334),
  [63] = {.entry = {.count = 1, .reusable = true}}, SHIFT(22),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(22),
  [67] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [69] = {.entry = {.count = 1, .reusable = false}}, SHIFT(206),
  [71] = {.entry = {.count = 1, .reusable = false}}, SHIFT(207),
  [73] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [75] = {.entry = {.count = 1, .reusable = true}}, SHIFT(14),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(14),
  [79] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [81] = {.entry = {.count = 1, .reusable = false}}, SHIFT(209),
  [83] = {.entry = {.count = 1, .reusable = false}}, SHIFT(177),
  [85] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [87] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [89] = {.entry = {.count = 1, .reusable = true}}, SHIFT(385),
  [91] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [93] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0),
  [95] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(12),
  [98] = {.entry = {.count = 1, .reusable = true}}, SHIFT(19),
  [100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(19),
  [102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [108] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_shell_command, 1, 0, 0),
  [110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(17),
  [112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(17),
  [114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(31),
  [116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(346),
  [118] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__healthcheck_cmd_instruction, 1, 0, 0),
  [120] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__shell_fragment, 2, 0, 0),
  [122] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__shell_fragment, 2, 0, 0), SHIFT_REPEAT(17),
  [125] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__shell_fragment, 2, 0, 0), SHIFT_REPEAT(17),
  [128] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__shell_fragment, 2, 0, 0), SHIFT_REPEAT(209),
  [131] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__shell_fragment, 2, 0, 0), SHIFT_REPEAT(177),
  [134] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__shell_fragment, 2, 0, 0), SHIFT_REPEAT(385),
  [137] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_add_instruction_repeat1, 2, 0, 0),
  [139] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_add_instruction_repeat1, 2, 0, 0),
  [141] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_add_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(391),
  [144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(25),
  [146] = {.entry = {.count = 1, .reusable = false}}, SHIFT(25),
  [148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(32),
  [150] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__healthcheck_raw_command_repeat1, 2, 0, 0),
  [152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__healthcheck_raw_command_repeat1, 3, 0, 0),
  [154] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__healthcheck_raw_fragment, 1, 0, 0),
  [156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(24),
  [158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(24),
  [160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_env_pair, 2, 0, 1),
  [162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [164] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_env_pair, 2, 0, 1),
  [166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(167),
  [170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [172] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__healthcheck_raw_fragment, 2, 0, 0),
  [174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(26),
  [176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(26),
  [178] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__shell_fragment, 2, 0, 0), SHIFT_REPEAT(25),
  [181] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__shell_fragment, 2, 0, 0), SHIFT_REPEAT(25),
  [184] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__shell_fragment, 2, 0, 0), SHIFT_REPEAT(206),
  [187] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__shell_fragment, 2, 0, 0), SHIFT_REPEAT(207),
  [190] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__shell_fragment, 2, 0, 0), SHIFT_REPEAT(399),
  [193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__healthcheck_raw_fragment_repeat1, 2, 0, 0),
  [195] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__healthcheck_raw_fragment_repeat1, 2, 0, 0), SHIFT_REPEAT(26),
  [198] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__healthcheck_raw_fragment_repeat1, 2, 0, 0), SHIFT_REPEAT(26),
  [201] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__healthcheck_raw_fragment_repeat1, 2, 0, 0), SHIFT_REPEAT(206),
  [204] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__healthcheck_raw_fragment_repeat1, 2, 0, 0), SHIFT_REPEAT(207),
  [207] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__healthcheck_raw_fragment_repeat1, 2, 0, 0), SHIFT_REPEAT(399),
  [210] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_param, 2, 0, 5),
  [212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [214] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_param, 2, 0, 5),
  [216] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_param, 4, 0, 15),
  [218] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_param, 4, 0, 15),
  [220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(20),
  [222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(20),
  [224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(34),
  [226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(27),
  [230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(33),
  [232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(28),
  [234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(28),
  [236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(70),
  [238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(21),
  [240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(21),
  [242] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_run_instruction_repeat1, 2, 0, 0),
  [244] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_run_instruction_repeat1, 2, 0, 0),
  [246] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_run_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(322),
  [249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(88),
  [253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(164),
  [255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(42),
  [257] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mount_param, 4, 0, 15),
  [259] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mount_param, 4, 0, 15),
  [261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [263] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mount_param, 5, 0, 18),
  [265] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mount_param, 5, 0, 18),
  [267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(387),
  [269] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mount_param_param, 1, 0, 0),
  [271] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mount_param_param, 1, 0, 0),
  [273] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_mount_param_repeat1, 2, 0, 0),
  [275] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_mount_param_repeat1, 2, 0, 0),
  [277] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_mount_param_repeat1, 2, 0, 0), SHIFT_REPEAT(361),
  [280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__spaced_label_value, 1, 0, 0),
  [282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(54),
  [286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__spaced_env_value, 1, 0, 0),
  [290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__shell_double_quoted_fragment, 2, 0, 0),
  [296] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__shell_double_quoted_fragment, 2, 0, 0),
  [298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__shell_double_quoted_fragment, 3, 0, 0),
  [300] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__shell_double_quoted_fragment, 3, 0, 0),
  [302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__shell_single_quoted_fragment, 2, 0, 0),
  [304] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__shell_single_quoted_fragment, 2, 0, 0),
  [306] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mount_param_param, 3, 0, 0),
  [308] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mount_param_param, 3, 0, 0),
  [310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__shell_single_quoted_fragment, 3, 0, 0),
  [314] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__shell_single_quoted_fragment, 3, 0, 0),
  [316] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__shell_fragment_atom, 2, 0, 0),
  [318] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__shell_fragment_atom, 2, 0, 0),
  [320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__spaced_env_value_continuation, 2, 0, 0),
  [322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__healthcheck_raw_fragment_atom, 2, 0, 0),
  [324] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__healthcheck_raw_fragment_atom, 2, 0, 0),
  [326] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__spaced_label_value_continuation, 2, 0, 0),
  [328] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__spaced_label_value_repeat1, 2, 0, 0),
  [330] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__spaced_label_value_repeat1, 2, 0, 0), SHIFT_REPEAT(314),
  [333] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__spaced_label_value_repeat1, 2, 0, 0), SHIFT_REPEAT(54),
  [336] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__spaced_env_value_repeat1, 2, 0, 0),
  [338] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__spaced_env_value_repeat1, 2, 0, 0), SHIFT_REPEAT(316),
  [341] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__spaced_env_value_repeat1, 2, 0, 0), SHIFT_REPEAT(56),
  [344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [352] = {.entry = {.count = 1, .reusable = false}}, SHIFT(144),
  [354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(397),
  [360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(376),
  [362] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unquoted_string, 1, 0, 0),
  [364] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unquoted_string, 1, 0, 0),
  [366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(85),
  [376] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_unquoted_string_repeat1, 2, 0, 0),
  [378] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unquoted_string_repeat1, 2, 0, 0), SHIFT_REPEAT(315),
  [381] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_unquoted_string_repeat1, 2, 0, 0),
  [383] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unquoted_string_repeat1, 2, 0, 0), SHIFT_REPEAT(69),
  [386] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0),
  [388] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(70),
  [391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(268),
  [395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(273),
  [397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(74),
  [401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(276),
  [405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(80),
  [407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(80),
  [409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(283),
  [411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(78),
  [415] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_image_name_repeat1, 2, 0, 0),
  [417] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_image_name_repeat1, 2, 0, 0),
  [419] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_image_name_repeat1, 2, 0, 0), SHIFT_REPEAT(284),
  [422] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_image_name_repeat1, 2, 0, 0), SHIFT_REPEAT(76),
  [425] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_image_name, 2, 0, 0),
  [427] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_image_name, 2, 0, 0),
  [429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(272),
  [435] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_label_instruction, 2, 0, 0),
  [437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(90),
  [441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [443] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_double_quoted_string_repeat1, 2, 0, 0), SHIFT_REPEAT(268),
  [446] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_double_quoted_string_repeat1, 2, 0, 0),
  [448] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_double_quoted_string_repeat1, 2, 0, 0), SHIFT_REPEAT(80),
  [451] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_double_quoted_string_repeat1, 2, 0, 0), SHIFT_REPEAT(80),
  [454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [456] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_label_instruction_repeat1, 2, 0, 0),
  [458] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_label_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(374),
  [461] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_label_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(90),
  [464] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_label_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(143),
  [467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [469] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unquoted_string_repeat1, 2, 0, 0), SHIFT_REPEAT(317),
  [472] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unquoted_string_repeat1, 2, 0, 0), SHIFT_REPEAT(84),
  [475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [477] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unquoted_string_repeat1, 2, 0, 0), SHIFT_REPEAT(320),
  [480] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unquoted_string_repeat1, 2, 0, 0), SHIFT_REPEAT(86),
  [483] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_image_name, 1, 0, 0),
  [485] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_image_name, 1, 0, 0),
  [487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(205),
  [491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(89),
  [495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(219),
  [497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(367),
  [499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [501] = {.entry = {.count = 1, .reusable = false}}, SHIFT(91),
  [503] = {.entry = {.count = 1, .reusable = false}}, SHIFT(395),
  [505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(267),
  [507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(93),
  [511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(270),
  [513] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_path_with_heredoc, 2, 0, 0),
  [515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(98),
  [517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [519] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__user_name_or_group_repeat1, 2, 0, 0),
  [521] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__user_name_or_group_repeat1, 2, 0, 0), SHIFT_REPEAT(95),
  [524] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__user_name_or_group_repeat1, 2, 0, 0), SHIFT_REPEAT(306),
  [527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(75),
  [531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [533] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_path_with_heredoc, 1, 0, 0),
  [535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(94),
  [537] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_path_repeat1, 2, 0, 0),
  [539] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_path_repeat1, 2, 0, 0), SHIFT_REPEAT(98),
  [542] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_path_repeat1, 2, 0, 0), SHIFT_REPEAT(318),
  [545] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__imm_expansion, 2, 0, 0),
  [547] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__imm_expansion, 2, 0, 0),
  [549] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_body, 3, 0, 0),
  [551] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expansion_body, 3, 0, 0),
  [553] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_image_tag, 2, 0, 0),
  [555] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_image_tag, 2, 0, 0),
  [557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(102),
  [561] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_image_tag_repeat1, 2, 0, 0),
  [563] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_image_tag_repeat1, 2, 0, 0),
  [565] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_image_tag_repeat1, 2, 0, 0), SHIFT_REPEAT(310),
  [568] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_image_tag_repeat1, 2, 0, 0), SHIFT_REPEAT(102),
  [571] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__user_name_or_group, 1, 0, 0),
  [573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(104),
  [575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [577] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__user_name_or_group, 2, 0, 0),
  [579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(95),
  [581] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_path_repeat1, 2, 0, 0), SHIFT_REPEAT(105),
  [584] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_path_repeat1, 2, 0, 0), SHIFT_REPEAT(303),
  [587] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__immediate_user_name_or_group, 2, 0, 0),
  [589] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__immediate_user_name_or_group, 2, 0, 0), SHIFT_REPEAT(258),
  [592] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__immediate_user_name_or_group, 2, 0, 0), SHIFT_REPEAT(306),
  [595] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_user_instruction, 4, 0, 14),
  [597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [599] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_add_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(397),
  [602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_path, 2, 0, 0),
  [604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(105),
  [606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [608] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_image_digest, 2, 0, 0),
  [610] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_image_digest, 2, 0, 0),
  [612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(115),
  [616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_image_digest_repeat1, 2, 0, 0),
  [618] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_image_digest_repeat1, 2, 0, 0),
  [620] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_image_digest_repeat1, 2, 0, 0), SHIFT_REPEAT(312),
  [623] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_image_digest_repeat1, 2, 0, 0), SHIFT_REPEAT(115),
  [626] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_path, 1, 0, 0),
  [628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(109),
  [630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [636] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_image_spec, 1, 0, 1),
  [638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__stopsignal_value_repeat1, 2, 0, 0),
  [644] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__stopsignal_value_repeat1, 2, 0, 0), SHIFT_REPEAT(119),
  [647] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__stopsignal_value_repeat1, 2, 0, 0), SHIFT_REPEAT(260),
  [650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_image_alias, 2, 0, 0),
  [658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [662] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_add_instruction_repeat2, 2, 0, 0), SHIFT_REPEAT(144),
  [665] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_add_instruction_repeat2, 2, 0, 0), SHIFT_REPEAT(302),
  [668] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_add_instruction_repeat2, 2, 0, 0), SHIFT_REPEAT(376),
  [671] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__stopsignal_value, 2, 0, 0),
  [673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [677] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_image_alias_repeat1, 2, 0, 0),
  [679] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_image_alias_repeat1, 2, 0, 0), SHIFT_REPEAT(319),
  [682] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_image_alias_repeat1, 2, 0, 0), SHIFT_REPEAT(128),
  [685] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__stopsignal_value, 1, 0, 0),
  [687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(116),
  [691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(116),
  [693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [695] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expansion, 2, 0, 0),
  [697] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expansion, 2, 0, 0),
  [699] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_image_alias, 1, 0, 0),
  [701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [703] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_expose_instruction_repeat1, 2, 0, 0),
  [705] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expose_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(293),
  [708] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expose_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(146),
  [711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [713] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expose_instruction, 2, 0, 0),
  [715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(373),
  [721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(162),
  [723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [729] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_copy_instruction, 3, 0, 0),
  [731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [735] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expose_port, 1, 0, 0),
  [737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [743] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_single_quoted_string_repeat1, 2, 0, 0),
  [745] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_single_quoted_string_repeat1, 2, 0, 0), SHIFT_REPEAT(152),
  [748] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_single_quoted_string_repeat1, 2, 0, 0), SHIFT_REPEAT(152),
  [751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(275),
  [753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(152),
  [755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [757] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_add_instruction, 3, 0, 0),
  [759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [761] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_env_instruction_repeat1, 2, 0, 0),
  [763] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_env_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(393),
  [766] = {.entry = {.count = 1, .reusable = false}}, SHIFT(277),
  [768] = {.entry = {.count = 1, .reusable = false}}, SHIFT(286),
  [770] = {.entry = {.count = 1, .reusable = false}}, SHIFT(154),
  [772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [774] = {.entry = {.count = 1, .reusable = false}}, SHIFT(382),
  [776] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_env_instruction, 2, 0, 0),
  [778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(393),
  [780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(174),
  [782] = {.entry = {.count = 1, .reusable = false}}, SHIFT(165),
  [784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [786] = {.entry = {.count = 1, .reusable = false}}, SHIFT(175),
  [788] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_path_repeat1, 2, 0, 0), SHIFT_REPEAT(166),
  [791] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_path_repeat1, 2, 0, 0), SHIFT_REPEAT(301),
  [794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(245),
  [796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(169),
  [798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [800] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_add_instruction, 4, 0, 0),
  [802] = {.entry = {.count = 1, .reusable = false}}, SHIFT(246),
  [804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(274),
  [806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(160),
  [808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [810] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_copy_instruction, 4, 0, 0),
  [812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
  [814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [816] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_single_quoted_string, 2, 0, 0),
  [818] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_single_quoted_string, 3, 0, 0),
  [820] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_shell_command, 2, 0, 0),
  [822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [826] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__spaced_label_value, 2, 0, 0),
  [828] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__healthcheck_raw_command_repeat1, 2, 0, 0), SHIFT_REPEAT(31),
  [831] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_run_instruction, 4, 0, 0),
  [833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__spaced_label_value_repeat2, 2, 0, 0),
  [839] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__spaced_label_value_repeat2, 2, 0, 0), SHIFT_REPEAT(73),
  [842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(358),
  [844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [846] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_run_instruction, 2, 0, 0),
  [848] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_run_instruction, 3, 0, 0),
  [850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [852] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_label_pair, 3, 0, 12),
  [854] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_json_string_repeat1, 2, 0, 0),
  [856] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_json_string_repeat1, 2, 0, 0), SHIFT_REPEAT(200),
  [859] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__continued_spaced_env_value, 1, 0, 0),
  [861] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__continued_spaced_label_value, 1, 0, 0),
  [863] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__spaced_env_value, 2, 0, 0),
  [865] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__spaced_env_value_repeat2, 2, 0, 0),
  [867] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__spaced_env_value_repeat2, 2, 0, 0), SHIFT_REPEAT(83),
  [870] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_double_quoted_string, 2, 0, 0),
  [872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [880] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__continued_spaced_label_value, 2, 0, 0),
  [882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(55),
  [888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(57),
  [892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [894] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_add_instruction, 5, 0, 0),
  [896] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_copy_instruction, 5, 0, 0),
  [898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [900] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__continued_spaced_env_value, 2, 0, 0),
  [902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(114),
  [904] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_image_spec, 2, 0, 6),
  [906] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_double_quoted_string, 3, 0, 0),
  [908] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__shell_double_quoted_fragment_repeat1, 2, 0, 0), SHIFT_REPEAT(220),
  [911] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__shell_double_quoted_fragment_repeat1, 2, 0, 0),
  [913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(62),
  [915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(71),
  [917] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__shell_single_quoted_fragment_repeat1, 2, 0, 0), SHIFT_REPEAT(223),
  [920] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__shell_single_quoted_fragment_repeat1, 2, 0, 0),
  [922] = {.entry = {.count = 1, .reusable = false}}, SHIFT(349),
  [924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [926] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_run_instruction_repeat2, 2, 0, 0),
  [928] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_run_instruction_repeat2, 2, 0, 0), SHIFT_REPEAT(266),
  [931] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_volume_instruction_repeat1, 2, 0, 0),
  [933] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_volume_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(173),
  [936] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_add_instruction_repeat2, 2, 0, 0),
  [938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4),
  [940] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_volume_instruction, 3, 0, 0),
  [942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(173),
  [944] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arg_pair, 1, 0, 1),
  [946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [948] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_arg_instruction_repeat1, 2, 0, 0),
  [950] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_arg_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(326),
  [953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__healthcheck_raw_command, 2, 0, 0),
  [959] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_source_file_repeat2, 2, 0, 0),
  [961] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat2, 2, 0, 0), SHIFT_REPEAT(6),
  [964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5),
  [966] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__healthcheck_raw_command_repeat1, 2, 0, 0), SHIFT_REPEAT(32),
  [969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2),
  [971] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_volume_instruction, 2, 0, 0),
  [973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(103),
  [975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [977] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__env_single_quoted_string_with_trailing_quote, 1, 0, 0),
  [979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arg_instruction, 2, 0, 0),
  [983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [985] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expose_port, 2, 0, 0),
  [987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__immediate_user_name_or_group, 1, 0, 0),
  [989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__healthcheck_raw_command, 1, 0, 0),
  [991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [995] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_block, 2, 0, 0),
  [997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [1001] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_json_string_array, 2, 0, 0),
  [1003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [1005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [1007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [1009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(407),
  [1011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(134),
  [1013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(366),
  [1015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [1017] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arg_pair, 3, 0, 17),
  [1019] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_heredoc_block_repeat1, 2, 0, 0), SHIFT_REPEAT(381),
  [1022] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_heredoc_block_repeat1, 2, 0, 0),
  [1024] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_image_spec, 2, 0, 7),
  [1026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [1028] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_from_instruction, 3, 0, 0),
  [1030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(241),
  [1032] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_image_spec, 3, 0, 11),
  [1034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [1036] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_user_instruction, 2, 0, 4),
  [1038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [1040] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_json_string_array, 4, 0, 0),
  [1042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_block, 3, 0, 0),
  [1044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(192),
  [1046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [1048] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__env_single_quoted_string_with_trailing_quote, 2, 0, 0),
  [1050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(402),
  [1052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(237),
  [1054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [1056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_json_string_array, 3, 0, 0),
  [1058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [1060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_from_instruction, 2, 0, 0),
  [1062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [1064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [1066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
  [1068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [1070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [1072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [1074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [1076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [1078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [1080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [1082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [1084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [1086] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_json_string_array_repeat1, 2, 0, 0), SHIFT_REPEAT(323),
  [1089] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_json_string_array_repeat1, 2, 0, 0),
  [1091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [1093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(410),
  [1095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [1097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [1099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [1101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [1103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [1105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [1107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [1109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [1111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(110),
  [1113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(413),
  [1115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [1117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(414),
  [1119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [1121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [1123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [1125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(155),
  [1127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [1129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [1131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [1133] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_env_pair, 3, 0, 13),
  [1135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(48),
  [1137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(362),
  [1139] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_workdir_instruction, 2, 0, 0),
  [1141] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__spaced_env_pair, 3, 0, 13),
  [1143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [1145] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_onbuild_instruction, 2, 0, 0),
  [1147] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__spaced_label_pair, 2, 0, 8),
  [1149] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_stopsignal_instruction, 2, 0, 0),
  [1151] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_env_instruction, 2, 0, 3),
  [1153] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_label_instruction, 2, 0, 2),
  [1155] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__healthcheck_cmd_instruction, 2, 0, 0),
  [1157] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_healthcheck_instruction, 2, 0, 0),
  [1159] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__spaced_env_pair, 2, 0, 9),
  [1161] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_from_instruction, 4, 0, 10),
  [1163] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_shell_instruction, 2, 0, 0),
  [1165] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entrypoint_instruction, 2, 0, 0),
  [1167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_from_instruction, 5, 0, 16),
  [1169] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_maintainer_instruction, 2, 0, 0),
  [1171] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cmd_instruction, 2, 0, 0),
  [1173] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cross_build_instruction, 2, 0, 0),
  [1175] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_json_string, 3, 0, 0),
  [1177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(39),
  [1179] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_healthcheck_instruction, 3, 0, 0),
  [1181] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_add_instruction, 2, 0, 0),
  [1183] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_copy_instruction, 2, 0, 0),
  [1185] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_json_string, 2, 0, 0),
  [1187] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__spaced_label_pair, 3, 0, 12),
  [1189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [1191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [1193] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [1195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [1197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [1199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [1201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [1203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [1205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [1207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [1209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [1211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [1213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [1215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [1217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [1219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [1221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(137),
  [1223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(50),
  [1225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(147),
  [1227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [1229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [1231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [1233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(153),
  [1235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(29),
  [1237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(197),
  [1239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(112),
  [1241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(100),
  [1243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [1245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(113),
  [1247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [1249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [1251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(142),
  [1253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [1255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [1257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(369),
  [1259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(377),
  [1261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [1263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [1265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [1267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [1269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(390),
  [1271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(392),
  [1273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(394),
  [1275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [1277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(398),
  [1279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(400),
  [1281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(401),
  [1283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token_comment = 0,
  ts_external_token_line_continuation = 1,
  ts_external_token_required_line_continuation = 2,
  ts_external_token__newline = 3,
  ts_external_token__invalid_json_array_shell_command = 4,
  ts_external_token_heredoc_marker = 5,
  ts_external_token__heredoc_line = 6,
  ts_external_token_heredoc_end = 7,
  ts_external_token__heredoc_nl = 8,
  ts_external_token_error_sentinel = 9,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token_comment] = sym_comment,
  [ts_external_token_line_continuation] = sym_line_continuation,
  [ts_external_token_required_line_continuation] = sym_required_line_continuation,
  [ts_external_token__newline] = sym__newline,
  [ts_external_token__invalid_json_array_shell_command] = sym__invalid_json_array_shell_command,
  [ts_external_token_heredoc_marker] = sym_heredoc_marker,
  [ts_external_token__heredoc_line] = sym__heredoc_line,
  [ts_external_token_heredoc_end] = sym_heredoc_end,
  [ts_external_token__heredoc_nl] = sym__heredoc_nl,
  [ts_external_token_error_sentinel] = sym_error_sentinel,
};

static const bool ts_external_scanner_states[14][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_continuation] = true,
    [ts_external_token_required_line_continuation] = true,
    [ts_external_token__newline] = true,
    [ts_external_token__invalid_json_array_shell_command] = true,
    [ts_external_token_heredoc_marker] = true,
    [ts_external_token__heredoc_line] = true,
    [ts_external_token_heredoc_end] = true,
    [ts_external_token__heredoc_nl] = true,
    [ts_external_token_error_sentinel] = true,
  },
  [2] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_continuation] = true,
    [ts_external_token__newline] = true,
  },
  [3] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_continuation] = true,
  },
  [4] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_continuation] = true,
    [ts_external_token__newline] = true,
    [ts_external_token_heredoc_marker] = true,
  },
  [5] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_continuation] = true,
    [ts_external_token_heredoc_marker] = true,
  },
  [6] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_continuation] = true,
    [ts_external_token__invalid_json_array_shell_command] = true,
    [ts_external_token_heredoc_marker] = true,
  },
  [7] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_continuation] = true,
    [ts_external_token_required_line_continuation] = true,
    [ts_external_token__newline] = true,
    [ts_external_token_heredoc_marker] = true,
    [ts_external_token__heredoc_nl] = true,
  },
  [8] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_continuation] = true,
    [ts_external_token_required_line_continuation] = true,
    [ts_external_token__newline] = true,
    [ts_external_token_heredoc_marker] = true,
  },
  [9] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_continuation] = true,
    [ts_external_token_required_line_continuation] = true,
    [ts_external_token__newline] = true,
  },
  [10] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_continuation] = true,
    [ts_external_token__newline] = true,
    [ts_external_token__heredoc_nl] = true,
  },
  [11] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_continuation] = true,
    [ts_external_token_required_line_continuation] = true,
  },
  [12] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_continuation] = true,
    [ts_external_token_required_line_continuation] = true,
    [ts_external_token__newline] = true,
    [ts_external_token__heredoc_nl] = true,
  },
  [13] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_continuation] = true,
    [ts_external_token__heredoc_line] = true,
    [ts_external_token_heredoc_end] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_containerfile_external_scanner_create(void);
void tree_sitter_containerfile_external_scanner_destroy(void *);
bool tree_sitter_containerfile_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_containerfile_external_scanner_serialize(void *, char *);
void tree_sitter_containerfile_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_containerfile(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_containerfile_external_scanner_create,
      tree_sitter_containerfile_external_scanner_destroy,
      tree_sitter_containerfile_external_scanner_scan,
      tree_sitter_containerfile_external_scanner_serialize,
      tree_sitter_containerfile_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "containerfile",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 0,
      .minor_version = 8,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
